Package: ELAplus
Language: en-US
Type: Package
Title: Energy Landscape Analysis for Ecological Communities
Version: 1.0.2
Date: 2026-05-15
Author: Kenta Suzuki [aut, cre, cph],
  Sotaro Takano [aut],
  Hiroaki Fujita [aut],
  Zeus Sato [ctb],
  Fumihiko Ayabe [ctb]
Authors@R: c(
    person("Kenta", "Suzuki",
           email = "kenta.suzuki.zk@riken.jp",
           role = c("aut","cre","cph")),
    person("Sotaro", "Takano",
           email = "sotaro.takano@riken.jp",
           role = "aut"),
    person("Hiroaki", "Fujita",
           email = "fujita.hiroaki.3x@kyoto-u.ac.jp",
           role = "aut"),
    person("Zeus", "Sato", role = "ctb"),
    person("Fumihiko", "Ayabe", role = "ctb")
           )
Maintainer: Kenta Suzuki <kenta.suzuki.zk@riken.jp>
Encoding: UTF-8
License: BSD_3_clause + file LICENSE
Description: Energy landscape analysis (ELA) is a systematic method for analyzing an energy landscape represented as a weighted network. This R package is especially designed to analyze ecological dynamics, i.e., transitions in community compositions. For details of the analysis framework, please visit Suzuki K et al. (2021) <doi:10.1002/ecm.1469>.
Depends: R (>= 4.1.0)
Imports: Rcpp (>= 1.0.11), doParallel, RColorBrewer, foreach, dplyr,
        magrittr, stringdist, gtools, ggplot2, purrr, rlang, tidyr,
        data.table, stringr, igraph
Suggests: cluster, plot3D, fpc, quanteda, slam, stm, topicmodels,
        tidyverse, Matrix, glmnet, gsubfn, zoo, ggsci, tidygraph,
        vegan, knitr, rmarkdown
VignetteBuilder: knitr
LinkingTo: Rcpp, RcppArmadillo
NeedsCompilation: yes
RoxygenNote: 7.3.3
LazyData: true
Packaged: 2026-05-15 08:49:40 UTC; str_mac
Repository: CRAN
Date/Publication: 2026-05-28 12:20:11 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-28 16:19:22 UTC; unix
Archs: ELAplus.so.dSYM

Package: ELAplus
Language: en-US
Type: Package
Title: Energy Landscape Analysis for Ecological Communities
Version: 1.0.2
Date: 2026-05-15
Author: Kenta Suzuki [aut, cre, cph],
  Sotaro Takano [aut],
  Hiroaki Fujita [aut],
  Zeus Sato [ctb],
  Fumihiko Ayabe [ctb]
Authors@R: c(
    person("Kenta", "Suzuki",
           email = "kenta.suzuki.zk@riken.jp",
           role = c("aut","cre","cph")),
    person("Sotaro", "Takano",
           email = "sotaro.takano@riken.jp",
           role = "aut"),
    person("Hiroaki", "Fujita",
           email = "fujita.hiroaki.3x@kyoto-u.ac.jp",
           role = "aut"),
    person("Zeus", "Sato", role = "ctb"),
    person("Fumihiko", "Ayabe", role = "ctb")
           )
Maintainer: Kenta Suzuki <kenta.suzuki.zk@riken.jp>
Encoding: UTF-8
License: BSD_3_clause + file LICENSE
Description: Energy landscape analysis (ELA) is a systematic method for analyzing an energy landscape represented as a weighted network. This R package is especially designed to analyze ecological dynamics, i.e., transitions in community compositions. For details of the analysis framework, please visit Suzuki K et al. (2021) <doi:10.1002/ecm.1469>.
Depends: R (>= 4.1.0)
Imports: Rcpp (>= 1.0.11), doParallel, RColorBrewer, foreach, dplyr,
        magrittr, stringdist, gtools, ggplot2, purrr, rlang, tidyr,
        data.table, stringr, igraph
Suggests: cluster, plot3D, fpc, quanteda, slam, stm, topicmodels,
        tidyverse, Matrix, glmnet, gsubfn, zoo, ggsci, tidygraph,
        vegan, knitr, rmarkdown
VignetteBuilder: knitr
LinkingTo: Rcpp, RcppArmadillo
NeedsCompilation: yes
RoxygenNote: 7.3.3
LazyData: true
Packaged: 2026-05-15 08:49:40 UTC; str_mac
Repository: CRAN
Date/Publication: 2026-05-28 12:20:11 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-28 13:31:18 UTC; unix
Archs: ELAplus.so.dSYM

# Generated by roxygen2: do not edit by hand

export(Bi)
export(DisconnectivityGraph)
export(ELA)
export(ELPruning)
export(Energy)
export(Findbp)
export(Formatting)
export(GELAObj)
export(GradELA)
export(GradientDescent)
export(HeatBath)
export(MinTippingPath)
export(OnestepHBS)
export(PCplot)
export(SSentropy)
export(Stability)
export(SteepestDescent)
export(bin2id)
export(bootstrap_ELA)
export(bootstrap_SA)
export(deltaE_histogram)
export(gStability)
export(hb.paramgen)
export(id2bin)
export(plotSAtest)
export(runSA)
export(sa2params)
export(saEndpoint)
export(showDG)
export(showGELA3D)
export(showIntrGraph)
export(showSSD)
export(showSSD_ggplot)
export(sstable)
export(tptable)
import(foreach)
importFrom(RColorBrewer,brewer.pal)
importFrom(Rcpp,sourceCpp)
importFrom(data.table,transpose)
importFrom(doParallel,registerDoParallel)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,group_map)
importFrom(dplyr,left_join)
importFrom(dplyr,select)
importFrom(foreach,"%do%")
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(ggplot2,aes)
importFrom(ggplot2,coord_cartesian)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_path)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,geom_ribbon)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_color_viridis_c)
importFrom(ggplot2,scale_size)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_classic)
importFrom(ggplot2,xlab)
importFrom(graphics,lines)
importFrom(igraph,components)
importFrom(igraph,graph_from_adjacency_matrix)
importFrom(magrittr,"%>%")
importFrom(parallel,clusterCall)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(purrr,map)
importFrom(rlang,.data)
importFrom(stats,embed)
importFrom(stats,median)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(stats,smooth.spline)
importFrom(stringdist,stringdist)
importFrom(stringr,str_c)
importFrom(tidyr,all_of)
importFrom(tidyr,gather)
importFrom(tidyr,pivot_longer)
importFrom(utils,combn)
importFrom(utils,head)
useDynLib(ELAplus, .registration=TRUE)

## ----setup, include=FALSE-----------------------------------------------------
library(knitr)
opts_chunk$set(
  fig.align = "center",
  out.extra = 'style="max-width:100%; overflow-x:auto; white-space: nowrap;"',
  tidy = FALSE
)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  message  = FALSE,
  warning  = FALSE
)

## -----------------------------------------------------------------------------
# load package
library(ELAplus)
# load test data set
# species abundance table
data("baseabtable")
head(baseabtable)

# metadata (environmental factors)
data("basemetadata")
head(basemetadata)

## -----------------------------------------------------------------------------
# load package
library(ELAplus)
# Generate random parameters
hb_params <- hb.paramgen(24, ne = 2) # the number of environmental factors
h.act <- hb_params[[1]]
j.act <- hb_params[[2]]
g.act <- hb_params[[3]]

# Simulate community composition data
rand_data <- HeatBath(100, 512, h.act, j.act, g = g.act)
rand_mat   <- rand_data[[1]]
rand_enmat <- rand_data[[2]]


## -----------------------------------------------------------------------------
formatted <- Formatting(baseabtable = baseabtable, basemetadata = basemetadata)

## ----eval=FALSE---------------------------------------------------------------
# formatted <- Formatting(baseabtable = baseabtable, basemetadata = basemetadata,
#                         normalize = 1, parameters = c(0.01, 0.01, 0.99))
# #> Processed 256 samples.
# #> Relative abundance threshold = 0.01
# #> Occurrence threshold (lower) = 0.01
# #> Occurrence threshold (upper) = 0.99
# #> Selected  16  out of  16 species.
# ocvecs <- formatted[[1]]  # community composition binary matrix
# abvecs <- formatted[[2]]  # community abundance matrix (not binarized)
# envecs <- formatted[[3]]  # environmental variables (optional)

## ----echo=FALSE, out.width="80%"----------------------------------------------
knitr::include_graphics("figures/ParameterMatrix.png")

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/EPMEM_abstract2.png")

## ----eval=FALSE---------------------------------------------------------------
# bpresult <- Findbp(ocvecs, rep = 2, threads = 2,
#                    cvmode="10fold",nrepeat = 10, we=c(0.001),totalit=4000,
#                    lmd=c(1e-05,1e-03,0.01),maxlrs = c(0.005),
#                    tol=0.03, runadamW=TRUE,sparse=TRUE,fastfitting=TRUE)
# #Start hyper-parameter search:
# #There are 300 tasks
# #Finished 300 tasks
# #SA: elapsed time 15.08 sec
# 
# bp <- bpresult[[1]]  # typically best bpresult
# bpm <- as.numeric(unlist(strsplit(names(bp)[1], split = "_")))
# allresults <- bpresult[[2]]

## ----plotSAtest, eval=FALSE---------------------------------------------------
# plotSAtest(allresults)
# # "id" corresponds to hyperparameter: lmd_we_maxlr

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/plotSAtest.png")

## ----eval=FALSE---------------------------------------------------------------
# sa <- runSA(
#   ocvecs,
#   enmat   = NULL,
#   rep     = 32,
#   getall  = FALSE,
#   lambda  = bpm[[1]],
#   we      = bpm[[2]],
#   maxlr = bpm[[3]],
#   totalit = bpm[[4]]
# )
# #> Start parameter fitting:
# #> .SA: elapsed time 0.13 sec

## ----eval=FALSE---------------------------------------------------------------
# ela <- ELA(
#   sa,
#   env            = NULL,
#   threads        = 1,
#   SS.itr         = 20000,
#   FindingTip.itr = 10000,
#   reporting      = TRUE
# )
# #> Start ELA:
# #> 7 stable states were found.
# #> Checking 21 tipping points.
# #> converting...
# #> ELA: elapsed time 2.19 sec

## ----eval=FALSE---------------------------------------------------------------
# ela[[1]] # stable state IDs
# #> [1] "09x" "01t" "EWB" "EY8" "1uV" "1mL" "17s"
# ela[[2]] # stable state energies
# #> [1] -11.099418 -10.297931  -9.544050  -8.961634  -8.005040  -6.440385  -6.269140

## ----eval=FALSE---------------------------------------------------------------
# elap <- ELPruning(ela, th=0.1, type="relative")
# #> Start pruning:
# #> *...
# #> ELPruning: elapsed time 0.05 sec

## ----eval=FALSE---------------------------------------------------------------
# elap[[1]][[1]] # updated stable state IDs
# #> [1] "09x" "EWB" "EY8" "1uV"
# elap[[1]][[2]] # updated stable state energies
# #> [1] -11.099418  -9.544050  -8.961634  -8.005040

## ----eval=FALSE---------------------------------------------------------------
# elap[[2]]
# #>  ss.before.pruning ss.after.pruning
# #> 1               09x              09x
# #> 2               01t              09x
# #> 3               EWB              EWB
# #> 4               EY8              EY8
# #> 5               1uV              1uV
# #> 6               1mL              09x
# #> 7               17s              09x

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/deltaE_scheme.png")

## ----eval=FALSE---------------------------------------------------------------
# elap <- ELPruning(ela, th=0.2, type="relative")
# #> Start pruning:
# #> *...
# #> ELPruning: elapsed time 0.05 sec
# elap[[1]][[1]] # updated stable state IDs
# #> [1] "09x" "EWB" "1uV"

## ----showDG, eval=FALSE-------------------------------------------------------
# # Before pruning
# showDG(ela, ocvecs, "example")

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/DGexample.png")

## ----showDG_pruned, eval=FALSE------------------------------------------------
# # After pruning
# showDG(elap[[1]], ocvecs, "example")

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/DGexample_pruned.png")

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/deltaE_hist.png")

## ----showDG_pvalue, eval=FALSE------------------------------------------------
# bs_params <- bootstrap_SA(ocvecs, rep = 16, threads=8,　bootstrap.it = 1000,
#                           lambda  = bpm[[1]],we = bpm[[2]],
#                           maxlr = bpm[[3]], totalit = bpm[[4]])
# 
# bootstrap <- bootstrap_ELA(bs_params,ocvecs,ela=ela,threads = 1)
# ela_rep <- bootstrap[[1]]
# bootstrap_ene <- bootstrap[[2]]
# showDG(ela_rep,ocvecs,bootstrap_ene = bootstrap_ene, min_pval = 0.01,
#        scale_labels = c(0.05,0.1,0.2,0.5))

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/DGexample_pval.png")

## ----showDG_IQR_pruned, eval=FALSE--------------------------------------------
# elap_rep <- ELPruning(ela, type = "bootstrap",bootstrap_ene=bootstrap_ene,lower_qr = 0.25)
# showDG(ela_rep,ocvecs,bootstrap_ene = bootstrap_ene, min_pval = 0.01,
#        scale_labels = c(0.05,0.1,0.2,0.5))

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/DGexample_pval_pruned.png")

## ----showDG_manual, eval=FALSE------------------------------------------------
# grobj_to_plot <- showDG(ela_rep,ocvecs,bootstrap_ene = bootstrap_ene,getGraphObj = TRUE)
# grobj_point <- grobj_to_plot[grobj_to_plot$point_str != "",]
# g <- ggplot(
#   grobj_to_plot,
#   aes(x=.data$nodes2xposi, y=.data$energy, label=.data$point_str)) +
#   xlab("") + geom_text(hjust=0.75, vjust=2, aes(fontface=2)) + geom_path() +
#   geom_point(
#     data = grobj_point,
#     mapping = aes(
#       x = .data$nodes2xposi,
#       y = .data$energy,
#       color = .data$energy,
#       shape = .data$type,
#     ),size = 4
#   ) +
#   scale_color_viridis_c(option = "plasma",alpha=0.8,
#                         direction=-1,na.value = "grey") +
#   coord_cartesian(xlim=c(0.75,7.5), ylim=c(-5.8,-11.5))
# plot(g)

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/DG_manual.png")

## ----ShowIntrGraph, eval=FALSE------------------------------------------------
# showIntrGraph(
#   elap[[1]],
#   sa,
#   th = 0.01,
#   annot_adj = c(0.75, 1.00)
# )

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/IntrGraph.png")

## ----eval=FALSE---------------------------------------------------------------
# bpresult <- Findbp(ocvecs, enmat = envecs, rep = 2, threads = 2,
#                    cvmode="10fold",nrepeat = 10, we=c(0.001),totalit=4000,
#                    lmd=c(1e-05,1e-03,0.01),maxlrs = c(0.005),
#                    tol=0.03, runadamW=TRUE,sparse=TRUE,fastfitting=TRUE)
# bp <- bpresult[[1]]
# bpm <- as.numeric(unlist(strsplit(names(bp)[1], split = "_")))
# 
# sa <- runSA(
#   ocvecs,
#   enmat   = envecs,
#   rep     = 16,
#   getall  = FALSE,
#   lambda = bpm[[1]],
#   we = bpm[[2]],
#   maxlr = bpm[[3]],
#   totalit = bpm[[4]]
# )
# 

## ----eval=FALSE---------------------------------------------------------------
# # This process takes few minutes with a small number of threads, and here pre-computed results are shown.
# gela <- GradELA(
#   sa      = sa,
#   eid     = "factor.1",
#   enmat   = envecs,
#   env     = NULL,
#   range   = c(-1, 1),
#   steps   = 32,
#   th = 0.05,
#   threads = 8,
#   pruning_type = "relative",
#   bs_params = bs_params,
#   ocmat = ocvecs)

## ----eval=FALSE---------------------------------------------------------------
# # This process takes few minutes with a small number of threads, and here pre-computed results are shown.
# bs_params <- bootstrap_SA(ocvecs, enmat = envecs,rep = 16, threads=8,bootstrap.it = 1000,
#                           lambda  = bpm[[1]],we = bpm[[2]],
#                           maxlr = bpm[[3]], totalit = bpm[[4]])
# 
# gela <- GradELA(
#   sa      = sa,
#   eid     = "factor.1",
#   enmat   = envecs,
#   env     = NULL,
#   range   = c(-1, 1),
#   steps   = 32,
#   lower_qr = 0.25,
#   threads = 8,
#   pruning_type = "bootstrap",
#   bs_params = bs_params,
#   ocmat = ocvecs)

## ----eval=FALSE---------------------------------------------------------------
# gela[[1]][[1]][[1]][[1]] # stable state IDs at 1st environmental parameter
# # [1] "0Ht" "EWB" "1uV"
# gela[[1]][[20]][[1]][[1]] # stable state IDs at 20th environmental parameter
# # [1] "09x" "EWB"

## ----showSSD, eval=FALSE, fig.width=8, fig.height=6, message=FALSE------------
# # GradELA result (ELPruning with type="relative").
# showSSD(gela)

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/SSD.png")

## ----showSSD_bootstrap, eval=FALSE, fig.width=8, fig.height=6, message=FALSE----
# # GradELA result (ELPruning with type="bootstrap").
# showSSD(gela)

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/SSD_pruned.png")

## ----showSSD_ggplot, eval=FALSE, fig.width=8, fig.height=6, message=FALSE-----
# ssd_df <- showSSD_ggplot(gela,plot = FALSE,getSSDobj = TRUE)
# g <- ggplot(ssd_df, aes(x = env, y = Energy, color = SS_ID)) +
#             geom_point(shape = 19) +
#             geom_line(aes(group = SS_ID))
# 

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/SSD_ggplot.png")

## ----GELA3D-demo, eval=FALSE, fig.width=8, fig.height=6, message=FALSE--------
# # This process takes few minutes, and here pre-computed result is drawn.
# gelaobj <- GELAObj(gela, sa=sa,threads=2)
# showGELA3D(gelaobj,theta = 30,phi = 30)

## ----echo=FALSE, out.width="95%"----------------------------------------------
knitr::include_graphics("figures/GELA3D.png")

#' Stochastic approximation for parameter fitting
#'
#' Estimate species interaction parameters and (optionally) environmental
#' preference parameters by stochastic approximation (SA). The function runs
#' multiple SA optimizations in parallel (when \code{threads > 1}) using
#' \pkg{foreach}.
#'
#' @param ocmat Numeric matrix (0/1). Rows are samples and columns are species.
#' @param enmat Optional numeric matrix. Environmental variables for each sample.
#'   Should be normalized/standardized beforehand.
#' @param rep Integer. Number of parallel runs (replicates) for optimization.
#' @param threads Integer. Number of cores used when running in parallel.
#' @param we Numeric. Weight decay parameter for AdamW.
#' @param totalit Integer. Total number of optimization iterations.
#' @param lambda Numeric. Sparsity penalty parameter (used when \code{sparse = TRUE}).
#' @param intv Integer. Interval of iterations at which intermediate results are recorded.
#' @param runadamW Logical. If \code{TRUE}, use AdamW-style optimization.
#' @param maxlr Numeric. Maximum learning rate.
#' @param sparse Logical. If \code{TRUE}, apply sparsity constraint.
#' @param getall Logical. If \code{TRUE}, return all replicate trajectories.
#' @param reporting Logical. If \code{TRUE}, print progress messages.
#'
#' @return If \code{getall = TRUE}, a list of SA trajectories (length \code{rep}).
#' Otherwise, a list with two elements, \code{c(params, enmat)} is returned.
#' \code{params} is the concatenated matrix of the estimated parameter set:
#' \describe{
#'   \item{h}{Implicit environmental effects on species occurrence}
#'   \item{J}{Species-species interaction matrix}
#'   \item{g}{(If \code{enmat} is not \code{NULL}) Explicit effects of the environmental factors.}
#' }
#'
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @importFrom Rcpp sourceCpp
#' @import foreach
#' @export
runSA <- function(ocmat, enmat=NULL,  rep=128, threads=1, we=0.001, 
                  totalit=1000, lambda=0.01,intv=100, runadamW=TRUE,
                  maxlr = 0.005, sparse=TRUE, getall=FALSE, reporting=TRUE){
    #if(!is.matrix(ocmat)){ocmat <- as.matrix(ocmat)}
    if(!sparse){lambda <- 0}
  
    #if(!runadamW){we <- 0}
  
    if (threads > 1) {
      cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
      parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
      parallel::clusterSetRNGStream(cluster)
      doParallel::registerDoParallel(cluster)
      on.exit(parallel::stopCluster(cluster), add = TRUE)
    }

    if(is.null(enmat)){

    	## ============================================== ##
        ## -- without explicit variables

        if(reporting){cat('Start parameter fitting:\n')}
        s <- proc.time()[3]
        if(threads>1){fittingMat <- foreach(i = 1:rep, .packages="ELAplus") %dopar% {
            SAres <- SA(ocData = ocmat, we = we, totalit = totalit, 
                        runadamW = runadamW, sparse = sparse,
                        lambda = lambda, intv = intv, maxlearningrate = maxlr)
            return(SAres)
        }}else{fittingMat <- foreach(i = 1:rep) %do% {
	      if(reporting){cat('.')}
            SAres <- SA(ocData = ocmat, we = we, totalit = totalit, 
                        runadamW = runadamW, sparse = sparse,
                        lambda = lambda, intv = intv,maxlearningrate = maxlr)
            return(SAres)
        }}

        if(reporting){cat(sprintf('SA: elapsed time %.2f sec\n\n', proc.time()[3]-s))}
        ## ============================================== ##
      
        if(getall){
          return(fittingMat)
        }else{
            return(list(saEndpoint(fittingMat,ocmat),enmat))}
          
	    }else{

        ## ============================================== ##
        ## -- with explicit variables
	      #if(!is.matrix(enmat)){enmat <- as.matrix(enmat)}

        if(reporting){cat('Start parameter fitting\n')}
        s <- proc.time()[3]
        if(threads>1){fittingMat <- foreach(i = 1:rep, .packages="ELAplus") %dopar% {
            SAres <- fullSA(ocData = ocmat, envData=as.matrix(enmat), 
                            runadamW = runadamW, sparse = sparse, we = we, 
                            totalit = totalit, lambda = lambda, intv = intv,
                            maxlearningrate = maxlr)
            return(SAres)
        }}else{fittingMat <- foreach(i = 1:rep) %do% {
            if(reporting){cat('.')}
            SAres <- fullSA(ocData = ocmat, envData=as.matrix(enmat),
                            runadamW = runadamW, sparse = sparse, we = we, 
                            totalit = totalit, lambda = lambda, intv = intv,
                            maxlearningrate = maxlr)
            return(SAres)
        }}

        if(reporting){cat(sprintf('SA: elapsed time %.2f sec\n\n', proc.time()[3]-s))}
        if(getall){
          return(fittingMat)
        }else{
          return(list(saEndpoint(fittingMat,ocmat,enmat=enmat),enmat))}
        ## ============================================== ##
	
    }
}


#' Extract endpoint parameters from \link{runSA} output
#'
#' Summarize the final parameter estimates across SA replicates by averaging
#' the last recorded block of parameters.
#'
#' @param fittingMat A list of SA trajectories returned by \code{runSA(getall = TRUE)}.
#' @param ocmat Numeric matrix (0/1). Rows are samples and columns are species.
#' @param enmat Optional environmental parameter matrix used in fitting.
#'
#' @return A numeric matrix of median endpoint parameters across serials. 
#' Rownames correspond to samples and Colnames correspond to species list
#'
#' @export
saEndpoint <- function(fittingMat,ocmat,enmat=NULL){
  
  num_cols <- ncol(ocmat)
  packed_results_end <- lapply(fittingMat, 
                               function(x) x[(nrow(x) - num_cols + 1):nrow(x),])
  params_all <- lapply(packed_results_end, function(x) x[, 1:(ncol(x) - 2)])
  #upd <- lapply(sa, function(x) x[, (ncol(x) - 1):ncol(x)]) 
  
  # median
  if (length(params_all) == 1) {
    params_mat <- params_all[[1]]
  } else {
    arr <- simplify2array(params_all)
    params_mat <- apply(arr, c(1, 2), median, na.rm = TRUE)
  }
  
  # mean
  #params_mat <- params_all[[1]]
  #if (length(params_all)>1){
  #  for (n in 2:length(params_all)) {
  #    params_mat <- params_mat + params_all[[n]]
  #  }
  #
  #  params_mat <- params_mat/length(params_all)
  #}
  rownames(params_mat) <- colnames(ocmat)
  
  if(is.null(enmat)){
    colnames(params_mat) <- c("h",paste('J',colnames(ocmat),sep='.'))
  }
  else{
    encols <- ncol(enmat)
    
    if( ncol(as.matrix(enmat))==1 ){
      gname <- paste('g', colnames(enmat), sep='.')
      ncols <- ncol(ocmat)+2
    }else{
      gname <- paste('g',1:ncol(enmat), sep='.')
      ncols <- ncol(ocmat)+ncol(enmat)+1
    }
    
    colnames(params_mat) <- c('h', gname, paste('J',colnames(ocmat),sep='.'))
  }
  
  return(params_mat)
}


#' Convert SA output to model parameter objects
#'
#' Convert the output of \link{runSA} / \link{saEndpoint} into parameter
#' objects in (extended) maximum pairwise entropy model:
#' \code{h}, \code{J}, and optionally \code{g} when environmental parameter
#' matrix is specified.
#'
#' @param sa A list in the form \code{list(param_matrix, enmat)} where
#'   \code{param_matrix} contains estimated parameters.
#' @param env Optional numeric vector of environmental values (column names correspond
#'    to environmental factors).Default is NULL.
#'
#' @return A list with elements \code{he} (h), \code{je} (J), \code{ge} (g or NULL),
#'   and \code{hge} (combined field).
#'
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
sa2params <- function(sa, env=NULL){
  sadim <- dim(sa[[1]])
  na <- rownames(sa[[1]])
  if(sadim[2] > sadim[1]+1){
      if(length(env)==0){env <- rep(0, sadim[2]-sadim[1]-1)} ## !!!
        he <- `names<-`(sa[[1]][, 1], na)
        je <- `dimnames<-`(sa[[1]][, ((sadim[2] - sadim[1]) + 1):sadim[2]],list(na, na))
        ge <- `dimnames<-`(as.matrix(sa[[1]][, 2:(sadim[2] - sadim[1])]),list(na, colnames(sa[[2]])))
        hge <- (he + ge %*% env)[,1]
        #hge <- t(t(envecs %*% t(ge)) + he)
  }else{
        he <- sa[[1]][, 1] %>% 'names<-'(na)
        je <- sa[[1]][, -1] %>% 'dimnames<-'(list(na,na))
        ge <- NULL
        hge <- he
  }
  return(list(he, je, ge, hge))}


#'testfittingSA: testrun of SA across given parameter sets using subset of ocmatrix
#'
#' Evaluate SA trajectories by validating the predicted occurrence matrix under
#' estimated parameters at each recorded time point.
#'
#' @param ocmat Numeric matrix (0/1). Observed community matrix.
#' @param sa_results List of SA parameter trajectories (e.g., from \code{runSA(getall=TRUE)}).
#' @param rep Integer. Number of replicates.
#' @param intv Integer. Recording interval used in fitting.
#' @param enmat Optional environmental matrix.
#'
#' @return A numeric matrix of validation scores (rows = replicates, columns = time points).
#'
#' @keywords internal
#' @noRd
testfittingSA <- function(ocmat,sa_results,rep,intv,enmat=NULL){
  nspecies = ncol(ocmat)
  num_timepoints <- nrow(sa_results[[1]]) / nspecies - 1
  timepoints <- sapply(0:num_timepoints, function(x) (x + 1) * intv)
  serials <- 1:rep
  res_serials <- matrix(0, ncol= (num_timepoints+1), nrow=rep,
                       dimnames=list(serials,timepoints))
  for(i in 1:length(sa_results)){
    rowidx <- 1
    timeidx <- 1
    SAparams <- sa_results[[i]]
    while(rowidx < nrow(SAparams)){
      
      params<-sa2params(list(SAparams[rowidx:(rowidx+nspecies-1),
                                                  1:(ncol(SAparams)-2)],enmat))
      res_serials[i,timeidx] <- validateSA(ocmat,params[[1]],params[[2]],
                                           g=params[[3]],enmat=enmat)
      rowidx <- rowidx + nspecies
      timeidx <- timeidx + 1
    }
  }
  return(res_serials)
}

#' Search best hyperparameters for SA fitting (lambda and weight decay) by cross-validation
#'
#' Repeating cross-validation over a grid of hyperparameters for
#' simulated annealing-based fitting, evaluates prediction/fitting errors on
#' held-out samples, and returns candidate best parameter/timepoint combinations.
#'
#' @param ocmat Numeric matrix (0/1). occurrence matrix of the communities.
#' @param enmat Optional environmental parameter matrix.
#' @param cvmode Cross-validation mode. One of `"10fold"`, `"5fold"`, or `"loo"`.
#'   `"loo"` performs leave-one-out cross-validation. Default is `"10fold"`.
#' @param nrepeat Number of repeated cross-validation runs for k-fold CV.
#'   Ignored when `cvmode = "loo"`. Default is `10`.
#' @param we Numeric vector. Candidate weight decay values.
#' @param totalit Integer. Total iterations for SA.
#' @param lmd Numeric vector. Candidate \code{lambda} values.
#' @param maxlrs Numeric vector. Maximum learning rate.
#' @param rep Integer. Number of replicates for \link{runSA}.
#' @param intv Integer. Recording interval.
#' @param threads Integer. Number of cores.
#' @param tol Numeric. Error tolerance for finding the best hyperparameters (default: 0.03)
#' @param runadamW Logical. Use AdamW.
#' @param sparse Logical. Use sparsity penalty.
#' @param reporting Logical. Print progress.
#' @param fastfitting Logical. If \code{TRUE}, returns early iterations within the range
#' specified by the parameter \code{tol} (default: 0.03) of best error.
#' @importFrom utils head
#'
#' @return A list with:
#' \describe{
#'   \item{best_candidates}{A named list of best hyperparameter settings and scores.}
#'   \item{allresults}{All validation results for each grid point.}
#' }
#'
#' @export

Findbp <- function(ocmat,enmat=NULL,cvmode=c("10fold","5fold","loo"),nrepeat=5,
                   we=c(0.001),maxlrs = c(0.005),
                   totalit=4000,lmd=c(0.0005,0.001,0.0025,0.005,0.0075),
                   rep=8,intv=100,threads=1,tol=0.03,
                   runadamW=TRUE,sparse=TRUE,fastfitting=TRUE,reporting=TRUE){
  min_points <- list()
  n_samples <- nrow(ocmat)
  if(n_samples < 2){
    stop("At least two samples are required.")
  }
  cvmode <- match.arg(cvmode)
  nrepeat <- as.integer(nrepeat)
  if(nrepeat < 1){
    stop("nrepeat must be >= 1.")
  }
  if(!runadamW){we <- c(0)}
  if(!sparse){lmd <- c(0)}
  ## ------------------------------------------------------------
  ## Make repeated cross-validation folds
  ## ------------------------------------------------------------
  split_list <- list()
  if(cvmode == "loo"){
    ## Leave-one-out is deterministic; nrepeat is ignored.
    for(i in seq_len(n_samples)){
      split_list[[length(split_list) + 1]] <- list(
        repeat_id = 1,
        fold_id = i,
        label = paste0("loo_", i),
        train = setdiff(seq_len(n_samples), i),
        test  = i
      )
    }
    split_prefix <- "loo"
  }else{
    k <- ifelse(cvmode == "10fold", 10, 5)
    if(n_samples < k){
      stop("Number of samples must be >= number of folds.")
    }
    for(r in seq_len(nrepeat)){
      ## Randomly assign samples to approximately equal-sized folds
      fold_id <- sample(rep(seq_len(k), length.out=n_samples))
      for(f in seq_len(k)){
        test_idx <- which(fold_id == f)
        train_idx <- setdiff(seq_len(n_samples), test_idx)
        split_list[[length(split_list) + 1]] <- list(
          repeat_id = r,
          fold_id = f,
          label = paste0("rep", r, "_fold", f),
          train = train_idx,
          test  = test_idx
        )
      }
    }
    split_prefix <- paste0(k, "fold")
  }
  n_splits <- length(split_list)
  ## ------------------------------------------------------------
  ## Make task grid:
  ## one task = one hyperparameter setting x one repeat/fold split
  ## ------------------------------------------------------------
  task_grid <- expand.grid(
    lambda = lmd,
    we = we,
    maxlr = maxlrs,
    split_id = seq_len(n_splits),
    stringsAsFactors = FALSE
  )
  n_tasks <- nrow(task_grid)
  ## ------------------------------------------------------------
  ## Required packages
  ## ------------------------------------------------------------
  #if(!requireNamespace("foreach", quietly=TRUE)){
  #  stop("Package 'foreach' is required.")
  #}
  #if(!requireNamespace("magrittr", quietly=TRUE)){
  #  stop("Package 'magrittr' is required.")
  #}
  #suppressPackageStartupMessages({
  #  library(foreach)
  #  library(magrittr)
  #})
  ## ------------------------------------------------------------
  ## Parallel setup
  ## ------------------------------------------------------------
  use_parallel <- 1
  if(use_parallel){
    if(!requireNamespace("doParallel", quietly=TRUE)){
      stop("Package 'doParallel' is required when threads > 1.")
    }
    n_workers <- min(threads, n_tasks)
    cluster <- parallel::makeCluster(n_workers)
    parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
    #parallel::clusterEvalQ(cluster, {
    #  suppressPackageStartupMessages({
    #    library(foreach)
    #    library(magrittr)
    #    library(ELAplus)
    #  })
    #  NULL
    #})
    doParallel::registerDoParallel(cluster)
    on.exit(parallel::stopCluster(cluster), add=TRUE)
  }
  ## ------------------------------------------------------------
  ## Run all tasks
  ## ------------------------------------------------------------
  run_one_task <- function(i){
    l <- task_grid$lambda[i]
    w <- task_grid$we[i]
    maxlr <- task_grid$maxlr[i]
    split_id <- task_grid$split_id[i]
    split <- split_list[[split_id]]
    train_idx <- split$train
    test_idx  <- split$test
    octraining <- ocmat[train_idx,,drop=FALSE]
    octest     <- ocmat[test_idx,,drop=FALSE]
    if(!is.null(enmat)){
      envtraining <- enmat[train_idx,,drop=FALSE]
      envtest     <- enmat[test_idx,,drop=FALSE]
    }else{
      envtraining <- NULL
      envtest     <- NULL
    }
    sa_results <- runSA(octraining, enmat=envtraining,
                        rep=rep, threads=1, we=w,maxlr=maxlr,
                        totalit=totalit, lambda=l, intv=intv,
                        runadamW=runadamW, sparse=sparse,
                        getall=TRUE, reporting=FALSE)
    current_results <- testfittingSA(octest, sa_results,
                                     rep=rep, intv=intv,
                                     enmat=envtest)
    ## Average across SA replicates within this split
    error_vec <- colMeans(current_results, na.rm=TRUE)
    list(
      lambda = l,
      we = w,
      maxlr = maxlr,
      split_id = split_id,
      split_label = split$label,
      repeat_id = split$repeat_id,
      fold_id = split$fold_id,
      error = error_vec
    )
  }
  if(reporting){cat('Start hyper-parameter search:\n')}
  if(reporting){cat(sprintf("There are %d tasks\n", n_tasks))}
  
  s <- proc.time()[3]
  
  task_results <- if(use_parallel){
    foreach(
      i = seq_len(n_tasks),
      .packages = c("foreach", "magrittr", "ELAplus"),
      .export = c("testfittingSA", "validateSA", "sa2params")
    ) %dopar% {
      run_one_task(i)
    }
  }else{
    foreach(
      i = seq_len(n_tasks),
      .packages = c("foreach", "magrittr", "ELAplus"),
      .export = c("testfittingSA", "validateSA", "sa2params")
    ) %do% {
      run_one_task(i)
    }
  }
  
  ## ------------------------------------------------------------
  ## Reconstruct allresults in the same format:
  ## allresults[[lambda_we]] = matrix of CV errors
  ## rows = repeat/fold splits, columns = timepoints
  ## ------------------------------------------------------------
  allresults <- list()
  min_points <- list()
  combo_names <- unique(paste(task_grid$lambda, task_grid$we, task_grid$maxlr, sep="_"))
  for(combo_name in combo_names){
    idx <- which(paste(task_grid$lambda, task_grid$we, task_grid$maxlr, sep="_") == combo_name)
    combo_results <- task_results[idx]
    current_results <- do.call(
      rbind,
      lapply(combo_results, function(x) x$error)
    )
    rownames(current_results) <- sapply(combo_results, function(x) x$split_label)
    allresults[[combo_name]] <- current_results
    mean_results <- apply(current_results, 2, mean, na.rm=TRUE)
    combo_parts <- strsplit(combo_name, "_", fixed=TRUE)[[1]]
    l <- combo_parts[1]
    w <- combo_parts[2]
    maxlr <- combo_parts[3]
    for (m in names(mean_results)){
      min_points[[paste(l, w, maxlr, as.integer(m), sep = "_")]] <- as.numeric(mean_results[m])
    }
  }
  if(reporting){cat(sprintf("Finished %d tasks\n", n_tasks))}
  if(reporting){cat(sprintf('SA: elapsed time %.2f sec\n\n', proc.time()[3]-s))}
  ## ------------------------------------------------------------
  ## Select best parameter/timepoint combinations
  ## ------------------------------------------------------------
  if(fastfitting){
    min_err <- min(unlist(min_points))
    min_points <- min_points[min_points < min_err*(1+tol)]
    sorted_min_points <- min_points[order(sapply(names(min_points),
                                                 function(x) {as.numeric(unlist(strsplit(x,split="_"))[[4]])}))]
  }else{
    sorted_min_points <- min_points[order(sapply(min_points, function(x) x[1]))]
  }
  return(list(head(sorted_min_points,min(length(sorted_min_points),5)),allresults))
}

#'plotSAtest: plot the results of SA fitting (Findbp output)
#' 
#' Plot mean validation scores over time for the results returned by
#' \link{Findbp}, including \code{95\%} confidence intervals.
#'
#' @param sa_results Output list from \link{Findbp}.
#' @param ylim Numeric length-2 vector (ymin, ymax). Default: \code{c("auto","auto")}.
#'
#' @return A \pkg{ggplot2} object.
#' @importFrom tidyr gather
#' @importFrom ggplot2 geom_line geom_ribbon scale_y_continuous
#' @details
#' This function uses \pkg{ggplot2} and tidyverse-style data manipulation.
#' Ensure required packages are installed.
#'@export
plotSAtest <- function(sa_results,ylim=c("auto","auto")){
  
  if(is.null(names(sa_results))){
    names(sa_results) <- paste0("set_", seq_along(sa_results))
  }
  
  summary_df <- data.frame()
  
  for(i in seq_along(sa_results)){
    tmp <- as.data.frame(sa_results[[i]])
    tmp$rowname <- rownames(tmp)
    rownames(tmp) <- NULL
    res_serials_melt <- tidyr::pivot_longer(
      tmp,
      cols = setdiff(names(tmp), "rowname"),
      names_to = "timepoint",
      values_to = "value"
    )
    
    res_serials_melt$timepoint <- as.numeric(res_serials_melt$timepoint)
    results_i <- res_serials_melt
    results_i$id <- names(sa_results)[i]
    summary_df <- rbind(summary_df,results_i)
  }
  
  summary_mean <- summary_df %>%
    dplyr::group_by(.data$id,.data$timepoint) %>%
    dplyr::summarise(n=sum(!is.na(.data$value)),
                     mean=mean(.data$value, na.rm=TRUE),
                     sd=stats::sd(.data$value, na.rm=TRUE),
                     .groups="drop") %>%
    dplyr::mutate(sem=ifelse(.data$n > 1, .data$sd / sqrt(.data$n), 0),
                  tcrit=ifelse(.data$n > 1, stats::qt(0.975, .data$n - 1), 0),
                  CI_lower=.data$mean - .data$tcrit * .data$sem,
                  CI_upper=.data$mean + .data$tcrit * .data$sem)
  
  if(ylim[[1]]=="auto"){
    ymin <- min(summary_mean$CI_lower, na.rm=TRUE)
  }else{
    ymin <- as.numeric(ylim[[1]])
  }
  
  if(ylim[[2]]=="auto"){
    ymax <- max(summary_mean$CI_upper, na.rm=TRUE)
  }else{
    ymax <- as.numeric(ylim[[2]])
  }
  
  ggplot(summary_mean, aes(x=.data$timepoint, y=.data$mean, color=.data$id)) +
    geom_line(aes(color=.data$id)) +
    geom_ribbon(aes(ymin=.data$CI_lower, ymax=.data$CI_upper, fill=.data$id),
                color="grey70", alpha=0.4) +
    scale_y_continuous(limits=c(ymin,ymax))
}


#'validateSA: to check the predictability of SA results by computing ocmatrix using
#'the estimated parameter set.
#'
#' @param ocmat Numeric matrix (0/1). Occurrence matrix.
#' @param h Numeric vector. Implicit environmental effects on species occurrence.
#' @param j Numeric matrix. Species-interaction matrix.
#' @param g Optional numeric matrix. Explicit environmental effects on species occurrence.
#' @param enmat Optional environmental parameters matrix.
#'
#' @return A numeric score (negative mean log of prediction error).
#'
#' @noRd
validateSA <- function(ocmat,h,j,g=NULL,enmat=NULL){
  if(is.null(enmat)){
    pp <- abs(1 - ocmat - 1 / (1 + exp(t(-h - t(ocmat %*% j)))))
  }
  else{
    pp <- abs(1 - ocmat - 1 / (1 + exp(-t(t(enmat %*% t(g)) + h) - ocmat %*% j)))
    #pp <- abs(1 - ocmat - 1 / (1 + exp(t(t(-hge - (ocmat %*% J))))))
  }
  return( -mean(log(pp[pp > 0])))
}

#' Parameter fitting with bootstrap sampling of datasets
#'
#' Perform \link{runSA} algorithm and estimate parameters using bootstrap resampled
#' data set (\code{ocmat} and \code{enmat}). 
#' The basic configuration is same as \link{runSA}.
#'
#' @param ocmat Numeric matrix (0/1). Rows are samples and columns are species.
#' @param enmat Optional numeric matrix. Environmental variables for each sample.
#'   Should be normalized/standardized beforehand.
#' @param rep Integer. Number of parallel runs (replicates) for optimization.
#' @param threads Integer. Number of cores used when running in parallel.
#' @param we Numeric. Weight decay parameter for AdamW.
#' @param totalit Integer. Total number of optimization iterations.
#' @param lambda Numeric. Sparsity penalty parameter (used when \code{sparse = TRUE}).
#' @param runadamW Logical. If \code{TRUE}, use AdamW-style optimization.
#' @param maxlr Numeric. Maximum learning rate.
#' @param sparse Logical. If \code{TRUE}, apply sparsity constraint.
#' @param bootstrap.it number of bootstrap trials (default: 1000).
#' @param reporting Logical. If \code{TRUE}, print progress messages.
#'
#' @return a list of SA parameters estimated with bootstrap resampling (length \code{bootstrap.it}).
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @importFrom Rcpp sourceCpp
#' @import foreach
#' @export
bootstrap_SA <- function(ocmat, enmat=NULL, bootstrap.it=2000, rep=128, threads=1, we=0.001, 
                         totalit=1000, lambda=0.01, runadamW=TRUE,maxlr = 0.005, sparse=TRUE, reporting=TRUE){
  if(reporting){cat('Start parameter fitting:\n')}
  s <- proc.time()[3]
  #if(!is.matrix(ocmat)){ocmat <- as.matrix(ocmat)}
  if(!sparse){lambda <- 0}
  
  #if(!runadamW){we <- 0}
  
  bootstrap_res <- vector("list", bootstrap.it)
  n_samples <- nrow(ocmat)
  
  for (i in 1:bootstrap.it){
    # bootstrap sampling
    if(reporting){
      if(stats::runif(1) > max((bootstrap.it - 100)/bootstrap.it, 0))
      cat('.')
      }
    randidxs <- sample(n_samples, size = n_samples, replace = TRUE)
    ocrand <- ocmat[randidxs, , drop = FALSE]
    
    if(!is.null(enmat)){
      envrand <- enmat[randidxs, , drop = FALSE]
    }
    else{envrand=NULL}
    
    sa_results <- runSA(ocrand, enmat=envrand, rep=rep, threads=threads, we=we, 
                        totalit=totalit, lambda=lambda,intv=100, runadamW=runadamW, 
                        sparse=sparse,maxlr=maxlr,getall=FALSE, reporting=FALSE)
    
    bootstrap_res[[i]] <- sa_results[[1]]
  }
  if(reporting){cat(sprintf('\nSA: elapsed time %.2f sec\n\n', proc.time()[3]-s))}
  return(bootstrap_res)
}

#' Check variability in the stochastic approximation algorithm. 
#'
#' Perform bootstrap resampling of the estimated paramters using \link{runSA} with \code{getall=TRUE}.
#' This bootstrap sampling reflects the intrinsic variability of the stochastic approximation algorithm 
#' and does not correspond to standard confidence intervals based on resampled data.
#' Use \link{bootstrap_SA} to assess finite-sample estimation error. 
#'
#' @param fittingMat output of runSA function (need all run results in serials) 
#' @param ocmat Numeric matrix (0/1). Rows are samples and columns are species.
#' @param enmat Optional numeric matrix. Environmental variables for each sample.
#'   Should be normalized/standardized beforehand.
#' @param resamplesize total resampling size (default: length(fittingMat)), necessary when \code{type="quantile"}.
#' @param iterations number of bootstrap iterations (default: 1000).
#'
#' @return a list of SA parameters estimated with bootstrap resampling (length \code{iterations}).
#' @noRd
compute_SA_variance <- function(fittingMat,ocmat,enmat=NULL,resamplesize=NULL,iterations = 1000){
  if(is.null(resamplesize)){resamplesize<-length(fittingMat)}
  rand_sample_idx <- lapply(1:iterations, function(i) {sample(1:length(fittingMat),size=resamplesize,replace=TRUE)})
  rand_sa_params <- vector("list", iterations)
  for (i in 1:iterations){
    rand_serial <- fittingMat[rand_sample_idx[[i]]]
    rand_sa_params[[i]] <- saEndpoint(rand_serial,ocmat,enmat=enmat)
  }
  return(rand_sa_params)
}

#' applymap
#' @description Apply a function to each element by row and return a matrix-like result.
#' @param df A matrix/data.frame to transform.
#' @param fun A function applied to each element.
#' @return A matrix (as returned by t(apply(...))) of transformed values.
#' @noRd
applymap <- function(df, fun){
  return(t(apply(df, 1, function(row){sapply(row, fun)})))
}

#' applyvec
#' @description Apply a function to each column (or vector) and preserve row/column names.
#' @param df A data.frame (or matrix coercible) to transform.
#' @param fun A function applied to each column (via lapply).
#' @return A data.frame with transformed columns and preserved names.
#' @noRd
applyvec <- function(df, fun){
  if(length(rownames(df)) > 1){
    output <- t(data.frame(lapply(t(df), fun))) %>% 'rownames<-'(rownames(df))
  }else(
    output <- data.frame(lapply(df, fun)) %>% 'colnames<-'(colnames(df))
  )
  return(output)
}

#' T
#' @description Transpose a data.frame and return it as a data.frame.
#' @param df A matrix/data.frame to transpose.
#' @return A transposed data.frame.
#' @noRd
T <- function(df){return(as.data.frame(t(df)))}

#' Formatting
#' @description Generate occurrence(binary)/abundance/factor matrices and labels from input abundance/metadata tables.
#' @param baseabtable Abundance table (samples x species) as data.frame/matrix.
#' @param basemetadata Optional environmental metadata table (samples x factors); NULL/empty for none.
#' @param normalize Integer 0 or 1. If 1, normalize within each sample to sum to 1.
#' @param grouping Integer 0 or 1. If 1, group similar occurrence patterns.
#' @param grouping_th Numeric threshold for grouping (relative Hamming distance).
#' @param reporting Logical. If TRUE, print processing summary.
#' @param parameters Numeric vector/list of thresholds (following 3 parameters).
#'   1. "0/1 threshold"
#'   2. "min occurrence threshold"
#'   3. "max occurrence threshold"
#'   (e.g., c(0.05, 0.05, 0.95) : species abundance > 0.05 is regarded as present,
#'   and species occurring in more than 0.05 and less than 0.95 in the all samples are kept.)
#' @return A list: (ocmatrix, abmatrix, fctmatrix, samplelabel, specieslabel, factorlabel).
#' @useDynLib ELAplus, .registration=TRUE
#' @export
#' @importFrom foreach foreach %do%
#' @importFrom stringdist stringdist

Formatting <- function(baseabtable, basemetadata=NULL, normalize=1, 
                       parameters=c(0.05,0.05,0.95), grouping=0, grouping_th=0, reporting=TRUE){
  ath <- parameters[1]     # Relative frequency threshold; if less than this value, set to 0
  minocth <- parameters[2] # Lower limit of mean occurence; species with mean occurence lower than this value are excluded.
  maxocth <- parameters[3] # Upper limit of mean occurence; species with mean occurence higher than this value are excluded.
  
  if(length(basemetadata) != 0){
    # with basemetadata
    # Get label information
    specieslabel = colnames(baseabtable)
    absamplelabel = rownames(baseabtable)
    mdsamplelabel = rownames(basemetadata)
    sharedsamplelabel = sort(intersect(absamplelabel, mdsamplelabel))

    # Generating occurence matrix
    patab = match(sharedsamplelabel, absamplelabel)
    patmd = match(sharedsamplelabel, mdsamplelabel)
    abnum = baseabtable[patab,]
    if (normalize == 1){ # Nomalization within samples (optional)
      abnum <- t(apply(abnum, 1, function(x){x / (sum(x) + 10^-16)}))
    }
    aboc = applymap(abnum, function(x){if(x >= ath){1}else{0}}) # 0/1 transformation with ath
    aboc_mean = apply(aboc, 2, mean)
    occrit = names(aboc_mean[(aboc_mean > minocth) & (aboc_mean < maxocth)]) # Calculation of mean occurence//determination of the target to be removed

    # Object generation
    ocmatrix = aboc[, occrit] # Occurence matrix
    abmatrix = abnum[, occrit] # Abundance matrix
    fctmatrix = as.matrix(as.data.frame(basemetadata[patmd, ])) # Environmental factors
    rownames(fctmatrix) <- rownames(ocmatrix) 
    colnames(fctmatrix) <- colnames(basemetadata)
    samplelabel = sharedsamplelabel # Sample labels
    specieslabel = occrit # Species labels
    factorlabel = colnames(basemetadata) # Factor labels
  }else{
    # without basemetadata
    # Get label information
    specieslabel = colnames(baseabtable)
    absamplelabel = rownames(baseabtable)
    sharedsamplelabel = sort(absamplelabel)

    # Generating occurence matrix
    patab = match(sharedsamplelabel, absamplelabel)
    abnum = baseabtable[patab,]
    if (normalize == 1){ # Nomalization within samples (optional)
      abnum <- t(apply(abnum, 1, function(x){x / (sum(x) + 10^-16)}))
    }
    aboc = applymap(abnum, function(x){if(x >= ath){1}else{0}}) # 0/1 transformation with ath
    aboc_mean = apply(aboc, 2, mean)
    occrit = names(aboc_mean[(aboc_mean > minocth) & (aboc_mean < maxocth)]) # Calculation of mean occurence//determination of the target to be removed

    # Object generation
    ocmatrix = aboc[, occrit] # Occurence matrix
    abmatrix = abnum[, occrit] # Abundance matrix
    fctmatrix = NULL # Environmental factors
    samplelabel = sharedsamplelabel # Sample labels
    specieslabel = occrit # Species labels
    factorlabel = NULL # Factor labels
  }

  # Grouping
  if(grouping == 1){
    ocvstr <- apply(ocmatrix, 2, function(y){return(paste(y, collapse = ""))})
    oo <- order(ocvstr)
    ov <- ocvstr[oo]
    j <- 0
    cc <- 0
    binlist <- list()
    aaa <- foreach (i=ov) %do% {
        cc <- cc + 1
        if(stringdist(i, j, method='hamming')/nchar(i) <= grouping_th){a <- rev(binlist)[[1]]
                aa <- append(a, cc)
                binlist <- append(binlist[-length(binlist)], list(aa))
                }else{binlist <- append(binlist, list(cc))}
        j=i
        }
    ltx <- length(specieslabel)
    specieslabel <- foreach (i=binlist, .combine='c') %do% {paste(specieslabel[oo][i], collapse=", ")}
    ocmatrix <- foreach(i=binlist, .combine='cbind') %do% {ocmatrix[,oo][,i[1]]}
    abmatrix <- foreach(i=binlist, .combine='cbind') %do% {if(length(i)>1){apply(abmatrix[,oo][,i],1,sum)}else{abmatrix[,oo][,i]}}
    colnames(ocmatrix) <- specieslabel
    colnames(abmatrix) <- specieslabel
  }

  if(reporting){
    cat('Processed', as.character(length(samplelabel)), 'samples.\n')
    cat('Relative abundance threshold =', as.character(ath), '\n')
    cat('Occurrence threshold (lower) =', as.character(minocth), '\n')
    cat('Occurrence threshold (upper) =', as.character(maxocth), '\n')
    if(grouping == 1){cat(as.character(ltx-length(specieslabel)), ' groups were found.\n')}
    cat('Selected ', as.character(length(specieslabel)), ' out of ',
                as.character(ncol(baseabtable)), 'species.\n')
  }
  return(list(ocmatrix, abmatrix, fctmatrix, samplelabel, specieslabel, factorlabel))}


#' str2i
#' @description Convert a binary string (e.g., "0101") to an integer.
#' @param x A character string consisting of 0/1 digits.
#' @return A numeric scalar giving the decoded integer value.
#' @noRd
str2i <- function(x) {
    y <- as.numeric(strsplit(x, "")[[1]])
    sum(y * 2^rev((seq_along(y)-1)))
}

#'bin2id: map a binary vector to a 64base characters
#' @description Encode a binary vector into a base-64 id string index.
#' @param bin A binary vector (0/1, community composition) to encode.
#' @return A character scalar representing the encoded id.
#' @export
#' @importFrom foreach foreach
#' @export
bin2id <- function(bin){
  digs <- "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_/"
  bi <- unlist(bin)
  modi <- length(bi)%%6
  if(modi != 0){b0 <- rep(0, 6-modi)}else{b0<-c()}
  bii <- c(b0, bi)
  lb <- length(bii)
  aa <- foreach(i=seq(1, length(bii), 6), .combine="c") %do% {a <- rev(bii)[i:(i+5)]
                                s <- str2i(paste(as.character(rev(a)), collapse=""))
                                substr(digs, s+1, s+1)}
  return(paste(rev(aa), collapse=""))}

#'id2bin: map a 64base characters to a binary vector
#' @description Decode a base-64 id string index to a binary vector of length s.
#' @param id A character scalar encoded by \link{bin2id}.
#' @param s Integer number of bits to return.
#' @return An integer vector of 0/1 bits of length s.
#' @export
#' @importFrom foreach foreach
#' @export
id2bin <-  function(id, s){
    digs <- "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_/"
    aa <- foreach (i=seq(1:nchar(id)), .combine="rbind") %do% {
        a <- substr(id,i,i)
        ia <- (unlist(gregexpr(a, digs))[1]) - 1
        CIntegerDigits(ia, 6)}
    uaa <- unlist(aa)
    aaa <- as.vector(uaa[(length(uaa)-s+1):length(uaa)])
    return(aaa)}

#' CIntegerDigits
#' @description Map an integer or encoded ID to a binary vector of length n (optionally with row names).
#' @param ssid Numeric state id or encoded id (character) to convert.
#' @param n Integer number of bits.
#' @param index Optional row names (default 1:n).
#' @return A data.frame of bits (0/1) with row names \code{index}.
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
CIntegerDigits <- function(ssid, n, index=NULL){
  if(is.null(index)){index <- 1:n}
  if(is.numeric(ssid)){
    return(data.frame(dec2bin(ssid, n), row.names=index))
  }
  else{
    return(data.frame(id2bin(ssid, n), row.names=index))
  }
}

#' dec2bin
#' @description Convert a non-negative integer to a binary vector of fixed length.
#' @param num Integer to convert.
#' @param digit Integer length of the output binary vector.
#' @return An integer vector of 0/1 bits (or NULL if both num<=0 and digit<=0).
#' @noRd
dec2bin <- function(num, digit=0){
  if(num <= 0 && digit <= 0){
    return(NULL)
  }else{
    return(append(Recall(num%/%2,digit-1), num%%2))
  }
}

#'PCplot:Scatter plot of community compositions colored by basin label
#'
#' @description Performs PCA on the community composition matrix (\code{ocmat}) and plots the first
#' two PCs. Points are colored by the stable state (basin) label estimated
#' from \link{Bi}. If \code{pruned = TRUE}, colors correspond to the
#' pruned/merged basin labels in \code{ssrep} (\code{stablestates2}), otherwise to the
#' original labels (\code{stablestates}). When \code{export = TRUE}, the plotting data
#' frame (stable-state labels + PCs + merged labels) is returned.
#'
#' @param ocmat Data frame (or matrix) of community compositions (rows = samples,
#'   columns = species/features).
#' @param sa Parameter set passed to \link{sa2params}.
#' @param env Optional environment object passed to \link{sa2params}.
#' @param ssrep Data frame-like object providing stable state label mapping with
#'   two columns: \code{stablestates} and \code{stablestates2}.
#' @param pruned Logical; if TRUE, use \code{stablestates2} for coloring.
#' @param export Logical; if TRUE, return the data used for plotting.
#' @importFrom dplyr select left_join bind_rows
#' @return NULL invisibly (or a data frame when \code{export = TRUE}).
#' @useDynLib ELAplus, .registration=TRUE
#' @export
PCplot <- function(ocmat, sa=sa, env=NULL, ssrep=NULL, pruned=TRUE, export=FALSE){
  ######################
  sa2p <- sa2params(sa, env)
  hestp <- sa2p[[1]]
  jestp <- sa2p[[2]]
  gestp <- sa2p[[3]]
  hgestp <- sa2p[[4]] 
  ####################
    if(length(ssrep)!=0){
    ssrep_ <- ssrep %>% as.data.frame() %>% 'colnames<-'(
    c('stablestates', 'stablestates2'))
    stablestates <- ssrep_$stablestates
    stablestates2 <- ssrep_$stablestates2
  PCA_ocmatrix <- as.data.frame(stats::prcomp(x=ocmat, scale=TRUE)[5]) %>%
    'colnames<-'(lapply(1: length(names(hgestp)), function(x){
      gsub(' ', '', paste('PC', as.character(x)), fixed=TRUE)}))
  states <- as.data.frame(t(as.data.frame(
    do.call(
      rbind,
      apply(ocmat, 1, function(x){
        t(Bi(x,
             hgestp, jestp))}))[, 1]))) %>% 'rownames<-'(
               rownames(ocmat)) %>% 'colnames<-'('stablestates')
  state_PC <- cbind(states, PCA_ocmatrix) %>% left_join(ssrep_, by='stablestates')
  if(pruned==FALSE){
    g <- ggplot(state_PC, aes(x=.data$PC1, y=.data$PC2, color=as.factor(stablestates)))}
  else{
    g <- ggplot(state_PC, aes(x=.data$PC1, y=.data$PC2, color=as.factor(stablestates2)))
  }
  g <- g + geom_point() + theme(aspect.ratio = 1)
  plot(g)
  if(export){return(state_PC)}else{return(NULL)}
    }else{
        cat("Please set ssrep")
    }}


#'temp.mat: supporting function for calc_ccc
#'
#' @description Builds an adjacency (0/1) matrix from an edge list \code{rips} with columns \code{V1}
#' and \code{V2}. The matrix size is determined by \code{max(rips)}, and entries
#' \code{matrix.p[V1, V2]} are set to 1.
#'
#' @param rips Edge list as a data frame with at least columns \code{V1} and \code{V2}.
#' @return A square adjacency matrix with dimnames \code{1:n}.
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
temp.mat <- function(rips){
  n <- max(rips)
  matrix.p <- matrix(0, nrow = n, ncol = n)
  for (i in 1:nrow(rips)) {
    matrix.p[rips$V1[i], rips$V2[i]] <- 1
  }
  dimnames(matrix.p) <- list(1:n, 1:n)
  return(matrix.p)}


#' calc_ccc: supporting function for DisconnectivityGraph
#'
#' @description Computes connected components (clusters) under an energy threshold \code{ww}.
#' Stable states with energy <= \code{ww} are included, and tipping connections with
#' energy <= \code{ww} are added as edges. The resulting graph is
#' decomposed into connected components, filtered to keep components of size >= 2
#' and (optionally) singleton components that correspond to a stable state.
#'
#' @param ww Numeric energy threshold.
#' @param stablestates List (or vector) of stable states.
#' @param ssenergies Energies for stable states.
#' @param tippingenergies List of energies in tipping points.
#' 
#' @importFrom dplyr group_by group_map filter
#' @importFrom igraph components graph_from_adjacency_matrix
#' @importFrom tidyr pivot_longer
#' @return A list of connected components (each component is a vector of node indices).
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
calc_ccc <- function(ww, stablestates, ssenergies, tippingenergies){
  sta <- data.frame(ssenergies) %>% subset(ssenergies <= ww) %>% rownames() %>%
    as.integer
  coordinates <- as.data.frame(gtools::permutations(n=length(
    stablestates), r=2, v=1:length(stablestates), repeats.allowed = TRUE))
  tip <- cbind(1: length(tippingenergies), as.data.frame(tippingenergies)) %>%
    #pivot_longer(cols=colnames(as.data.frame(tippingenergies))) %>% cbind(
    #  coordinates) %>% subset(value <= ww) %>% select(c('V1', 'V2'))
    pivot_longer(cols=colnames(as.data.frame(tippingenergies))) %>%
    dplyr::bind_cols(coordinates) %>%
    dplyr::filter(.data$value <= ww) %>%
    dplyr::select(c('V1', 'V2'))
  rips <- rbind(as.data.frame(cbind(sta, sta)) %>% 'colnames<-'(c('V1', 'V2')), tip)
  tempmat <- temp.mat(rips)
  ccc_pre1 <- as.vector(components(graph_from_adjacency_matrix(
    adjmatrix=as.matrix(tempmat) + t(tempmat), mode='undirected', diag=F))$membership)
  ccc_pre2 <- T(as.data.frame(rbind(ccc_pre1, 1: length(ccc_pre1)))) %>% 
    group_by(ccc_pre1) %>% group_map(as.list) %>% lapply(function(x){x$V2})
  ccc_pre2 <- ccc_pre2 %>% 'names<-'(1: length(ccc_pre2))
  ccc_pre2_length <- T(as.data.frame(lapply(
    ccc_pre2, function(x){length(x)}))) %>% 'rownames<-'(1: length(ccc_pre2))
  #ccc_final1 <- ccc_pre2[ccc_pre2_length %>% subset(V1 >= 2) %>% rownames]
  #ccc_final2_cond1 <- ccc_pre2_length %>% subset(V1 == 1) %>% rownames
  ccc_final1 <- ccc_pre2[rownames(ccc_pre2_length)[ccc_pre2_length[, "V1"] >= 2]]
  ccc_final2_cond1 <- rownames(ccc_pre2_length)[ccc_pre2_length[, "V1"] == 1]
  #print(ccc_final2_cond1)
  ccc_final2_cond2 <- cbind(ccc_pre2[ccc_final2_cond1], lapply(
    ccc_pre2[ccc_final2_cond1], function(x){length(intersect(x, sta))})) %>% 
    'colnames<-'(c('node', 'len')) %>% applymap(function(x){x[1]}) %>%
    as.data.frame
  ####################################################
  if(length(ccc_final2_cond2)>0){
      ccc_final2 <- ccc_pre2[rownames(ccc_final2_cond2['len'] == 1)]
      }else{ccc_final2 <- NULL}
  ####################################################
  ccc <- c(ccc_final1, ccc_final2)
  return(ccc)
}

#' Assort: supporting function for DisconnectivityGraph
#'
#' @description Sorts (assorts) a list of daughter sets according to their first matching
#' parent set, then by descending set size, then by the first element.
#'
#' @param parent List of parent sets.
#' @param daughter List of daughter sets.
#' @return A one-element list containing the reordered daughter sets.
#' @noRd
Assort <- function(parent, daughter){
  parent_ <- parent %>% unlist(recursive=FALSE)
  daughter_ <- daughter %>% unlist(recursive=FALSE)
  orderkey <- do.call(
    rbind, 
    daughter_  %>%
      lapply(function(x){list(
        match(TRUE, lapply(parent_, function(y){all(x %in% y)})), -length(x), x[1]
        )})) %>% as.data.frame %>% applymap(function(x){x[1]}) %>% as.data.frame %>%
    'colnames<-'(c('setloc', 'length', 'firstN')) %>% 
    cbind(index=1: length(daughter_))
  return(list(daughter_[order(orderkey$setloc, orderkey$length, orderkey$firstN)]))
}

#' calc_cee: supporting function for DisconnectivityGraph
#'
#' @description Iteratively constructs an ordered list of connected component sets
#' from hierarchical edge/component information, using \code{Assort} to keep
#' a consistent parent-daughter ordering.
#'
#' @param hee List of component sets (hierarchical sequence).
#' @return A list of ordered component sets.
#' @noRd
calc_cee <- function(hee){
  cee <- list()
  for(daughter in hee){
    daughter_ <- list(daughter)
    if(length(cee) == 0){
      cee <- append(cee, daughter_)
    }else{
      parent <- cee[length(cee)]
      cee <- append(cee, Assort(parent, daughter_))
    }
  }
  return(cee)
}

#' calc_xpos1: supporting function for DisconnectivityGraph
#'
#' @description Updates x-positions for components when moving from a previous level
#' to a refined partition. If the partition size is
#' unchanged, positions are kept; otherwise, new positions are computed as the
#' average of member positions.
#'
#' @param xpos_prev Previous x-position list.
#' @param rcee_ Current level component sets.
#' @importFrom tidyr all_of
#' @importFrom data.table transpose
#' @return A list containing component ids and their updated x-positions.
#' @noRd
calc_xpos1 <- function(xpos_prev, rcee_){
  rcee__ <- rcee_
  xpos_prev_ <- unlist(xpos_prev, recursive=FALSE)
  if(length(rcee__) == length(unlist(xpos_prev_[1], recursive=FALSE))){
    output <- xpos_prev
  }else{
    xx <- xpos_prev_
    yy <- rcee__
    # pp_pre <- lapply(yy, "length<-", max(lengths(yy))) %>% as.data.frame %>% T %>%
    #  'colnames<-'(1: max(unlist(lapply(yy, length)))) %>%
    #  cbind(index=1: length(yy)) %>%
    #  pivot_longer(cols=-index, names_to='columns', values_to='yy_') %>%
    #  stats::na.omit() %>% applymap(as.integer) %>% as.data.frame() %>%
    #  column_to_rownames(var='yy_')
    tmp <- lapply(yy, "length<-", max(lengths(yy))) %>% as.data.frame %>% T %>%
      'colnames<-'(1: max(unlist(lapply(yy, length)))) %>%
      cbind(index=1: length(yy))
    pp_pre <- pivot_longer(tmp,cols = setdiff(names(tmp),"index"),
                           names_to = "columns",values_to = "yy_") %>%
      stats::na.omit() %>% applymap(as.integer) %>% as.data.frame()
    
    rownames(pp_pre) <- pp_pre$yy_
    pp_pre$yy_ <- NULL
    
    pp <- pp_pre[as.character(unlist(xx[1])), 'index']
    pp2 <- cbind(pp, unlist(xx[2])) %>% as.data.frame %>% 'colnames<-'(c('i', 'j'))
    pp3 <- unique(pp) %>% map(function(x){pp2[pp2$i == x,]})
    output <- list(list(
      yy %>% map(function(x){x[1]}),
      pp3 %>% lapply(function(x){
        sum(unlist(do.call(cbind, x)[, 'j'])) / length(unlist(x$j))})))
  }
  return(output)
}

#' calc_xpos: supporting function for DisconnectivityGraph
#'
#'
#' @param rcee List of component partitions across energy levels.
#' @return A list of x-position specifications for each level.
#' @noRd
calc_xpos <- function(rcee){
  xpositions <- list()
  for(rcee_ in rcee){
    if(length(xpositions) == 0){
      xpositions <- append(
        xpositions, 
        list(list(rcee_ %>% lapply(function(x){x[1]}), 1: length(rcee_))))
    }else{
      xpos_prev <- xpositions[length(xpositions)]
      xpositions <- append(xpositions, calc_xpos1(xpos_prev, rcee_))
    }
  }
  return(xpositions)
}

#' GraphObj: Generating graph objects for DisconnectivityGraph
#'
#' @description Creates a data frame describing nodes/branches for a disconnectivity graph.
#' From ELA output (\code{ela} = stable states, their energies, tipping points, and
#' tipping energies), this function:
#'  - constructs connectivity under decreasing energy thresholds,
#'  - derives component hierarchy,
#'  - assigns x-positions to components for visualization,
#'  - maps energies to representative points,
#' and returns a tidy object used by \link{DisconnectivityGraph}.
#' If \code{bootstrap_ene} is provided, statistical significance level of 
#' energy gap (i.e., energy difference between stable state and tipping point) 
#' in each stable state is computed. 
#'
#' @param ela Tuple/list: (stablestates, ssenergies, tippingpoints, tippingenergies).
#' @param bootstrap_ene Optional matrix/data frame of bootstrap energies
#' @importFrom stats quantile
#' @return A data frame containing component sets, key, energy,
#'   x-position, and representative point,
#'   (plus \code{qt_min/qt_max} when \code{bootstrap_ene} is given).
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
GraphObj <- function(ela,bootstrap_ene=NULL){
  stablestates <- ela[[1]]
  ssenergies <- ela[[2]]
  tippingpoints <- ela[[3]]
  tippingenergies <- ela[[4]]
  energyall <- setdiff(union(ssenergies, unlist(tippingenergies)), Inf)
  rre_pre <- rbind(
    cbind(ssenergies, stablestates) %>% 'colnames<-'(c('energy', 'point')),
    cbind(
      tippingenergies %>% as.data.frame %>% 'rownames<-'(
        1: length(tippingenergies)) %>% 'colnames<-'(
        1: length(t(tippingenergies))) %>% pivot_longer(
          cols=1: length(t(tippingenergies))),
    tippingpoints %>% as.data.frame %>% 'rownames<-'(
      1: length(tippingpoints)) %>% 'colnames<-'(
        1: length(t(tippingpoints))) %>% pivot_longer(
          cols=1: length(t(tippingpoints)))) %>%
    'colnames<-'(c('V1.1', 'energy', 'V1.2', 'point')) %>%
    select(c('energy', 'point')))
  rre <- rre_pre[!duplicated(rre_pre),]
  eee <- cbind(energyall, lapply(energyall, function(ww){calc_ccc(
    ww, stablestates, ssenergies, tippingenergies)})) %>% 'colnames<-'(c('energy', 'ccc')) %>%
    as.data.frame
  eee$energy <- unlist(eee[,'energy'])
  eee <- eee[order(eee$energy, decreasing=TRUE),]
  ese <- eee[!duplicated(eee$ccc, fromLast=TRUE),]
  eps <- ese$ccc
  ori <- eps[1][[1]]$'1'
  hee <- apply(cbind(eps %>% map(function(x){t(t(x))}), eps %>% map(
    function(x){t(setdiff(ori, unlist(x)))})), 1, function(x){unlist(
      x, recursive=FALSE)})
  cee <- calc_cee(hee)
  rcee <- cee[length(cee): 1]
  xpositions <- do.call(rbind, calc_xpos(rcee))[, 2][length(rcee): 1]
  invalid_shape <- lapply(xpositions, function(x){dim(data.frame(x))[1]}) != 1
  xpositions[invalid_shape] <- lapply(
    xpositions[invalid_shape],
    function(x){lapply(unlist(x), function(y){y})})
  nodes2xposi_pre <- cbind(
    cee=unlist(cee, recursive=FALSE),
    xpositions=unlist(xpositions, recursive=FALSE)) %>% as.data.frame
  nodes2xposi_pre$xpositions <- nodes2xposi_pre$xpositions %>% as.double
  nodes2xposi <- nodes2xposi_pre[!duplicated(nodes2xposi_pre),]
  ff_pre <- do.call(
    rbind,
    ese %>% apply(1, function(x){cbind(ccc=x$ccc, energy=x$energy)})) %>%
    as.data.frame
  ff_pre2 <- cbind(
    ccc=ff_pre$ccc,
    ccc_str=ff_pre$ccc %>% as.character,
    energy=ff_pre$energy %>% as.double)
  ff <- ff_pre2 %>% as.data.frame %>% left_join(
    cbind(
      cee=nodes2xposi$cee,
      cee_str=nodes2xposi$cee %>% as.character,
      nodes2xposi=nodes2xposi$xpositions) %>% as.data.frame,
    by=c('ccc_str' = 'cee_str'))
  ff <- ff[, c('cee', 'ccc_str', 'energy', 'nodes2xposi')]
  ff$ccc_str <- ff$ccc_str %>% as.character
  ff[, c('energy', 'nodes2xposi')] <- ff[, c('energy', 'nodes2xposi')] %>%
    apply(1, as.double) %>% T
  graphinfo_pre <- ff[order(ff$ccc_str, ff$energy),]
  graphinfo_pre2 <- graphinfo_pre[!duplicated(graphinfo_pre$ccc_str),]
  graphinfo_pre2$len_cee <- lapply(
    graphinfo_pre2$cee, function(x){length(x)}) %>% as.integer
  graphinfo_pre2$sum_cee <- lapply(graphinfo_pre2$cee, sum) %>% as.integer
  graphinfo <- graphinfo_pre2[order(
    graphinfo_pre2$len_cee, graphinfo_pre2$sum_cee),]
  mm <- as.data.frame(as.character(graphinfo$energy))
  colnames(mm) <- "energy"
  mmm <- (mm  %>% left_join(as.data.frame(rre),by='energy'))
  point <- mmm$point
  mmm <- mmm[!duplicated(mmm),]
  rownames(mmm) <- mmm$point
  graphinfo <- cbind(graphinfo, point)
  if(!is.null(bootstrap_ene)){
    ssobj <- graphinfo[graphinfo$len_cee==1,]
    for(i in 1:dim(ssobj)[1]){
      ss_str <- ssobj[i,"ccc_str"]
      ssid <- ssobj[i,"point"]
      tipp_pre <-  graphinfo[sapply(graphinfo$ccc_str, function(x) {
        v <- eval(parse(text=x))
        ss_str %in% unlist(v)
      }), ]
      tipp_pre <- tipp_pre[tipp_pre$len_cee > 1,]
      tippid <- tipp_pre[order(tipp_pre$len_cee),][1,"point"]
      graphinfo[graphinfo$ccc_str==ss_str,"tippid"] <- tippid
      deltaE <- bootstrap_ene[tippid,] - bootstrap_ene[ssid,]
      graphinfo[graphinfo$ccc_str==ss_str,'pvalue'] <- length(deltaE[deltaE < 0])/length(deltaE)
    }
    graphinfo_final <- graphinfo[, c(
      'cee', 'ccc_str', 'energy', 'nodes2xposi', 'point', 'tippid', 'pvalue')]
  }else{
    graphinfo <- cbind(graphinfo, point)
    graphinfo_final <- graphinfo[, c(
      'cee', 'ccc_str', 'energy', 'nodes2xposi', 'point')]
  }
  return(graphinfo_final)}


#' DisconnectivityGraph: Generate a disconnectivity graph
#'
#' @description
#' Draws a disconnectivity graph from the data frame returned by \code{GraphObj}.
#' Nodes are plotted at their assigned x-positions and energy levels, and are
#' connected to their parent components by paths.
#'
#' If \code{show_pvalue = TRUE}, approximate p-values for the energy gaps of
#' stable states are displayed. The p-values are shown on a log10 scale using
#' the values specified in \code{scale_labels}.
#'
#' @param grobj A data frame returned by \code{GraphObj}.
#' @param s Integer; number of species/features (used for labeling by \link{id2bin}).
#' @param DG_sample Character; title of the plot.
#' @param annot_adj Numeric vector of length 2; horizontal and vertical adjustment for labels.
#' @param min_pval Numeric (must be > 0); minimum p-value shown in the visualization. This is
#' usually set to the minimum detection level in bootstrap sampling
#' (e.g., if the number of bootstrap iterations is 1000, the minimum level is
#' 1/1000 = 0.001).
#' @param scale_labels Numeric vector of values between 0 and 1; labels for the
#' p-value scale. These values are converted to log-scaled values for plotting.
#' @param show_pvalue Logical; if \code{TRUE}, plots approximate p-values based
#' on the energy gaps between each stable state and its corresponding tipping
#' point estimated by bootstrap resampling.
#' @param getGraphObj Logical; if \code{TRUE}, returns the data frame used for
#' plotting the disconnectivity graph, so that users can manually modify the
#' graphical settings.
#' @param plot Logical; if TRUE, show disconnectivity graph 
#' @return
#' If \code{getGraphObj = TRUE}, returns a data frame containing energy values,
#' x-positions, and p-values for the energy gaps of the identified stable states
#' and tipping points. Otherwise, returns \code{NULL}.
#'
#' @importFrom purrr map
#' @importFrom ggplot2 ggplot aes geom_rect geom_point geom_text geom_path xlab ggtitle theme coord_cartesian scale_size scale_color_viridis_c
#' @importFrom rlang .data
#' @useDynLib ELAplus, .registration=TRUE
#' @export
DisconnectivityGraph <- function(
    grobj, s, DG_sample='DG_sample', annot_adj=c(0.75, 2.00), min_pval = 5e-02, scale_labels = c(0.05,0.1,0.2,0.5),
    show_pvalue=FALSE, getGraphObj=FALSE, plot =TRUE){

  range_ <- c(min(grobj$energy), max(grobj$energy))
  grobj_pre <- grobj
  grobj_pre$len_cee <- lapply(
    grobj_pre$cee, function(x){length(x)}) %>% as.integer
  grobj_ <- cbind(
    grobj_pre,
    do.call(
      cbind,
      grobj_pre$point %>% map(function(x){id2bin(x, s)})) %>% t
      )
  jun_pre <- grobj_[grobj_$len_cee == 1,]
  jun <- jun_pre[order(jun_pre$nodes2xposi),]
  jen_pre <- grobj_[grobj_$len_cee > 1,]
  jen <- jen_pre[order(jen_pre$nodes2xposi),]

  # grobj to plot
  grobj_to_plot <- data.frame()
  if(nrow(grobj_) != 1){
    for(i in 1: (nrow(grobj_))){
      aa <- grobj_[i,]
      #aa$point_str <- paste0('C', aa$point)
      aa$point_str <- aa$point
      bb_pre <- grobj_[grobj_$cee %>% map(
        function(x){all(unlist(aa$cee) %in% unlist(x))}) %>% unlist(),]
      bb <- bb_pre[bb_pre$ccc_str != aa$ccc_str,][1,]
      bb$point_str <- rep("", length(bb$point))
      # bb$point_str <- paste0('C', bb$point)
      between_aa_bb <- aa
      between_aa_bb$energy <- bb$energy
      between_aa_bb$point_str <- ''
      line_break <- between_aa_bb
      line_break$nodes2xposi <- NA
      line_break$energy <- NA
      grobj_to_plot_ <- rbind(aa, between_aa_bb, bb, line_break)
      grobj_to_plot_$id <- i
      grobj_to_plot <- grobj_to_plot %>% rbind(grobj_to_plot_)
    }
  }

  # scatter with annot and line
  g <- ggplot(
    grobj_to_plot,
    aes(x=.data$nodes2xposi, y=.data$energy, label=.data$point_str))
  xrange <- range(grobj_to_plot$nodes2xposi, na.rm=TRUE)
  yrange <- range(grobj_to_plot$energy, na.rm=TRUE)
  xlim <- c(xrange[[1]]-0.05,xrange[[2]]+0.05)
  ylim <- c(yrange[[1]]-0.25,yrange[[2]]+0.25)
  
  g <- g + xlab("") + geom_text(
    hjust=annot_adj[1], vjust=annot_adj[2], aes(fontface=2)) + geom_path()
  
  grobj_to_plot <- grobj_to_plot[!is.na(grobj_to_plot$len_cee), ]
  grobj_to_plot[as.logical((grobj_to_plot$len_cee > 1)*(grobj_to_plot$point_str != "")),"type"] <- "tipp"
  grobj_to_plot[as.logical((grobj_to_plot$len_cee == 1)*(grobj_to_plot$point_str != "")),"type"] <- "ss"
  
  grobj_point <- grobj_to_plot[grobj_to_plot$point_str != "",]
  grobj_tipp <- grobj_point[grobj_point$len_cee > 1,]
  grobj_ss <- grobj_point[grobj_point$len_cee == 1,]
  
  if (show_pvalue){
    #log_scale_labels <- scale_labels
    log_scale_labels <- log10(scale_labels)
    nom_scale_labels <- (log_scale_labels - min(log_scale_labels))/(max(log_scale_labels) - min(log_scale_labels))*3 + 1
    
    min_pval <- max(min(scale_labels),min_pval)
    max_pval <- max(scale_labels)
    grobj_ss[grobj_ss$pvalue < min_pval,"pvalue"] <- min_pval
    grobj_ss[grobj_ss$pvalue > max_pval,"pvalue"] <- max_pval
    log_pvalue <- log10(grobj_ss$pvalue)
    #log_pvalue <- grobj_ss$pvalue
    nom_pvalue <- (log_pvalue - min(log_scale_labels))/(max(log_scale_labels) - min(log_scale_labels))*3 + 1
    print(nom_pvalue)
    
    log_scale_labels <- trunc(log_scale_labels * 100) / 100
    
    grobj_ss[,"pval_size"] <- nom_pvalue
    g <- g + 
      geom_point(data = grobj_tipp, color = "black",shape = 15,size=2)
    
    g <- g +
      geom_point(
        data = grobj_ss,
        mapping = aes(
          x = .data$nodes2xposi,
          y = .data$energy,
          color = .data$pval_size,
          size = .data$pval_size
        )
      ) +
      scale_size(range = c(3, 7),limits = c(1,4), name="",
                 breaks=nom_scale_labels,labels=as.character(log_scale_labels)) +
      scale_color_viridis_c(option = "viridis",alpha=0.8,
                            direction=-1,na.value = "grey",limits=c(1,4),
                            name="log10(Bootstrap-approx. p-val)",
                            breaks=nom_scale_labels,labels=as.character(log_scale_labels))
  
  }else{
    g <- g + 
      geom_point(data = grobj_tipp, color = "black",shape = 15, size=2.5)
    
    g <- g +
      geom_point(data = grobj_ss,color="black",shape = 19, size = 3)
  }
  # pie
  #g <- g + geom_scatterpie(
  #  aes(x=nodes2xposi, y=energy),
  #  data=grobj_, cols=d_list)

  # plot
  g <- g + ggtitle(DG_sample) + coord_cartesian(xlim=xlim, ylim=ylim)
  if (plot){
    plot(g)
  }
  if (getGraphObj){
    return(grobj_to_plot)}
  else{
    return(NULL)
  }
}

#' showDG: draw DisconnectivityGraph
#'
#' @description Computing graph objects from \code{ela} (and optionally
#' bootstrap energies) and draw the disconnectivity graph for a given community
#' composition matrix \code{oc}. More than two stable state are required.
#'
#' @param ela ELA output tuple/list used by \code{GraphObj}.
#' @param oc Community composition matrix/data frame (used to infer \code{s = ncol(oc)}).
#' @param label Character; plot title.
#' @param bootstrap_ene Optional bootstrap energy matrix/data frame for computing statistical significance levels of energy gaps.
#' @param min_pval Numeric; minimum p-value shown in the visualization. This is
#' usually set to the minimum detection level in bootstrap sampling
#' (e.g., if the number of bootstrap iterations is 1000, the minimum level is
#' 1/1000 = 0.001).
#' @param scale_labels Numeric vector of values between 0 and 1; labels for the
#' p-value scale. These values are converted to log-scaled values for plotting.
#' @param plot Logical; if TRUE, show disconnectivity graph 
#' @param getGraphObj Logical; if TRUE, returns data.frame object used for plotting DisconnectivityGraph
#' Users can manually change the graphical configurations.
#' @return A data frame containing energy values, xpositions, pvalues for energy gaps in identified stable states and tipping points.
#' @useDynLib ELAplus, .registration=TRUE
#' @export
showDG <- function(ela, oc, label="",bootstrap_ene=NULL,min_pval = 5e-02, plot=TRUE,
                   scale_labels = c(0.05,0.1,0.2,0.5), getGraphObj=FALSE){
  if(length(ela[[1]])>1){
  s <- ncol(oc)
  if(is.null(bootstrap_ene)){
    grobj <- GraphObj(ela)
    plotObj <- DisconnectivityGraph(grobj, s, label, annot_adj=c(0.75, 2.00), min_pval = min_pval,
                                    scale_labels = scale_labels, getGraphObj = TRUE, plot =plot)
  }else{
    grobj <- GraphObj(ela,bootstrap_ene=bootstrap_ene)
    plotObj <- DisconnectivityGraph(grobj, s, label, annot_adj=c(0.75, 2.00), min_pval = min_pval,
                                    scale_labels = scale_labels,show_pvalue = TRUE,
                                    getGraphObj = TRUE, plot =plot)
  }
  if (getGraphObj){
    return(plotObj)
  }else{
    return(NULL)
  }
  }else{
      return(cat("only one stable state found\n"))
  }}

#' Plot histograms of bootstrap deltaE values
#'
#' For each stable state, this function computes the difference between the energy values 
#' of the corresponding tipping point and the stable state across bootstrap samples,
#' then draws a histogram of the resulting deltaE values.
#'
#' @param bootstrap_ene A data frame containing energies in stable states and tipping points,
#' output of \link{bootstrap_ELA}.
#' @param grobj_to_plot A data frame containing stable state - tipping point relationship,
#' output of \link{showDG}
#' @param getData Logical. If `TRUE`, return the computed deltaE values as a list.
#'   If `FALSE`, return `NULL`.
#'
#' @return
#' If `getData = TRUE`, a list of numeric vectors, one per deltaE for each stable state.
#' If `getData = FALSE`, `NULL`.
#'
#' @export

deltaE_histogram <- function(bootstrap_ene, grobj_to_plot, getData = TRUE) {
  grobj_point <- grobj_to_plot[grobj_to_plot$point_str != "", ]
  grobj_tipp  <- grobj_point[grobj_point$len_cee > 1, ]
  ssobj       <- grobj_point[grobj_point$len_cee == 1, ]
  
  n <- nrow(ssobj)
  nr <- ceiling(sqrt(n))
  nc <- ceiling(n / nr)
  
  op <- graphics::par(mfrow = c(nr, nc), mar = c(4, 4, 2, 1))
  on.exit(graphics::par(op))
  
  deltaE_list <- vector("list", n)
  
  for (i in seq_len(n)) {
    ssid   <- ssobj[i, "point_str"]
    tippid <- ssobj[i, "tippid"]
    deltaE <- bootstrap_ene[tippid, ] - bootstrap_ene[ssid, ]
    
    graphics::hist(as.numeric(deltaE),
         main = paste0(ssid, "-", tippid),
         xlab = "deltaE")
    
    deltaE_list[[i]] <- deltaE
  }
  
  if (getData) {
    return(deltaE_list)
  } else {
    return(NULL)
  }
}


#' showIntrGraph: draw interaction matrix on a PCoA plot
#'
#' Computes a 2D PCoA embedding from the interaction matrix (\code{jest}) and overlays
#' interaction links above a threshold \code{th}. Positive and negative links are
#' drawn separately, and species names are annotated at their coordinates.
#'
#' @param ela \link{ELA} output tuple/list (used to obtain stable states etc.; currently
#'   mainly used for context).
#' @param sa Parameter set passed to \link{sa2params}.
#' @param th Numeric threshold for drawing links (absolute value).
#' @param annot_adj Numeric length-2; horizontal/vertical adjustment for labels.
#' @param env target environmental factor.
#' @return NULL.
#' @useDynLib ELAplus, .registration=TRUE
#' @importFrom ggplot2 scale_color_manual theme_classic
#' @export
showIntrGraph <- function(ela, sa, env=NULL, th=0., annot_adj=c(0.75, 2.00)){
  ######################
  sa2p <- sa2params(sa, env)
  hestp <- sa2p[[1]]
  jestp <- sa2p[[2]]
  gestp <- sa2p[[3]]
  hgestp <- sa2p[[4]] 
  ####################
  jest <- jestp
  s <- length(hestp)
  tss <- ela[[1]]
  tssen <- ela[[2]]
  tti <- ela[[3]]
  tien <- ela[[4]]
  #memberq <- do.call(cbind, tss %>% map(
  #  function(x){CIntegerDigits(ssid=x, n=s)})) %>%
  #  'colnames<-'(tss) %>%
  #  apply(2, function(x){names(x[x == 1]) %>% as.integer()}) %>%
  #  as.data.frame() %>% as.list() %>% 'names<-'(NULL)
  
  ############
  rn <- rownames(jest)
  colnames(jest) <- rn
  jj <- jest
  jest <- as.data.frame(jest)
  ############
  
  #fmq <- memberq %>% unlist() %>% unique() %>% sort()
  #tname <- colnames(jest)[fmq]
  tname <- colnames(jest)
  mat <- jest[tname, tname]
  mat$species1 <- rownames(mat)
  link_ <- mat %>% pivot_longer(cols=tname, names_to='species2') %>% as.data.frame()
  plink_pre <- link_[link_$value > th, c("species1", "species2"), drop = FALSE]
  qlink_pre <- link_[link_$value < -th, c("species1", "species2"), drop = FALSE]
  pcp <- as.data.frame(stats::cmdscale(jj, k=2)) %>%
    'colnames<-'(lapply(1:2, function(x){
      gsub(' ', '', paste('PCo', as.character(x)), fixed=TRUE)}))
  lmap <- cbind(
    link_[, c('species1', 'species2')],
    lmap=link_$value %>% map(function(x){min(max(-1, x), 1)}) %>% unlist())
  tmap <- cbind(
    link_[, c('species1', 'species2')],
    tmap=link_$value %>% map(function(x){min(max(-1, x), 1)}) %>% unlist())
  ltmap <- cbind(lmap, tmap=tmap$tmap)
  plink <- plink_pre %>% left_join(ltmap, by=c('species1', 'species2'))
  qlink <- qlink_pre %>% left_join(ltmap, by=c('species1', 'species2'))
  
  # grobj to plot
  grobj_to_plot <- data.frame()
  if(nrow(plink) != 1){
    for(i in 1: nrow(plink)){
      grobj_ <- pcp[unlist(plink[i, c('species1', 'species2')]),]
      grobj_$lmap <- plink[i, 'lmap']
      grobj_$tmap <- plink[i, 'tmap']
      grobj_$id <- 'positive'
      rownames(grobj_) <- NULL
      line_break <- grobj_[1,]
      line_break$PCo1 <- NA
      line_break$PCo2 <- NA
      line_break$tmap <- NA
      line_break$lmap <- NA
      grobj_to_plot <- rbind(grobj_to_plot, grobj_, line_break)
    }
  }
  if(nrow(qlink) != 1){
    for(i in 1: nrow(qlink)){
      grobj_ <- pcp[unlist(qlink[i, c('species1', 'species2')]),]
      grobj_$lmap <- qlink[i, 'lmap']
      grobj_$tmap <- qlink[i, 'tmap']
      grobj_$id <- 'negative'
      rownames(grobj_) <- NULL
      line_break <- grobj_[1,]
      line_break$PCo1 <- NA
      line_break$PCo2 <- NA
      line_break$tmap <- NA
      line_break$lmap <- NA
      grobj_to_plot <- rbind(grobj_to_plot, grobj_, line_break)
    }
  }
  grobj_to_plot$species <- NA
  for(i in 1: nrow(pcp)){
    pcp_ <- pcp[i,]
    grobj_ <- cbind(pcp_, lmap=Inf, tmap=1, id='species', species=rownames(pcp_))
    line_break <- as.list(grobj_ %>% sapply(function(x){NA}))
    line_break$id <- 'species'
    line_break$species <- rownames(pcp_)
    grobj_to_plot <- rbind(grobj_to_plot, grobj_, line_break)
  }
  grobj_to_plot$lmap2 <- grobj_to_plot$lmap %>% sapply(function(x){stats::pnorm(x)})
  grobj_to_plot$tmap2 <- grobj_to_plot$tmap %>% sapply(function(x){1 * abs(x)})
  
  g <- ggplot(
    grobj_to_plot,
    aes(x=.data$PCo1, y=.data$PCo2, label=.data$species, 
        color=.data$id, alpha=.data$lmap2))
  g <- g + geom_point() + geom_text(
    hjust=annot_adj[1], vjust=annot_adj[2], aes(fontface=2)) + geom_path() +
    scale_color_manual(values = c('red','blue',  'black'))
  g <- g + theme_classic()
  
  # plot
  plot(g)
  
  return(NULL)}


#' showSSD: draw stable state diagram
#'
#' @description Plots energy values of stable states across a factor (e.g., environmental or
#' control parameter) using the output from \code{gela}. Each unique stable state is
#' drawn as a separate line with points, and a legend is added.
#'
#' @param gela The output of \link{GradELA}. Object containing stable state lists, energies, the factor,
#'   and environment labels as used in the code.
#' @return NULL (produces a plot as side effect).
#' @importFrom RColorBrewer brewer.pal
#' @importFrom graphics lines
#' @useDynLib ELAplus, .registration=TRUE
#' @export
showSSD <- function(gela){
  mem <- foreach(i=gela[[1]]) %do% {
      i[[1]][[1]]
  }
  uqss <- unique(unlist(mem))

  memen <- foreach(i=gela[[1]]) %do% {
    i[[1]][[2]]
    }
  minen <- min(unlist(memen))
  maxen <- max(unlist(memen))

  cols <- rep(brewer.pal(8, "Set1"), times=100)

  yrange <- c(minen-abs(0.05*minen), maxen+abs(0.05*maxen))
  graphics::par(xpd=T)
  graphics::par(mar=c(6,6,3,12))

  cc <- 0
  factor <- gela[[2]]
  env <- gela[[3]]
  for(j in uqss){
    cc <- cc + 1
    
    po <- foreach(i = gela[[1]]) %do% {
      pp <- which(i[[1]][[1]] == j)
      if(identical(pp, integer(0))) NA else i[[1]][[2]][pp]
    }
    
    energy <- unlist(po)
    
    if(cc == 1){
      plot(factor, energy,
           type = "o",
           ylim = yrange,
           col = cols[cc],
           xlab = names(env[is.na(env)]),
           ylab = "Energy")
    } else {
      lines(factor, energy, type = "o", col = cols[cc])
    }
  }

  graphics::legend(graphics::par()$usr[2] + 0.2, graphics::par()$usr[4], legend = uqss, col = cols, pch=1, lty=1)}


#' showSSD_ggplot: draw stable state diagram using ggplot
#'
#' @description Plots energy values of stable states across a factor (e.g., environmental or
#' control parameter) using the output from \code{gela}. Each unique stable state is
#' drawn as a separate line with points, and a legend is added.
#'
#' @param gela The output of \link{GradELA}. Object containing stable state lists, energies, the factor,
#'   and environment labels as used in the code.
#' @param plot Logical; if TRUE, show SSD plot 
#' @param getSSDobj Logical; if TRUE, returns data.frame object used for plotting SSD diagram
#' Users can manually change the graphical configurations.
#' @return A data frame used for plotting SSD diagram.
#' @importFrom RColorBrewer brewer.pal
#' @useDynLib ELAplus, .registration=TRUE
#' @export
showSSD_ggplot <- function(gela, plot = TRUE, getSSDobj=TRUE) {
  egrad <- gela[[2]]
  ssd_df <- foreach(i = 1:length(egrad), .combine = "rbind") %do% {
    data.frame(
      env = egrad[[i]],
      SS_ID = gela[[1]][[i]][[1]][[1]],
      Energy = gela[[1]][[i]][[1]][[2]]
    )
  }
  
  if (plot) {
    g <- ggplot(ssd_df, aes(x = .data$env, y = .data$Energy, color = .data$SS_ID)) +
      geom_point(shape = 21) +
      geom_line(aes(group = .data$SS_ID))
    print(g)
  }
  if(getSSDobj){
    return(ssd_df)
  }else{return(NULL)}
}

#' GELA3D: draw 3D plot of energy landscape
#'
#' @description Draws a 3D surface of the energy landscape and overlays
#' line trajectories using \pkg{plot3D} function.
#'
#' @param gelsobj List with two elements: surface data frame and line data frame.
#' @param theta numeric value. plot3D option: Azimuth angle (in degrees) controlling rotation around the z-axis.
#' @param phi numeric value. plot3D option: Elevation angle (in degrees) controlling the vertical viewing position.
#' @importFrom stringr str_c
#' @return NULL (produces a 3D plot as side effect).
#' @useDynLib ELAplus, .registration=TRUE
#' @export
showGELA3D <- function(gelsobj,theta=25,phi=50) {
    # surface data
    surf3d <- gelsobj[[1]]
    # line data
    l3d <- gelsobj[[2]]
    GetMesh <- function (surf3d) {
        nx <- length(union(surf3d$x, c()))
        ny <- length(union(surf3d$y, c()))
        x <- matrix(0, nrow = nx, ncol = ny)
        y <- matrix(0, nrow = nx, ncol = ny)
        z <- matrix(0, nrow = nx, ncol = ny)
        for (i in 1:nx) {
            for (j in 1:ny) {
                x[i, j] <- surf3d[i + (j - 1) * nx, ]$x
                y[i, j] <- surf3d[i + (j - 1) * nx, ]$y
                z[i, j] <- surf3d[i + (j - 1) * nx, ]$z
            }
        }
        list(x, y, z)
    }
    x <- GetMesh(surf3d)[[1]]
    y <- GetMesh(surf3d)[[2]]
    z <- GetMesh(surf3d)[[3]]
    # surface plot
    plot3D::persp3D(
        x, y, z, 
        theta = theta, phi = phi, expand = 0.5, facets=NA, ticktype = "detailed",
        alpha = 0.1, 
        main = "GradELA with plot3D",
        xlab = "X", ylab = "Y", zlab = "Energy"
    )
    # lines plot
    for (a_l3d in split(l3d, l3d$Index)) {
        ifelse(length(a_l3d$x) == 1, type <- "p", type <- "l")
        x <- unlist(a_l3d$x)
        y <- unlist(a_l3d$y)
        z <- unlist(a_l3d$z)
        plot3D::scatter3D(
            x, y, z,
            type=type, add=TRUE, col='green', lwd=4
        )
        i <- a_l3d$Index[[1]]
        plot3D::text3D(
            mean(x), mean(y), mean(z)*0.95, 
            labels = str_c("C", i), add=TRUE, cex=1.2
        )
    }
}

#' Compute heat-bath update probabilities
#'
#' Internal helper for heat-bath sampling. Given a binary state matrix \code{y},
#' parameters \code{h} and \code{J}, and optionally environmental effects \code{g} with an
#' environmental matrix \code{enmat}, this function returns the per-entry update
#' probabilities used by the heat-bath step.
#'
#' @param y Numeric matrix of size \code{nproc x nn} (typically 0/1), current states.
#' @param h Numeric vector of length \code{nn}, implicit environmental coefficients.
#' @param J Numeric matrix of size \code{nn x nn}, interaction matrix.
#' @param g Optional numeric matrix of size \code{nn x ne}, explicit environmental coefficients.
#' @param enmat Optional numeric matrix of size \code{nproc x ne}, environmental values.
#'
#' @return A numeric matrix of update probabilities (same shape as \code{y} after
#'   internal transposition), with values in (0, 1).
#'
#' @noRd
logmon <- function(y, h, J, g = NULL, enmat = NULL) {
  si <- y %*% J
  if (is.null(enmat)) {
    term <- t(sweep(-si, 2, h, FUN = "-"))
  } else {
    gei <- enmat %*% t(g)
    hei <- sweep(gei, 2, h, FUN = "+")
    term <- t(-hei - si)
  }
  result <- 1 / (1 + exp(term))
  return(result)
}

#' Perform one heat-bath update step
#'
#' Updates each row (sample) of a binary state matrix by choosing one random
#' position per row and resampling it according to precomputed heat-bath
#' probabilities.
#' 
#' @param y Numeric matrix of size \code{nproc x nn} (typically 0/1), current states.
#' @param logmo Numeric matrix of probabilities compatible with \code{y}. Each row
#'   corresponds to a row in \code{y}; entries are used to sample 0/1 updates.
#'
#' @return A numeric matrix with the same dimensions as \code{y}, updated states.
#'
#' @export
OnestepHBS <- function(y, logmo) {
  m <- ncol(y)
  
  update_row <- function(row_y, row_logmo) {
    position <- sample.int(m, 1)
    ne <- ifelse(runif(1) < row_logmo[position], 1, 0)
    row_y[position] <- ne
    row_y
  }
  
  updated_y <- t(mapply(update_row, as.data.frame(t(y)), as.data.frame(t(logmo))))
  return(updated_y)
}

#' Heat-bath sampling of binary states
#'
#' Generates samples of binary state vectors using heat-bath updates under
#' an extended maximum pairwise entropy model parameterized 
#' by implicit environmental parameters \code{h} and interaction matrix \code{J}.
#' Optionally includes environmental effects \code{g}, in which case an environmental
#' matrix is generated internally.
#'
#' @param steps Integer. Number of heat-bath iterations to run.
#' @param nproc Integer. Number of parallel samples (rows) to maintain.
#' @param h Numeric vector of length \code{nn}, implicit environmental parameters.
#' @param J Numeric matrix of size \code{nn x nn}, interaction matrix.
#' @param g Optional numeric matrix of size \code{nn x ne}, environmental coefficients.
#'
#' @return A list with components:
#' \describe{
#'   \item{y}{Numeric matrix (\code{nproc x nn}) of sampled states (0/1).}
#'   \item{enmat}{Numeric matrix (\code{nproc x ne}) of environmental values, or \code{NULL}
#'     if \code{g} is \code{NULL}.}
#' }
#'
#' @examples
#' set.seed(1)
#' nn <- 5
#' pars <- hb.paramgen(nn)
#' res <- HeatBath(steps = 10, nproc = 8, h = pars[[1]], J = pars[[2]])
#' str(res)
#'
#' @importFrom stats runif
#' @export
HeatBath <- function(steps, nproc, h, J, g = NULL) {
  nn <- length(h)
  speciesname <- paste0("species.", seq_len(nn))
  samplename <- paste0("sample.", seq_len(nproc))
  
  y <- array(0, dim = c(nproc, nn), dimnames = list(samplename, speciesname))
  
  if (!is.null(g)) {
    ne <- dim(g)[2]
    factorname <- paste0("factor.", seq_len(ne))
    enmat <- matrix(runif(nproc * ne, min = 0, max = 1),
                    nrow = nproc, ncol = ne,
                    dimnames = list(samplename, factorname))
  } else {
    enmat <- NULL
  }
  
  for (i in seq_len(steps)) {
    logmo <- logmon(y, h = h, J = J, g = g, enmat = enmat)
    y <- OnestepHBS(y, t(logmo))
  }
  
  colnames(y) <- speciesname
  return(list(y = y, enmat = enmat))
}

#' Randomly generate model parameters for heat-bath sampling
#'
#' Generates random implicit environmental parameters \code{h} and a interaction matrix \code{J}. 
#' Optionally generates explicit environmental coefficients \code{g}. 
#' The \code{j_connectivity} parameter controls the fraction of non-zero off-diagonal interactions.
#'
#' @param nn Integer. Number of units (species, taxa, or members).
#' @param ne Integer. Number of environmental factors. If \code{ne > 0}, \code{g} is generated.
#' @param j_connectivity Numeric in [0, 1]. Fraction of possible pairwise
#'   interactions set to non-zero.
#' @importFrom stats setNames
#' @return A list containing at least:
#' \describe{
#'   \item{h}{Named numeric vector of length \code{nn}. Implicit environmental parameters}
#'   \item{J}{Named numeric matrix (\code{nn x nn}), symmetric. Interaction matrix}
#' }
#' If \code{ne > 0}, also includes:
#' \describe{
#'   \item{g}{Named numeric matrix (\code{nn x ne}). Explicit environmental parameters}
#' }
#'
#'
#' @export
hb.paramgen <- function(nn, ne = 0, j_connectivity = 1) {
  speciesname <- paste0("species.", seq_len(nn))
  
  h.act <- setNames(runif(nn, min = -2, max = 2), speciesname)
  
  rnp <- function() {
    tuples <- t(combn(nn, 2))
    rnp_list <- lapply(seq_len(nrow(tuples)), function(i) {
      if (runif(1) < 0.5) rev(tuples[i, ]) else tuples[i, ]
    })
    do.call(rbind, rnp_list)
  }
  
  generate_matrix <- function(rnp_list) {
    nonzerolist <- sample.int(nrow(rnp_list), nrow(rnp_list) * j_connectivity)
    ma <- matrix(0, nn, nn)
    for (i in seq_len(nrow(rnp_list))) {
      index <- rnp_list[i, ]
      ma[index[1], index[2]] <- ifelse(i %in% nonzerolist, runif(1, min = -2, max = 2), 0)
    }
    ma
  }
  
  ma <- generate_matrix(rnp())
  J <- `dimnames<-`(ma + t(ma), list(speciesname, speciesname))
  
  if (ne > 0) {
    gname <- paste0("g.", seq_len(ne))
    g <- matrix(runif(nn * ne, min = -2, max = 2), nrow = nn, ncol = ne)
    g <- `dimnames<-`(g, list(speciesname, gname))
    return(list(h = h.act, J = J, g = g))
  }
  
  return(list(h = h.act, J = J))
}

#' Gradient descent estimation of model parameters
#'
#' Estimates parameters \code{h} and \code{J} by iterative gradient updates to match
#' first and second moments of the observed binary data under an maximum
#' pairwise entropy model.
#'
#' @param data Numeric matrix (typically 0/1). Rows are samples and columns are species or taxa.
#' @param maxit Integer. Maximum number of gradient descent iterations.
#'
#' @return A list with elements:
#' \describe{
#'   \item{h.est}{Numeric vector of estimated implicit environmental parameters.}
#'   \item{j.est}{Numeric matrix of estimated interaction matrix.}
#' }
#'
#' @details
#' This function calls \code{Energy()} to compute energies for candidate states.
#' Ensure that \code{Energy} is available in the package namespace.
#'
#' @export
GradientDescent <- function(data, maxit) {
  units <- ncol(data)
  e <- 10^-8
  datalength <- nrow(data)
  tuples <- as.matrix(expand.grid(rep(list(0:1), units)))
  
  semp <- colMeans(data)
  ssemp <- t(data) %*% data / datalength * abs(diag(units) - 1)
  
  htt <- rep(0, units)
  jtt <- matrix(0, units, units)
  ttt <- 0
  
  while (ttt < maxit) {
    ht <- htt
    jt <- jtt
    
    e0 <- apply(tuples, 1, Energy, htt, jtt)
    en <- exp(-e0)
    prb <- en / sum(en)
    
    pt <- prb * tuples
    smod <- colSums(pt)
    ssmod <- (t(pt) %*% tuples) * abs(diag(units) - 1)
    
    dh <- 0.1 * ((semp + e) - (smod + e))
    dj <- 0.1 * ((ssemp + e) - (ssmod + e))
    
    htt <- ht + dh
    jtt <- jt + dj
    ttt <- ttt + 1
  }
  
  list(h.est = htt, j.est = jtt)
}

#' Abundance table for 16 taxa across 256 samples
#'
#' `baseabtable` is a data frame containing abundance for 16 taxa across 256 samples.
#'
#' Each row corresponds to a sample, and each column represents
#' the abundance of a specific taxon.
#'
#' @format A data frame with 256 rows and 16 variables:
#'
#' @details
#' The values represent abundance measurements for each taxon in each sample.
#' The exact definition of "abundance" (e.g., counts, relative abundance, or normalized values)
#' depends on the data preprocessing pipeline used.
#'
#' @source Generated for example purposes (replace with actual source if applicable)
#'
#' @examples
#' data(baseabtable)
#' head(baseabtable)
#' formatted <- Formatting(baseabtable = baseabtable)
"baseabtable"

#' Environmental metadata for 256 samples
#'
#' `basemetadata` is a data frame containing environmental parameters
#' associated with the same 256 samples as in \code{baseabtable}.
#'
#' Each row corresponds to a sample, and each column represents
#' an environmental variable.
#'
#' @format A data frame with 256 rows and 2 variables:
#'
#' @details
#' The rows in this data frame correspond to the same samples and
#' ordering as in \code{baseabtable}. These variables can be used
#' for ELA with environmental parameters such as \code{gradELA}
#'
#' @source Generated for example purposes (replace with actual source if applicable)
#'
#' @examples
#' data(baseabtable)
#' data(basemetadata)
#' head(basemetadata)
#' formatted <- Formatting(baseabtable = baseabtable, basemetadata = basemetadata)
"basemetadata"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

SA <- function(ocData, we = 0.001, totalit = 1000L, runadamW = TRUE, sparse = TRUE, lambda = 0.01, intv = 100L, maxlearningrate = 0.005, warmup_tt = 1) {
    .Call(`_ELAplus_SA`, ocData, we, totalit, runadamW, sparse, lambda, intv, maxlearningrate, warmup_tt)
}

fullSA <- function(ocData, envData, runadamW = TRUE, sparse = TRUE, we = 0.001, totalit = 1000L, lambda = 0.01, intv = 100L, maxlearningrate = 0.005, warmup_tt = 1) {
    .Call(`_ELAplus_fullSA`, ocData, envData, runadamW, sparse, we, totalit, lambda, intv, maxlearningrate, warmup_tt)
}

cEnergy <- function(state, alpha, beta) {
    .Call(`_ELAplus_cEnergy`, state, alpha, beta)
}

SteepestDescent_cpp <- function(state, alpha, beta) {
    .Call(`_ELAplus_SteepestDescent_cpp`, state, alpha, beta)
}

SSestimate <- function(alpha, beta, itr = 20000L) {
    .Call(`_ELAplus_SSestimate`, alpha, beta, itr)
}

FindingTippingpoint_cpp <- function(s1, s2, alpha, jj, tmax = 10000L) {
    .Call(`_ELAplus_FindingTippingpoint_cpp`, s1, s2, alpha, jj, tmax)
}

TPestimate <- function(comb, minset, alpha, beta, tmax = 10000L) {
    .Call(`_ELAplus_TPestimate`, comb, minset, alpha, beta, tmax)
}

SSentropy_cpp <- function(uoc, ss, alpha, beta, seitr = 1000L, convTime = 10000L) {
    .Call(`_ELAplus_SSentropy_cpp`, uoc, ss, alpha, beta, seitr, convTime)
}

#' Energy landscape analysis in parallel mode
#' @description Identifying stable states and tipping points.
#'
#' This function estimates stable states by repeated steepest-descent searches
#' \code{SSestimate} and then checks tipping points between all pairs of stable
#' states using \code{FindingTippingpoint_cpp}. 
#' The result is summarized and converted by \code{elaconvert}
#' into a list of stable state IDs, their energies, tipping point IDs, and
#' tipping energies.
#'
#' @param sa output of runSA function.
#' @param env vector of environmental factors for specifying a single environmental condition (optional).
#' @param SS.itr number of iterations of the process to identify the stable states.
#' @param FindingTip.itr number of iterations of the process to identify the tipping points.
#' @param threads number of cores (threads) used for calculation.
#' @param reporting enable/disable the reporting function.
#' @param fork logical; if TRUE and \code{threads > 1}, a cluster is created internally for parallel execution.
#'
#' @return A list returned by \code{elaconvert}: (1) stable state IDs, (2) stable state energies,
#'   (3) tipping point ID matrix/list, and (4) tipping energies matrix/list.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @importFrom Rcpp sourceCpp
#' @importFrom foreach foreach
#' @export
ELA <- function(sa=sa, env=NULL, SS.itr=20000, FindingTip.itr=10000,
                 threads=1, reporting=TRUE, fork=TRUE){
  ######################
  sa2p <- sa2params(sa, env)
  hestp <- sa2p[[1]]
  jestp <- sa2p[[2]]
  gestp <- sa2p[[3]]
  hgestp <- sa2p[[4]]
  if(threads>1&&fork==TRUE){cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
      parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
      parallel::clusterSetRNGStream(cluster)
      doParallel::registerDoParallel(cluster)
      on.exit(parallel::stopCluster(cluster), add = TRUE)}
	    
  ####################
  ## ||||||||||||||||||||||||||||||||||||| ##
  if(reporting){cat('Start ELA:\n')}
  start <- proc.time()[3]
	speciesName <- rownames(jestp)

	## ||||||||||||||||||||||||||||||||||||| ##
	## -- Stable state estimatin
	minsets <- as.data.frame(SSestimate(hgestp, jestp, itr = SS.itr))
    uqm <- unique(minsets[,-length(minsets[1,])])
    energy <- apply(uqm,1,function(x){Energy(x,hgestp,jestp)})
    minsets <- cbind(uqm, energy)
    minsets <- as.matrix(minsets[order(minsets[,ncol(minsets)]),])
    colnames(minsets) <- c(speciesName, "energy")
    if(reporting){cat(nrow(minsets),"stable states were found.\n")}
	sampleSS <- data.frame(minsets)
	ssid <- sprintf("SS_%s", formatC(1:nrow(minsets), width = nchar(ncol(minsets)),
	                                 flag = "0"))
	rownames(minsets) <- ssid

	## ||||||||||||||||||||||||||||||||||||| ##
	## -- Tipping point estimation

	if(length(ssid)>1){
	comb <- expand.grid(1:nrow(minsets), 1:nrow(minsets))
	comb <- comb[comb[, 1] <comb[, 2], 2:1]

	tpnodeID <- sprintf("TPnode_%s", formatC(1:nrow(comb), width = nchar(nrow(comb)),
	                             flag = "0"))
  if(reporting){cat('Checking', nrow(comb), 'tipping points.\n')}
    if(threads>1){
	FindTip <- foreach (idx = 1:nrow(comb), .combine=rbind, .packages="ELAplus") %dopar% {

	    minsetsub <- minsets[as.integer(comb[idx, ]), ]
	    ss1 = minsetsub[1, -ncol(minsets)]
	    ss2 = minsetsub[2, -ncol(minsets)]

	    tippoint.tmp = FindingTippingpoint_cpp(s1 = ss1,  s2 = ss2,
	                                           alpha = hgestp, jj = jestp,
	                                           tmax = FindingTip.itr)

	    colnames(tippoint.tmp) <- c(speciesName, "energy")

	    ss <- data.frame(SS1=rownames(minsetsub)[1], SS1.energy=minsetsub[1,ncol(minsetsub)],
	               SS2=rownames(minsetsub)[2], SS2.energy=minsetsub[2,ncol(minsetsub)])
	    tp <- data.frame(TP=tpnodeID[idx], tippoint.tmp)
	    return( cbind(ss, tp))
	}}else{
		FindTip <- foreach (idx = 1:nrow(comb), .combine=rbind) %do% {

	    minsetsub <- minsets[as.integer(comb[idx, ]), ]
	    ss1 = minsetsub[1, -ncol(minsets)]
	    ss2 = minsetsub[2, -ncol(minsets)]

	    tippoint.tmp = FindingTippingpoint_cpp(s1 = ss1,  s2 = ss2,
	                                           alpha = hgestp, jj = jestp,
	                                           tmax = FindingTip.itr)

	    colnames(tippoint.tmp) <- c(speciesName, "energy")

	    ss <- data.frame(SS1=rownames(minsetsub)[1], SS1.energy=minsetsub[1,ncol(minsetsub)],
	               SS2=rownames(minsetsub)[2], SS2.energy=minsetsub[2,ncol(minsetsub)])
	    tp <- data.frame(TP=tpnodeID[idx], tippoint.tmp)
	    return( cbind(ss, tp))
	}}
	}

	## ||||||||||||||||||||||||||||||||||||| ##
	## -- Summarize
	if(length(ssid)>1){
	tpState <- (FindTip[,-c(1:5)]);
    uqt <- unique(tpState[,-length(tpState[1,])])
    energy <- apply(uqt,1,function(x){Energy(x,hgestp,jestp)})
    tpStateUni <- cbind(uqt, energy)
	tpid <- sprintf("TP_%s", formatC(1:nrow(tpStateUni), width = nchar(nrow(tpStateUni)),
	                                 flag = "0"))
	rownames(tpStateUni) <- tpid

	stateInfo <- rbind(data.frame(stateID=rownames(minsets), state=rep('Stable state', nrow(minsets)), minsets),
	                   data.frame(stateID=rownames(tpStateUni), state=rep('Tipping point', nrow(tpStateUni)), tpStateUni))
	                   }
	                   else{
	                   stateInfo <- data.frame(stateID=rownames(minsets), state=rep('Stable state', nrow(minsets)), minsets)
	                   }

	# ||||||||||||||||||||||||||||||||||||| ##
	if(length(ssid)>1){
	Stable <- FindTip[,c(1:4)]

	tmp <- rownames(tpStateUni); names(tmp) <-apply(tpStateUni[,-ncol(tpStateUni)], 1, paste, collapse='')
	ordered <- apply(tpState[,-ncol(tpState)], 1, paste, collapse='')
	tpidVec <- tmp[ordered]

	network <- data.frame(Stable, TP=tpidVec, Tp.energy=FindTip[,ncol(FindTip)])
	elasummary <- list(stateInfo = stateInfo, network =network)
	}else{
	elasummary <- list(stateInfo = stateInfo, network =NULL)
	}

	# ||||||||||||||||||||||||||||||||||||| ##
	if(reporting==TRUE){cat("converting...\n")}
  elaconv <- elaconvert(elasummary)
  end <- proc.time()[3]
	if(reporting==TRUE){cat(sprintf("ELA: elapsed time %.2f sec\n", end - start))}
	return(elaconv)}


#' ELA convert
#' @description Converting output of the ELA function.
#'
#' Converts the raw ELA summary into the standard
#' \code{ela} tuple used in downstream visualization/analysis functions:
#' stable state IDs (integer IDs converted from binary states), their energies,
#' tipping point IDs (with "Inf" for missing pairs), and tipping energies.
#'
#' @param ela output of the ELA function.
#' @return A list: (stablestates, ssenergy, tippingpoints, tpenergy).
#' @importFrom magrittr %>%
#' @noRd
elaconvert <- function(ela){
  stablestates <- apply(ela$stateInfo[ela$stateInfo$state=='Stable state', 3:(length(colnames(ela$stateInfo))-1)],
                        1,
                        bin2id) |> as.vector()

  ssenergy <- ela$stateInfo[ela$stateInfo$state=='Stable state','energy']

  # number of stable state
  num_ss <- nrow(ela$stateInfo[grepl('SS', ela$stateInfo$stateID),])
  num_d <- nchar(ncol(ela$stateInfo)-2)
  get_tp <- function(i, j) {
      ss1 <- sprintf("SS_%s", formatC(i, width = num_d,flag="0"))
      ss2 <- sprintf("SS_%s", formatC(j, width = num_d,flag="0"))
      return(ela$network[ela$network$SS1==ss1 & ela$network$SS2==ss2,
                        "TP"])}

  # tippingpoints
  if(num_ss>1){
  tippingpoints <- lapply(matrix(1:num_ss), function(i)
  apply(matrix(1:num_ss), 1, function(j) {
      tp <- get_tp(i, j)
      if(length(tp)==0){return("Inf")}
      bin <- ela$stateInfo[ela$stateInfo$stateID==tp, 3:(length(colnames(ela$stateInfo))-1)]
                                          return(apply(bin,1, bin2id))
                                          })) %>% 'names<-'(sapply(1:num_ss, \(i) paste(i, sep="")))
  }else{
      tippingpoints <- lapply(matrix(1:1),function(i) apply(matrix(1:1),1, function(j) {
      return("Inf")
      })) %>% 'names<-'(sapply(1:1, \(i) paste(i, sep="")))
  }

  ## tpenergy
  if(num_ss>1){
  tpenergy <- lapply(matrix(1:num_ss),function(i) apply(matrix(1:num_ss),1, function(j) {
      tp <- get_tp(i, j)
      if(length(tp)==0){return(Inf)}
      return(ela$stateInfo[ela$stateInfo$stateID==tp, 'energy'])
      })) %>% 'names<-'(sapply(1:num_ss, \(i) paste(i, sep="")))
  }else{
      tpenergy <- lapply(matrix(1:1),function(i) apply(matrix(1:1),1, function(j) {
      return(Inf)
      })) %>% 'names<-'(sapply(1:1, \(i) paste(i, sep="")))
  }

  return(list(stablestates, ssenergy, tippingpoints, tpenergy))}


#' ELPruning
#' @description Pruning shallow stable states of energy landscape.
#'
#' Iteratively prunes stable states (merges basins). 
#' If \code{type="relative"},  stable state whose energy depth is smaller than \code{th} times the deepest
#' basin depth are pruned and merged to the deeper basins. 
#' If \code{type="bootstrap"}, stable state whose bootstrap confidence interval (default: 95% quantile) of energy gap between tipping point
#' cross zero is pruned. 
#' The function updates stable states and corresponding tipping
#' structures, and returns both the pruned ELA object and a mapping table
#' from stable states before pruning to those after pruning.
#'
#' @param ela output of ELA function.
#' @param th ratio of the depth of the basin to be pruned with respect to the deepest basin.
#' @param type the type of pruning \code{"relative"}(use with \code{th}) or \code{"bootstrap"} (use with \code{lower_qr} and \code{bootstrap_ene}) 
#' @param lower_qr Numeric value. Quantile used for constructing confidence intervals from bootstrap samples to check 
#' whether the estimated energy gaps include zero. Default is \code{lower_qr = 0.05}.
#' @param bootstrap_ene a data frame of bootstrap energies generated by \link{bootstrap_ELA}. Necessary when \code{type="quantile"}.
#' @param threads number of cores (threads) used for calculation.
#' @param reporting enable/disable the reporting function.
#' @param fork logical; if TRUE and \code{threads > 1}, a cluster is created internally for parallel execution.
#'
#' @return A list of two elements: (1) pruned ELA tuple (same format as \link{ELA} output),
#'   and (2) a data frame with columns \code{ss.before.pruning} and \code{ss.after.pruning}.
#'
#' @export
ELPruning <- function(ela, type="relative",th=0.05, lower_qr=0.05, bootstrap_ene = NULL,
                      threads=1, reporting=TRUE, fork=TRUE){
  start <- proc.time()[3]
  if(reporting){cat('Start pruning:\n')}
  
  if(threads>1&&fork==TRUE){cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
  parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
  parallel::clusterSetRNGStream(cluster)
  doParallel::registerDoParallel(cluster)
  on.exit(parallel::stopCluster(cluster), add = TRUE)}
  
  tss <- ela[[1]]
  tssen <- ela[[2]]
  tti <- ela[[3]]
  ttien <- ela[[4]]
  if(length(tss)>1){
    rat <- as.data.frame(gtools::permutations(
      n=length(tss), r=2, v=1:length(tss), repeats.allowed = TRUE))
    rat <- rat[rat$V1 < rat$V2, , drop = FALSE]
    tss0 <- tss
    ssrep <- tss
    if(type=="relative"){
      #%>% subset(V1 < V2)
      if(reporting==TRUE){cat("*")}
      cmpax <- calc_mist_pax(rat, ttien, tssen, threads)
      if(reporting==TRUE){cat(".")}
      mist <- as.numeric(unlist(cmpax[,(dim(cmpax)[2]-1):(dim(cmpax)[2])]))
      pax <- cmpax[,-dim(cmpax)[2]:-(dim(cmpax)[2]-1)]
      paxmax <- max(mist)
      tt <- 0
      while(min(pax['val1']) < th * paxmax){
          tt <- tt + 1
          rat <- as.data.frame(gtools::permutations(
              n=length(tss), r=2, v=1:length(tss), repeats.allowed = TRUE))
          #%>% subset(V1 < V2)
          rat <- rat[rat$V1 < rat$V2, , drop = FALSE]
          cmpax <- calc_mist_pax(rat, ttien, tssen, threads)
          if(reporting==TRUE){if(tt%%10 == 0){cat("*")}else{cat(".")}}
          mist <- as.numeric(unlist(cmpax[,(dim(cmpax)[2]-1):(dim(cmpax)[2])]))
          pax <- cmpax[,-dim(cmpax)[2]:-(dim(cmpax)[2]-1)]
          paxmax <- max(mist)
        if(min(pax['val1']) >= th * paxmax){break}else{
            pp <- pax[order(pax$val1),][1, c('val2_1', 'val2_2')] %>% as.list() %>% unlist()
            ssrep <- replace_value(tss[pp[2]], ssrep, tss[pp[1]])
            tss <- tss[tss != tss[pp[2]]]
            tssen <- tssen[tssen != tssen[pp[2]]]
            tti_pre <- tti %>% as.data.frame()
            tti <- tti_pre[rownames(tti_pre) != rownames(tti_pre)[pp[2]],
                           colnames(tti_pre) != colnames(tti_pre)[pp[2]]] %>% as.list()
            ttien_pre <- ttien %>% as.data.frame()
            ttien <- ttien_pre[rownames(ttien_pre) != rownames(ttien_pre)[pp[2]],
                               colnames(ttien_pre) != colnames(ttien_pre)[pp[2]]] %>% as.list()
          }
        if(length(unique(ssrep)) == 1){break}
      }
    }else if (type=="bootstrap"){
      if(is.null(bootstrap_ene)){cat("bootstrap results of ela is missing. Please specify by 'bootstrap_ene='")}
      # Using quantile range in bootstrap energy calculation for pruning
      done <- FALSE
      while(done != TRUE){
        rat <- as.data.frame(gtools::permutations(
          n=length(tss), r=2, v=1:length(tss), repeats.allowed = TRUE))
        rat <- rat[rat$V1 < rat$V2, , drop = FALSE]
        #qr_overlap <- calc_overlap_qr(tss, rat, tssen, tti, ttien, bootstrap_ene, lower_qr, threads)
        qr_overlap <- calc_deltaE_range(tss, rat, tssen, tti, ttien, bootstrap_ene, lower_qr, threads)
        qr_overlap <- qr_overlap[order(qr_overlap$val1),]
        #print(qr_overlap)
        for (i in 1:dim(qr_overlap)[1]) {
          #if (qr_overlap[i,"overlap1"] || qr_overlap[i,"overlap2"]){
          if (qr_overlap[i,"pvalue1"] > lower_qr || qr_overlap[i,"pvalue2"] > lower_qr){
            pp <- qr_overlap[i, c('val2_1', 'val2_2')]%>% as.list() %>% unlist()
            ssrep <- replace_value(tss[pp[2]], ssrep, tss[pp[1]])
            tss <- tss[tss != tss[pp[2]]]
            tssen <- tssen[tssen != tssen[pp[2]]]
            tti_pre <- tti %>% as.data.frame()
            tti <- tti_pre[rownames(tti_pre) != rownames(tti_pre)[pp[2]],
                           colnames(tti_pre) != colnames(tti_pre)[pp[2]]] %>% as.list()
            ttien_pre <- ttien %>% as.data.frame()
            ttien <- ttien_pre[rownames(ttien_pre) != rownames(ttien_pre)[pp[2]],
                               colnames(ttien_pre) != colnames(ttien_pre)[pp[2]]] %>% as.list()
            done <- FALSE
            break
          }
          done <- TRUE
        }
        if(length(unique(ssrep)) == 1){
          done <- TRUE
          break
        }
      }
    }
  tti <- tti %>% as.data.frame() %>% 'colnames<-'(1: length(tti)) %>% as.list()
  ttien <- ttien %>% as.data.frame() %>% 'colnames<-'(1: length(ttien)) %>% as.list()
  end <- proc.time()[3]
  if(reporting==TRUE){cat(sprintf("\nELPruning: elapsed time %.2f sec\n", end - start))}
    states <- as.data.frame(cbind(tss0, ssrep))
    colnames(states) <- c("ss.before.pruning", "ss.after.pruning")
    return(list(list(tss, tssen, tti, ttien), states))
  }else{
    end <- proc.time()[3]
    if(reporting==TRUE){cat(sprintf("\nELPruning: elapsed time %.2f sec\n", end - start))}
    states <- as.data.frame(cbind(tss, tss))
    colnames(states) <- c("ss.before.pruning", "ss.after.pruning")
    return(list(list(tss, tssen, tti, ttien), states))
  }}


#'Calc_overlap_qr
#'@description supporting function for ELpruning with absolute value
#' @param tss A string vector (stable states) 
#' @param rat A data.frame
#' @param tti A matix (tipping points).
#' @param bootstrap_ene A data.frame bootstrap results of estimated energies
#' @param qrange A vector (length: 2) quantile range 
#' @param tssen A numeric vector (energy of stable states).
#' @param ttien A matrix-like object convertible to a numeric matrix.
#' @param threads Integer. If greater than 1, parallel execution will be run.
#' 
#' @return A data frame
#' @importFrom stats quantile
#' @noRd
calc_overlap_qr <- function(tss, rat, tssen, tti, ttien, bootstrap_ene, qrange, threads){
  
  overlap_interval <- function(a_min, a_max, b_min, b_max) {
    v <- c(a_min, a_max, b_min, b_max)
    if (anyNA(v) || any(!is.finite(v))) return(FALSE)
    (a_min <= b_max) && (b_min <= a_max)
  }
  qr_ela <- as.matrix(t(apply(bootstrap_ene, 1, function(x) quantile(x, probs = qrange,na.rm=TRUE))))
  qr_col <- sapply(qrange,function(x){paste0(as.character(x*100),"%")})
  pax <- rat
  npax <- max(pax['V2'])
  tti.m <- as.matrix(tti)
  ttien.m <- as.matrix(ttien)
  ### requires starting clusters out side the function ###
  if(threads>1){a <- foreach(i = pax[,'V1'], j = pax[,'V2'], .combine="rbind", .packages=c("tidyverse","ELAplus")) %dopar% {
    row_ <- as.character(npax*(i-1) + j)
    tipp_id <- tti.m[[j]][i]
    te <- ttien.m[[j]][i]
    tipp_qr <- c(qr_ela[tipp_id,qr_col[1]],qr_ela[tipp_id,qr_col[2]])
    se <- c(tssen[i], tssen[j])
    id1 <- tss[[i]]
    id2 <- tss[[j]]
    # check overlap in the quantile range
    pax[row_, 'val1'] <- min(te - se[1], te - se[2])
    val2_pre <- pax[row_, c('V1', 'V2')] %>% T %>% cbind(se)
    val2_pre <- val2_pre[order(val2_pre$se)][,row_]
    pax[row_, 'val2_1'] <- val2_pre[1]
    pax[row_, 'val2_2'] <- val2_pre[2]
    pax[row_, 'overlap1'] <- overlap_interval(tipp_qr[1],tipp_qr[2],qr_ela[id1,qr_col[1]], qr_ela[id1,qr_col[2]])
    pax[row_, 'overlap2'] <- overlap_interval(tipp_qr[1],tipp_qr[2],qr_ela[id2,qr_col[1]], qr_ela[id2,qr_col[2]])
    pax[row_,]
  }}else{a <- foreach(i = pax[,'V1'], j = pax[,'V2'], .combine="rbind") %do% {
    row_ <- as.character(npax*(i-1) + j)
    tipp_id <- tti.m[[j]][i]
    te <- ttien.m[[j]][i]
    tipp_qr <- c(qr_ela[tipp_id,qr_col[1]],qr_ela[tipp_id,qr_col[2]])
    se <- c(tssen[i], tssen[j])
    id1 <- tss[[i]]
    id2 <- tss[[j]]
    # check overlap in the quantile range
    pax[row_, 'val1'] <- min(te - se[1], te - se[2])
    val2_pre <- pax[row_, c('V1', 'V2')] %>% T %>% cbind(se)
    val2_pre <- val2_pre[order(val2_pre$se)][,row_]
    pax[row_, 'val2_1'] <- val2_pre[1]
    pax[row_, 'val2_2'] <- val2_pre[2]
    pax[row_, 'overlap1'] <- overlap_interval(tipp_qr[1],tipp_qr[2],qr_ela[id1,qr_col[1]], qr_ela[id1,qr_col[2]])
    pax[row_, 'overlap2'] <- overlap_interval(tipp_qr[1],tipp_qr[2],qr_ela[id2,qr_col[1]], qr_ela[id2,qr_col[2]])
    pax[row_,]
  }}
  return(a)}



#'Calc mist pax
#'@description supporting function for ELpruning
#' @param rat A data.frame 
#' @param ttien A matrix-like object convertible to a numeric matrix.
#' @param tssen A numeric vector.
#' @param threads Integer. If greater than 1, parallel execution will be run.
#'
#' @return A data frame
#' @noRd
calc_mist_pax <- function(rat, ttien, tssen, threads){
  pax <- rat
  npax <- max(pax['V2'])
  ttien.m <- as.matrix(ttien)
  ### requires starting clusters out side the function ###
  if(threads>1){a <- foreach(i = pax[,'V1'], j = pax[,'V2'], .combine="rbind", .packages=c("tidyverse","ELAplus")) %dopar% {
    row_ <- as.character(npax*(i-1) + j)
    te <- ttien.m[[j]][i]
    se <- c(tssen[i], tssen[j])
    pax[row_, 'val1'] <- min(te - se[1], te - se[2])
    val2_pre <- pax[row_, c('V1', 'V2')] %>% T %>% cbind(se)
    val2_pre <- val2_pre[order(val2_pre$se)][,row_]
    pax[row_, 'val2_1'] <- val2_pre[1]
    pax[row_, 'val2_2'] <- val2_pre[2]
    pax[row_, 'mist1'] <- te - se[1]
    pax[row_, 'mist2'] <- te - se[2]
    pax[row_,]
    }}else{a <- foreach(i = pax[,'V1'], j = pax[,'V2'], .combine="rbind") %do% {
    row_ <- as.character(npax*(i-1) + j)
    te <- ttien.m[[j]][i]
    se <- c(tssen[i], tssen[j])
    pax[row_, 'val1'] <- min(te - se[1], te - se[2])
    val2_pre <- pax[row_, c('V1', 'V2')] %>% T %>% cbind(se)
    val2_pre <- val2_pre[order(val2_pre$se)][,row_]
    pax[row_, 'val2_1'] <- val2_pre[1]
    pax[row_, 'val2_2'] <- val2_pre[2]
    pax[row_, 'mist1'] <- te - se[1]
    pax[row_, 'mist2'] <- te - se[2]
    pax[row_,]
    }}
  return(a)}


#'Calc_deltaE_range
#'@description supporting function for ELpruning with absolute value
#' @param tss A string vector (stable states) 
#' @param rat A data.frame
#' @param tti A matix (tipping points).
#' @param bootstrap_ene A data.frame bootstrap results of estimated energies
#' @param qrange A numeric value for quantile range (confidence interval).
#' @param tssen A numeric vector (energy of stable states).
#' @param ttien A matrix-like object convertible to a numeric matrix.
#' @param threads Integer. If greater than 1, parallel execution will be run.
#' 
#' @return A data frame
#' @importFrom stats quantile
#' @noRd
calc_deltaE_range <- function(tss, rat, tssen, tti, ttien, bootstrap_ene, qrange, threads){
  
  pax <- rat
  npax <- max(pax['V2'])
  tti.m <- as.matrix(tti)
  ttien.m <- as.matrix(ttien)
  ### requires starting clusters out side the function ###
  if(threads>1){a <- foreach(i = pax[,'V1'], j = pax[,'V2'], .combine="rbind", .packages=c("tidyverse","ELAplus")) %dopar% {
    row_ <- as.character(npax*(i-1) + j)
    tipp_id <- tti.m[[j]][i]
    te <- ttien.m[[j]][i]
    se <- c(tssen[i], tssen[j])
    id1 <- tss[[i]]
    id2 <- tss[[j]]
    # check overlap in the quantile range
    pax[row_, 'val1'] <- min(te - se[1], te - se[2])
    val2_pre <- pax[row_, c('V1', 'V2')] %>% T %>% cbind(se)
    val2_pre <- val2_pre[order(val2_pre$se)][,row_]
    pax[row_, 'val2_1'] <- val2_pre[1]
    pax[row_, 'val2_2'] <- val2_pre[2]
    deltaE1 <- bootstrap_ene[tipp_id,] - bootstrap_ene[id1,]
    deltaE2 <- bootstrap_ene[tipp_id,] - bootstrap_ene[id2,]
    #pax[row_, 'overlap1'] <- as.numeric(quantile(deltaE1,probs=qrange,na.rm=TRUE)) < 0
    pax[row_, 'pvalue1'] <- length(deltaE1[deltaE1 < 0])/length(deltaE1)
    pax[row_, 'pvalue2'] <- length(deltaE1[deltaE2 < 0])/length(deltaE2)
    pax[row_,]
  }}else{a <- foreach(i = pax[,'V1'], j = pax[,'V2'], .combine="rbind") %do% {
    row_ <- as.character(npax*(i-1) + j)
    tipp_id <- tti.m[[j]][i]
    te <- ttien.m[[j]][i]
    se <- c(tssen[i], tssen[j])
    id1 <- tss[[i]]
    id2 <- tss[[j]]
    # check overlap in the quantile range
    pax[row_, 'val1'] <- min(te - se[1], te - se[2])
    val2_pre <- pax[row_, c('V1', 'V2')] %>% T %>% cbind(se)
    val2_pre <- val2_pre[order(val2_pre$se)][,row_]
    pax[row_, 'val2_1'] <- val2_pre[1]
    pax[row_, 'val2_2'] <- val2_pre[2]
    deltaE1 <- bootstrap_ene[tipp_id,] - bootstrap_ene[id1,]
    deltaE2 <- bootstrap_ene[tipp_id,] - bootstrap_ene[id2,]
    pax[row_, 'pvalue1'] <- length(deltaE1[deltaE1 < 0])/length(deltaE1)
    pax[row_, 'pvalue2'] <- length(deltaE1[deltaE2 < 0])/length(deltaE2)
    pax[row_,]
  }}
  return(a)}


#'Replace value
#'@description supporting function for ELpruning
#' @param value_before A value to be replaced.
#' @param list_before A vector (or list) to be copied and modified.
#' @param value_after A value to replace with.
#' @return A modified copy of input with replacements applied.
#' @noRd
replace_value <- function(value_before, list_before, value_after){
  list_after <- list_before
  for(i in 1: length(list_after)){
    if(list_after[i] == value_before){list_after[i] <- value_after}
  }
  return(list_after)
}

#' SteepestDescent
#' @description Perform steepest descent on the energy landscape.
#'
#' Starting from an initial binary state, the steepest descent algorithm 
#' will find a local minimum of the energy landscape. 
#' The computation is delegated to the C++ implementation
#' \code{SteepestDescent_cpp}.
#'
#' @param state Binary vector representing the initial state.
#' @param alpha vector of implicit/explicit environmental effects.
#' @param beta Matrix of species interactions.
#'
#' @return A numeric vector of the reached binary stable state and its energy.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
SteepestDescent <- function(state, alpha, beta){
	res <- SteepestDescent_cpp(state, alpha, beta)
	return(res)
}

#' Bi
#' @description Identify basin (stable state) and its energy for a given state.
#'
#' Applies \link{SteepestDescent} to the input state and returns the basin (stable state) reached,
#' together with its energy. The stable state is encoded as an integer ID using \link{bin2id}.
#'
#' @param xi Binary vector representing the initial state.
#' @param h vector of implicit/explicit environmental effects on species.
#' @param j Matrix of species interactions.
#'
#' @return A list with two elements:
#'   (1) integer stable state ID (basin assignment),
#'   (2) energy of the corresponding stable state.
#' @useDynLib ELAplus, .registration=TRUE
#' @export
Bi <- function(xi, h, j){
  sd <- SteepestDescent(xi,h,j)
  stateyen <- list(as.vector(sd[-length(sd)]), sd[length(sd)])
  state <- stateyen[[1]]
  energy <- stateyen[[2]]
  return(
    list(bin2id(state),
      energy))}


#' Energy
#' @description Calculate energy of a given state.
#'
#' Computing energy in \code{state} using given parameters,\code{alpha} and \code{beta}
#' based on the (extended) maximum pairwise entropy model.
#'  
#' @param state Binary vector representing a state (community composition).
#' @param alpha vector of implicit/explicit environmental effects on species.
#' @param beta Matrix of species interactions.
#'
#' @return A numeric scalar giving the energy of the state.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
Energy <- function(state, alpha, beta){
	res <- cEnergy(state, alpha, beta)
	return(res)
}


#' GradELA
#' @description Estimate energy landscape across environmental gradients.
#'
#' Runs \link{ELA} and \link{ELPruning} comprehensively across specified environmental parameters.
#' One environmental parameter can be changed by specifying \code{eid} over \code{steps} within \code{range}. 
#' If \code{env} is not given, the mean vector of \code{enmat} is used as the reference environmental condition.
#'
#' @param sa output of runSA function.
#' @param eid target environmental factor: the specified environmental factor to be changed.
#' @param enmat array/matrix of environmental factors for each sample (normalization required).
#' @param pruning_type the type of pruning \code{"relative"}(use with \code{th}) or \code{"bootstrap"} (use with \code{lower_qr},\code{ocmat} and \code{bs_params}) 
#' @param env array of environmental factor values. 
#' For the environmental parameters that are not specified for \code{eid}, these values are used as references.
#' If not specified, then the mean of values in \code{enmat} are used as references.
#' @param range range over which the environmental factor specified by eid is changed.
#' @param steps number of steps in the range.
#' @param th ratio of the depth of the basin to be pruned with respect to the deepest basin.
#' @param threads number of cores (threads) used for calculation.
#' @param SS.itr number of iterations to identify stable states.
#' @param FindingTip.itr number of iterations to identify tipping points.
#' @param bs_params output of \link{bootstrap_SA}. list of estimated parameters using bootstrap resampled data set.
#' necessary when \code{type="bootstrap"}.
#' @param ocmat community composition matrix.
#' @param lower_qr Numeric value. Quantile used for constructing confidence intervals from bootstrap samples to check 
#' whether the estimated energy gaps include zero. Default is \code{lower_qr = 0.05},necessary when \code{type="bootstrap"}.
#' @param threads number of cores (threads) used for calculation.
#'
#' @return A list of three elements: (1) ELA results for each step (pruned),
#'   (2) the sequence of environmental values, and (3) the reference env vector.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
GradELA <- function(sa=sa, pruning_type="relative",eid=NULL, enmat=enmat, env=NULL, 
                    range=NULL, steps=16, 
                    th=0.05, threads=1, SS.itr=20000, FindingTip.itr=10000,
                    lower_qr=0.05,bs_params=NULL,ocmat=NULL){
  start <- proc.time()[3]
  
  if(threads>1){cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
  parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
  parallel::clusterSetRNGStream(cluster)
  doParallel::registerDoParallel(cluster)
  on.exit(parallel::stopCluster(cluster), add = TRUE)}
  
  if(is.null(eid)){return(cat("eid not specified prease set eid=position/colname\n"))}
  if(is.character(eid)){ei <- which(colnames(enmat)==eid)}else{ei<-eid}
  if(is.null(env)){refe <- apply(enmat, 2, mean)
  cat("env not specified, the mean of enmat is used\n")
  }else{refe <- env}
  if(is.null(range)){mi <- min(enmat[,ei])}else{mi <- range[1]}
  if(is.null(range)){ma <- max(enmat[,ei])}else{ma <- range[2]}
  dm <- (ma - mi)/(steps - 1)
  de <- seq(mi, ma, dm)
  cat("processing(")
  cat(length(de))
  cat(") |")
  els <- foreach(i=de) %dopar% {
      cat("=")
      ee <- replace(refe,ei,i)
      elanp <- ELA(sa, ee, SS.itr=SS.itr, FindingTip.itr=FindingTip.itr, threads=1, reporting=FALSE, fork=FALSE)
      if(pruning_type=="relative"){
        elap <- ELPruning(elanp, type = "relative", th=th, threads=1, reporting=FALSE, fork=FALSE)
      }
      else if(pruning_type=="bootstrap"){
        bootstrap <- bootstrap_ELA(bs_params,ocmat,ela=elanp,enmat=enmat,env=ee,threads=1,reporting=FALSE)
        bootstrap_ene <- bootstrap[[2]]
        elap <- ELPruning(elanp, type = "bootstrap",bootstrap_ene=bootstrap_ene,threads=1,
                          lower_qr=lower_qr,reporting = FALSE,fork=FALSE)
      }
      return(elap)
  }
  cat("|\n")
  end <- proc.time()[3]
  refe[eid] <- NA
  cat(sprintf("Elapsed time %.2f sec\n", end - start))
  return(list(els, de, refe))
}


#' SSentropy
#' @description Estimate stable state entropy for a given state.
#'
#' Evaluating the entropy of transitions among stable states
#' starting from a given community state. 
#'
#' The returned value includes the estimated entropy, the convergence time,
#' and the number of successful convergences within the given iteration limits.
#'
#' @param state Binary vector (or matrix with one row) representing the current state.
#' @param ss Matrix of stable states.
#' @param alpha Vector of implicit or explicit environmental effects.
#' @param beta Matrix of species interactions.
#' @param seitr Number of stochastic search iterations.
#' @param convTime Maximum number of iterations for each search.
#'
#' @return A numeric matrix with three columns:
#'   (1) stable state entropy,
#'   (2) convergence time,
#'   (3) number of converged searches.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
SSentropy <- function(state, ss, alpha, beta, seitr=1000, convTime=10000){ 
  res <- SSentropy_cpp(uoc= state, ss= ss, alpha= alpha, beta=beta, 
                       seitr=seitr, convTime=convTime) 
  return(res) 
  }


#' stability
#' @description Calculate energy gap and stable state entropy as stability indices and returns a dataframe.
#'
#' For each sample state, this function identifies the basin (stable state) by \link{Bi},
#' then computes (i) the energy gap and (ii) stable state entropy
#' estimated by \link{SSentropy}. 
#' If \code{enmat} is provided and \code{sa} was estimated with environmental
#' factors, the basin assignment is computed for each sample-specific \code{env}. 
#'
#' Note: the returned stable state id corresponds to the basin (stable state) the sample is assigned to.
#'
#' @param sa : output of runSA function.
#' @param ocmat : array of the vectors of presence/absence status for each sample. 
#' @param enmat : array of the vectors of environmental factors for each sample. Normalization required.
#' @param threads : number of cores (threads) used for calculation.
#' @param seitr : Number of stochastic search iterations.
#' @param seconvTime : Maximum number of iterations for each search.
#' @param reporting : enable/disable the reporting function. 
#' @return A data frame with stability indices. Columns include at least
#'   \code{energy.gap}, \code{ss.entropy}, \code{e.realize}, \code{e.stable}, \code{state.id}, and \code{stable.state.id}
#'   (and may include \code{conv_time} and \code{convergence} depending on the branch).
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export

Stability <- function(sa, ocmat, enmat=NULL, seitr=1000, seconvTime=10000, threads=1, reporting=TRUE){
  if(reporting){cat('Start Stability:\n')}
  start <- proc.time()[3]
  sadim <- dim(sa[[1]])
  if(reporting && sadim[2] > sadim[1]+1 && is.null(enmat)){cat("sa obtained with enmat, but enmat is not given.\nassigned a zero vector as the environmental condition for all data points.\n")}
  if(reporting && sadim[2] == sadim[1]+1 && !is.null(enmat)){cat("sa obtained without enmat, but enmat is given. enmat is omitted.\n")}
  
  no <- ncol(ocmat)
  if(threads>1){cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
  parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
  parallel::clusterSetRNGStream(cluster)
  doParallel::registerDoParallel(cluster)
  on.exit(parallel::stopCluster(cluster), add = TRUE)}
  
  if(!is.null(enmat) && sadim[2] > sadim[1]+1){
      ne <- ncol(enmat)
      
      if(threads>1){stab <- foreach(x=t(cbind(as.data.frame(ocmat), enmat)), .combine="rbind", .packages=c("ELAplus", "tidyverse", "foreach")) %dopar%{
        env <- x[(no+1):(no+ne)]
        sa2p <- sa2params(sa, env)
        je <- sa2p[[2]]
        hge <- sa2p[[4]]
        state <- bin2id(x[1:no]) 
        # basin #
        bi <- Bi(x[1:no], hge, je)
        ssid <- bi[[1]]
        e.stable <- bi[[2]] 
        # energy gap #
        e.realize <- Energy(x[1:no], hge, je)
        energy.gap <- as.numeric(e.realize - e.stable)
        # ss entropy #
        minsets <- SSestimate(hge, je, itr = 10000)
        minsets <- unique(minsets[,-ncol(minsets)])
        sse <- SSentropy(
          as.matrix(t(as.data.frame(x[1:no]))),
          ss=minsets, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
        ss.entropy <- sse[,1]
        conv_time <- sse[,2]
        convergence <- sse[,3]
        cbind(as.data.frame(t(as.matrix(c(energy.gap, ss.entropy, conv_time, convergence, e.realize, e.stable)))),
              as.data.frame(t(as.matrix(c(state, ssid))))) %>% "colnames<-" (c("energy.gap", "ss.entropy","conv_time","convergence","e.realize","e.stable","state.id","stable.state.id"))
      }
      
      }else{
        stab <- foreach(x=t(cbind(as.data.frame(ocmat), enmat)), .combine="rbind") %do% {
        env <- x[(no+1):(no+ne)]
        sa2p <- sa2params(sa, env)
        je <- sa2p[[2]]
        hge <- sa2p[[4]]
        state <- bin2id(x[1:no]) 
        # basin #
        bi <- Bi(x[1:no], hge, je)
        ssid <- bi[[1]]
        e.stable <- bi[[2]] 
        # energy gap #
        e.realize <- Energy(x[1:no], hge, je)
        energy.gap <- as.numeric(e.realize - e.stable)
        # ss entropy #
        minsets <- SSestimate(hge, je, itr = 10000)
        minsets <- unique(minsets[,-ncol(minsets)])
        sse <- SSentropy(
            as.matrix(t(as.data.frame(x[1:no]))),
                         ss=minsets, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
        ss.entropy <- sse[,1]
        conv_time <- sse[,2]
        convergence <- sse[,3]
        cbind(as.data.frame(t(as.matrix(c(energy.gap, ss.entropy, conv_time, convergence, e.realize, e.stable)))),
              as.data.frame(t(as.matrix(c(state, ssid))))) %>% "colnames<-" (c("energy.gap", "ss.entropy","conv_time","convergence","e.realize","e.stable","state.id","stable.state.id"))
      }
      }
     
      # ss entropy check #
      if(reporting){cat(sprintf("Convergence rate: %.2f \n",mean(stab$convergence)/seitr))}
      if(mean(stab$convergence)/seitr < 0.9){cat("more than 10% searches were not converged. convTime may be too small, try larger seconvTime.\n")}
      
      rownames(stab) <- rownames(ocmat)
     
     }else{
      sa2p <- sa2params(sa)
      je <- sa2p[[2]]
      hge <- sa2p[[4]]
      sampleSS <- t(apply(ocmat, 1, SteepestDescent, alpha=hge, beta=je))
      minsets <- unique(sampleSS[,-ncol(sampleSS)])
      if(threads>1){
        sse <- foreach(i = 1:dim(ocmat)[1], .combine = 'rbind', .packages = c("ELAplus", "tidyverse", "foreach")) %dopar% {
          SSentropy(t(as.matrix(ocmat[i,])), ss=minsets, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
        }
      }
      else{
        sse <- SSentropy(ocmat, ss=minsets, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
      }
      if(reporting){cat(sprintf("Convergence rate: %.2f \n",mean(sse[,3])/seitr))}
      if(mean(sse[,3])/seitr < 0.9){cat("more than 10% searches were not converged. convTime may be too small, try larger convTime.\n")}
      ss.entropy <- sse[,1]
      names(ss.entropy) <- rownames(ocmat)
      e.stable <- sampleSS[, ncol(sampleSS)]
      e.realize <- apply(ocmat, 1, Energy, hge, je)
      energy.gap <- e.realize - e.stable
      state.id <- apply(ocmat, 1, bin2id)
      stable.state.id <- apply(sampleSS[, -ncol(sampleSS)], 1, bin2id)
      stab <- cbind(as.data.frame(cbind(energy.gap, ss.entropy, e.realize, e.stable)),
      as.data.frame(cbind(state.id, stable.state.id)))}
  end <- proc.time()[3]
  if(reporting==TRUE){cat(sprintf("Stability: elapsed time %.2f sec\n\n", end - start))}
  return(stab)}

#' gstability
#' @description Calculate energy gap, energy barrier and stable state entropy.
#'
#' This function evaluates stability indices for both (i) the pruned and (ii) the non-pruned
#' energy landscape. For each sample, it assigns the basin by \link{Bi} and computes:
#'  - energy gap (sample energy minus basin energy),
#'  - energy barrier (minimum tipping energy connected to the basin minus basin energy),
#'  - stable state entropy (\code{SSentropy}).
#'
#' For basin connectivity: when only one stable state exists, tipping energy is treated as \code{Inf}
#' and therefore energy barrier is returned as \code{Inf}.
#'
#' @param sa : output of runSA function.
#' @param ocmat : matrix of presence/absence status for each sample. 
#' @param enmat : matrix of environmental factors for each sample. Normalization required.
#' @param th : Numerical (0~1). Ratio of the depth of the basin to be pruned with respect to the deepest basin in each pair.
#' @param pruning_type the type of pruning \code{"relative"}(use with \code{th}) or \code{"bootstrap"} (use with \code{lower_qr} and \code{bootstrap_ene}) 
#' @param lower_qr Numeric value. Quantile used for constructing confidence intervals from bootstrap samples to check 
#' whether the estimated energy gaps include zero. Default is \code{lower_qr = 0.05},necessary when \code{type="bootstrap"}.
#' @param bs_params output of \link{bootstrap_SA}. list of estimated parameters using bootstrap resampled data set,
#' necessary when \code{type="bootstrap"}.
#' @param threads number of cores (threads) used for calculation.#' @param threads : number of cores (threads) used for calculation.
#' @param seitr : Number of stochastic search iterations.
#' @param seconvTime : Maximum number of iterations for each search.
#' @param reporting : enable/disable the reporting function. 
#' 
#' @return A list of four elements: (1) pruned stability data frame, (2) non-pruned stability data frame,
#'   (3) auxiliary object for pruned landscape (parameters + stable states table), and
#'   (4) auxiliary object for sample-wise evaluation (inputs and ELA objects).
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
gStability <- function(sa, ocmat, enmat=NULL, seitr=1000, seconvTime=10000, pruning_type="relative",
                       th=0.1,lower_qr=0.05,threads=1, reporting=TRUE,bs_params=NULL){
  if(reporting){cat('Start gStability:\n')}
  start <- proc.time()[3]
  sadim <- dim(sa[[1]])
  if(reporting && sadim[2] > sadim[1]+1 && is.null(enmat)){cat("sa obtained with enmat, but enmat is not given.\nassigned a zero vector as the environmental condition for all data points.\n")}
  if(reporting && sadim[2] == sadim[1]+1 && !is.null(enmat)){cat("sa obtained without enmat, but enmat is given. enmat is omitted.\n")}
  cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
  parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
  parallel::clusterSetRNGStream(cluster)
  doParallel::registerDoParallel(cluster)
  on.exit(parallel::stopCluster(cluster), add = TRUE)
  if(!is.null(enmat) && sadim[2] > sadim[1]+1){
    no <- ncol(ocmat)
    mo <- nrow(ocmat)
    ne <- ncol(enmat)
    ### ela
    elall <- foreach(x=t(cbind(as.data.frame(ocmat), enmat)), .combine="rbind", .packages=c("ELAplus", "tidyverse", "foreach", "gtools")) %dopar%{
        env <- x[(no+1):(no+ne)]
        elanp <- ELA(sa, env=env, threads=1, reporting=FALSE, fork=FALSE)
        if(pruning_type=="relative"){
          elap <- ELPruning(elanp, type = "relative", th=th, threads=1, reporting=FALSE, fork=FALSE)
        }
        else if(pruning_type=="bootstrap"){
          if(is.null(bs_params)){stop("'bs_params' is missing. Please get bootstrap SA results by running 'bootstrap_SA()'.")}
          bootstrap <- bootstrap_ELA(bs_params,ocmat,ela=elanp,enmat=enmat,env=env,threads=1,reporting=FALSE)
          bootstrap_ene <- bootstrap[[2]]
          elap <- ELPruning(elanp, type = "bootstrap",bootstrap_ene=bootstrap_ene,threads=1,
                            lower_qr=lower_qr,reporting = FALSE,fork=FALSE)
        }
        #ela <- ELPruning(elanp, th=th, threads=1, reporting=FALSE, fork=FALSE)
        list(elanp, elap)}
    #print("Finish pruning.")
    ### stability evaluation (1)
    stab <- foreach(x=t(cbind(as.data.frame(ocmat), enmat)), ela=elall[,2], elanp=elall[,1], .combine="rbind") %do% {
        env <- x[(no+1):(no+ne)]
        sa2p <- sa2params(sa, env)
        je <- sa2p[[2]]
        hge <- sa2p[[4]]
        state <- bin2id(x[1:no])
        #####
        stablestates <- ela[[1]][[1]]
        stablen <- ela[[1]][[2]]
        tippingpoints <- ela[[1]][[3]]
        tippingen <- ela[[1]][[4]]
        stablestates.np <- elanp[[1]]
        stablen.np <- elanp[[2]]
        tippingpoints.np <- elanp[[3]]
        tippingen.np <- elanp[[4]]
        # basin #
        bi <- Bi(x[1:no], hge, je)
        ssid.np <- bi[[1]]
        ssid <- ela[[2]][,2][ela[[2]][,1]==bi[[1]]]
        e.stable <- stablen[stablestates==ssid]
        e.stable.np <- bi[[2]]
        # energy gap #
        e.realize <- Energy(x[1:no], hge, je)
        energy.gap <- as.numeric(e.realize - e.stable)
        energy.gap.np <- as.numeric(e.realize - e.stable.np)
        # energy barrier #
        if(length(stablestates) > 1){
            e.tipping <- min(unlist(c(as.data.frame(tippingen)[stablestates==ssid,],as.data.frame(tippingen)[,stablestates==ssid])))
            energy.barrier <- e.tipping - e.stable}else{
            e.tipping <- Inf
            energy.barrier <- Inf}
        if(length(stablestates.np) > 1){
            e.tipping.np <- min(unlist(c(as.data.frame(tippingen.np)[stablestates.np==ssid.np,],as.data.frame(tippingen.np)[,stablestates.np==ssid.np])))
            energy.barrier.np <- e.tipping.np - e.stable.np}else{
            e.tipping.np <- Inf
            energy.barrier.np <- Inf}
        # ss entropy #
        minsets <- as.matrix(t(sapply(stablestates, function(x){id2bin(x, no)})))
        sse <- SSentropy(as.matrix(t(as.data.frame(x[1:no]))), ss=minsets, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
        ss.entropy <- sse[,1]
        conv_time <- sse[,2]
        convergence <- sse[,3]
        minsets.np <- as.matrix(t(sapply(stablestates.np, function(x){ELAplus::id2bin(x, no)})))
        sse.np <- SSentropy(as.matrix(t(as.data.frame(x[1:no]))), ss=minsets.np, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
        ss.entropy.np <- sse.np[,1]
        conv_time.np <- sse.np[,2]
        convergence.np <- sse.np[,3]
        cbind(as.data.frame(t(as.matrix(c(energy.gap, ss.entropy, energy.barrier, e.realize, e.stable, e.tipping, conv_time, convergence)))),
              as.data.frame(t(as.matrix(c(state, ssid)))),
              as.data.frame(t(as.matrix(c(energy.gap.np, ss.entropy.np, energy.barrier.np, e.realize, e.stable.np, e.tipping.np, conv_time.np, convergence.np)))),
              as.data.frame(t(as.matrix(c(state, ssid.np))))) %>% "colnames<-" (
                  c("energy.gap", "ss.entropy", "energy.barrier", "e.realize", "e.stable", "e.tipping",
                    "conv_time","convergence","state.id", "stable.state.id",
                    "energy.gap.np", "ss.entropy.np", "energy.barrier.np", "e.realize", "e.stable.np", "e.tipping.np", 
                    "conv_time.np","convergence.np","state.id.np", "stable.state.id.np"))}
     rownames(stab) <- rownames(ocmat)
     if(reporting){cat(sprintf("Convergence rate: %.2f \n",mean(stab$convergence)/seitr))}
     if(reporting){cat(sprintf("Convergence rate (no pruning): %.2f \n",mean(stab$convergence.np)/seitr))}
     if(mean(stab$convergence)/seitr < 0.9){cat("more than 10% searches were not converged. convTime may be too small, try larger seconvTime.\n")}
     if(mean(stab$convergence.np)/seitr < 0.9){cat("more than 10% searches were not converged(no pruning). convTime may be too small, try larger seconvTime.\n")}

     ### Stability evaluation (2)
     stab2 <- foreach(x=t(cbind(as.data.frame(ocmat), enmat)),  ela=elall[,2], elanp=elall[,1]) %do% {
        env <- x[(no+1):(no+ne)]
        sa2p <- sa2params(sa, env)
        stablestates <- ela[[1]][[1]]
        stablen <- ela[[1]][[2]]
        bin <- as.list(lapply(stablestates, function(x){id2bin(x, ncol(ocmat))}))
        sstable <- as.data.frame(cbind(stablestates,
                                       stablen, t(as.data.frame(bin)))) %>% 'colnames<-'(c('ID', 'Energy', colnames(ocmat))) %>% 'rownames<-'(1: length(stablestates))
        list(list(sa2p, sstable), list(ocmat, env, sa, ela, elanp))}
     stab2 <- list(lapply(stab2, function(x) x[[1]]), lapply(stab2, function(x) x[[2]]))
     }else{
      no <- ncol(ocmat)
      sa2p <- sa2params(sa)
      je <- sa2p[[2]]
      hge <- sa2p[[4]]
      ### ela
      elanp <- ELA(sa, env=NULL, threads=threads, reporting=FALSE, fork=FALSE)
      if(pruning_type=="relative"){
        ela <- ELPruning(elanp, type = "relative", th=th, threads=threads, reporting=FALSE, fork=FALSE)
      }
      else if(pruning_type=="bootstrap"){
        bootstrap <- bootstrap_ELA(bs_params,ocmat,ela=elanp,enmat=enmat,env=NULL,threads=1,reporting=FALSE)
        bootstrap_ene <- bootstrap[[2]]
        ela <- ELPruning(elanp, type = "bootstrap",bootstrap_ene=bootstrap_ene,threads=1,
                          lower_qr=lower_qr,reporting = FALSE,fork=FALSE)
      }
      
      ### Stability evaluation (1)
      stablestates <- ela[[1]][[1]]
      stablen <- ela[[1]][[2]]
      tippingpoints <- ela[[1]][[3]]
      tippingen <- ela[[1]][[4]]
      stablestates.np <- elanp[[1]]
      stablen.np <- elanp[[2]]
      tippingpoints.np <- elanp[[3]]
      tippingen.np <- elanp[[4]]
      stab <- foreach(x=t(as.data.frame(ocmat)), .combine="rbind", .packages=c("ELAplus", "tidyverse", "foreach")) %dopar% {
        state <- bin2id(x[1:no])
        # basin #
        bi <- Bi(x[1:no], hge, je)
        ssid.np <- bi[[1]]
        ssid <- ela[[2]][,2][ela[[2]][,1]==bi[[1]]]
        e.stable <- stablen[stablestates==ssid]
        e.stable.np <- bi[[2]]
        # energy gap #
        e.realize <- Energy(x[1:no], hge, je)
        energy.gap <- as.numeric(e.realize - e.stable)
        energy.gap.np <- as.numeric(e.realize - e.stable.np)
        # energy barrier #
        if(length(stablestates) > 1){
            e.tipping <- min(unlist(c(as.data.frame(tippingen)[stablestates==ssid,],as.data.frame(tippingen)[,stablestates==ssid])))
            energy.barrier <- e.tipping - e.stable}else{
            e.tipping <- Inf
            energy.barrier <- Inf}
        if(length(stablestates.np) > 1){
            e.tipping.np <- min(unlist(c(as.data.frame(tippingen.np)[stablestates.np==ssid.np,],as.data.frame(tippingen.np)[,stablestates.np==ssid.np])))
            energy.barrier.np <- e.tipping.np - e.stable.np}else{
            e.tipping.np <- Inf
            energy.barrier.np <- Inf}
        # ss entropy #
        minsets <- as.matrix(t(sapply(stablestates, function(x){id2bin(x, no)})))
        sse <- SSentropy(as.matrix(t(as.data.frame(x[1:no]))), ss=minsets, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
        ss.entropy <- sse[,1]
        conv_time <- sse[,2]
        convergence <- sse[,3]
        minsets.np <- as.matrix(t(sapply(stablestates.np, function(x){ELAplus::id2bin(x, no)})))
        sse.np <- SSentropy(as.matrix(t(as.data.frame(x[1:no]))), ss=minsets.np, alpha= hge, beta=je, seitr=seitr, convTime=seconvTime)
        ss.entropy.np <- sse.np[,1]
        conv_time.np <- sse.np[,2]
        convergence.np <- sse.np[,3]
        data <- cbind(as.data.frame(t(as.matrix(c(energy.gap, ss.entropy, energy.barrier, e.realize, e.stable, e.tipping, conv_time, convergence)))),
              as.data.frame(t(as.matrix(c(state, ssid)))),
              as.data.frame(t(as.matrix(c(energy.gap.np, ss.entropy.np, energy.barrier.np, e.realize, e.stable.np, e.tipping.np, conv_time.np, convergence.np)))),
              as.data.frame(t(as.matrix(c(state, ssid.np))))) %>% "colnames<-" (
                c("energy.gap", "ss.entropy", "energy.barrier", "e.realize", "e.stable", "e.tipping",
                  "conv_time","convergence","state.id", "stable.state.id",
                  "energy.gap.np", "ss.entropy.np", "energy.barrier.np", "e.realize", "e.stable.np", "e.tipping.np", 
                  "conv_time.np","convergence.np","state.id.np", "stable.state.id.np"))}
      rownames(stab) <- rownames(ocmat)
      if(reporting){cat(sprintf("Convergence rate: %.2f \n",mean(stab$convergence)/seitr))}
      if(reporting){cat(sprintf("Convergence rate (no pruning): %.2f \n",mean(stab$convergence.np)/seitr))}
      if(mean(stab$convergence)/seitr < 0.9){cat("more than 10% searches were not converged. convTime may be too small, try larger seconvTime.\n")}
      if(mean(stab$convergence.np)/seitr < 0.9){cat("more than 10% searches were not converged(no pruning). convTime may be too small, try larger seconvTime.\n")}

      ### Stability evaluation (2)
      bin <- as.list(lapply(stablestates, function(x){id2bin(x, ncol(ocmat))}))
      sstable <- as.data.frame(cbind(stablestates,
                                      stablen, t(as.data.frame(bin)))) %>% 'colnames<-'(c('ID', 'Energy', colnames(ocmat))) %>% 'rownames<-'(1: length(stablestates))
      stab2 <- list(list(sa2p, sstable), list(ocmat, NULL, sa, ela, elanp))
     }
  end <- proc.time()[3]
  if(reporting==TRUE){cat(sprintf("gStability: elapsed time %.2f sec\n\n", end - start))}
  return(list(as.data.frame(stab)[1:8], as.data.frame(stab)[9:16], stab2[[1]], stab2[[2]]))
  }


#' sstable
#' @description Summary of stable states.
#'
#' Builds a table of stable states and their energies from an ELA output.
#' The returned data frame contains stable state IDs and the corresponding binary
#' presence/absence vector (per species). This function is typically used for
#' inspecting stable states after pruning (if \code{ela} is pruned) or before pruning.
#'
#' @param ela output of ELA or the first return value of ELPruning.
#' @param ocmat community composition matrix.
#'
#' @return A data frame with columns \code{ID}, \code{Energy}.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
sstable <- function(ela,ocmat){
  stablestates  <- ela[[1]]
  stablen <- ela[[2]]
  tippingpoints <- ela[[3]]
  tippingen <- ela[[4]]
  bin <- as.list(lapply(stablestates, function(x){id2bin(x, ncol(ocmat))}))
  names(bin) <- stablestates
  sstable <- as.data.frame(cbind(stablestates, stablen, t(as.data.frame(bin)))) %>%
    'colnames<-'(c('ID', 'Energy', colnames(ocmat))) %>%
    'rownames<-'(1: length(stablestates))
  return(sstable)}

#' tptable
#' @description Summary of tipping points.
#'
#' Builds a table of tipping points (transition states) and their energies from an ELA output.
#' This function lists only finite tipping points
#' (i.e., entries not equal to \code{Inf}), and returns the corresponding binary state vectors.
#'
#' Two stable states are required at least.
#'
#' @param ela  output of \link{ELA} or the first return value of \link{ELPruning}.
#' @param ocmat community composition matrix.
#'
#' @return A data frame with columns \code{TP}, \code{SS1}, \code{SS2}, \code{Energy}, 
#' and one column per species (binary state vector).
#' If no tipping points exist, returns NULL (invisibly) after printing a message.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
tptable <- function(ela,ocmat){
  stablestates  <- ela[[1]]
  stablen <- ela[[2]]
  tippingpoints <- ela[[3]]
  tippingen <- ela[[4]]
  tpid.df <- as.data.frame(tippingpoints)
  tpid <- unlist(tpid.df[upper.tri(tpid.df, diag = FALSE)], use.names = FALSE)
  tpen.df <- as.data.frame(tippingen)
  tpen <- unlist(tpen.df[upper.tri(tpen.df, diag = FALSE)], use.names = FALSE)
  tpbin <- as.list(lapply(tpid, function(x){id2bin(x, ncol(ocmat))}))
  if(length(tpbin) == 0){
    cat(sprintf('only one stable state is found.\nNo tipping points exist.\n'))
    return()
  }
  tp.ss1 <- stablestates[unlist(lapply(seq_along(tippingpoints), function(i) {
      non_inf_count <- sum(tippingpoints[[i]] != "Inf")
      return(rep(i, non_inf_count))}))]
  tp.ss2 <-stablestates[unlist(lapply(tippingpoints, function(x) which(x != "Inf")))]

  tptable <- as.data.frame(cbind(tpid, tp.ss1, tp.ss2, tpen, t(as.data.frame(tpbin)))) %>%
    'colnames<-'(c('TP','SS1','SS2', 'Energy', colnames(ocmat))) %>%
    'rownames<-'(1: length(tpid))
  return(tptable)
}

####################################
####################################

#'Calculation stability indices.
#'@description Calculating stable state energy, difference between sample's energy and stable state energy and stable state entropy.
#'
#'@param data data is a matrix.
#'@param alpha alpha is a vector which explicit/implicit preference. Use output of runSA function.
#'@param J J is a matrix which shows species preference. Use output of runSA function.
#'@param prun Numeric 0 ~ 1. ratio of the depth of the basin to be pruned with respect to the deepest basin.
#'. If 0, then pruning will be skipped (default = 0).
#'@param seitr serials for searching stable states
#'@param pruning_type the type of pruning \code{"relative"}(use with \code{th}) or \code{"bootstrap"} (use with \code{lower_qr} and \code{bootstrap_ene}) 
#' @param lower_qr Numeric value. Quantile used for constructing confidence intervals from bootstrap samples to check 
#' whether the estimated energy gaps include zero. Default is \code{lower_qr = 0.05},necessary when \code{type="bootstrap"}.
#'@param bootstrap_ene a data frame of bootstrap energies generated by \link{bootstrap_ELA}. Necessary when \code{type="bootstrap"}.
#'@param threads the number of cores (threads) for parallel computing
#'@param convTime total iterations for searching stable points
#'@param threads the number of cores used for parallel computing.
#'@useDynLib ELAplus, .registration=TRUE
#'@noRd
calcStability <- function(data=NULL, alpha, J,
						  prun=0,threads=1,pruning_type="relative",
						  lower_qr=0.05,bs_params=NULL,seitr=1000, convTime=10000){
						  	
    start <- proc.time()[3]
    sampleSS <- t(apply(data, 1, SteepestDescent_cpp, alpha=alpha, beta=J))
   	
   	uniqueSS <- data.frame(unique(sampleSS[,-ncol(sampleSS)]))
	uniqueSS$energy <- NA
	
	for(i in 1:nrow(uniqueSS)){
	    
	    state <- uniqueSS[i,-ncol(uniqueSS)]
	    detect <- apply(sampleSS[,-ncol(sampleSS)], 1, function(x){ all(x==state) })
	    
	    uniqueSS[i, ncol(uniqueSS)] <- mean(sampleSS[detect, ncol(sampleSS)])
	}
	sampleSS <- uniqueSS[order(uniqueSS$energy),]
	
    ## ================================ ##
	if(prun>0 || !is.null(bs_params)){
		## |||||||||||||||||||||||||||| ##
		## pruned landscape
	    sa_dummy = list(cbind(alpha, J),NULL)
	    elasummary <- ELA(sa = sa_dummy, threads = threads,FindingTip.itr = 10000,
	                      SS.itr = 10000)
		#redela <- ELPruning(elasummary, th=prun)
		if(pruning_type=="relative"){
		  redela <- ELPruning(elasummary, type="relative",th=prun, reporting=FALSE, fork=FALSE)
		}
	  else if(pruning_type=="bootstrap"){
	    print("Pruning starts")
	    bootstrap <- bootstrap_ELA(bs_params,data,ela=elasummary,
	                               enmat=NULL,env=NULL,threads=1,reporting=FALSE)
	    print("bootstrap finished")
	    bootstrap_ene <- bootstrap[[2]]
	    redela <- ELPruning(elasummary, type = "bootstrap",bootstrap_ene=bootstrap_ene,threads=1,
	                     lower_qr=lower_qr,reporting = FALSE,fork=FALSE)
	  }
	 print("Pruning Finished")
		stateInfo <- elasummary[[1]]
		ssInfo <- stateInfo[stateInfo$state=='Stable state',]
		
		## |||||||||||||||||||||||||||| ##
		## -- Convert to deeper landscape
		
		sampleSSpruned <- as.matrix(sampleSS)
		for(i in 1:nrow(ssInfo)){
		    
		    state <- ssInfo[i,-c(1, 2, ncol(ssInfo))]
		    detect <- apply(sampleSS[,-ncol(sampleSS)], 1, function(x){ all(x==state) })
		    
		    if(sum(detect)>1){
		        id <- ssInfo[i,1]
		        
		        if( !any(redela$log[,2]==id)){
		            log <- redela$log
		            change <- log[id,2]
		            
		            sampleSSpruned[detect,] <- matrix( unlist(ssInfo[change, -c(1,2)]), 
		            								   ncol=ncol(sampleSSpruned), nrow=sum(detect), 
		            								   byrow=TRUE)
		        }
		        
		    }
		
		}
		sampleSS <- sampleSSpruned
        ## |||||||||||||||||||||||||||| ##
	}
    
    stablestate=t(apply(sampleSS, 1, 
                    function(x){ 
                        sep <- round(length(x)/20)
                        line <- c()
                        for(i in 1:sep){
                            y <- x[-length(x)]
                            bi <- stats::na.omit(y[((1+20*(i-1)):(20*i))])
                            line <- c(line, str2i( paste0(bi, collapse='')))
                            
                        }
                        
                        ssid=paste(line, collapse='-')
                        c(ssid, x[length(x)]) }))
                  
    ssenergy <- sampleSS[, ncol(sampleSS)]
    energy <- apply(data, 1, Energy, alpha= alpha, beta=J)
    
    energy.gap <- energy-ssenergy
    ssent <- SSentropy_cpp(uoc=data, ss=unique(sampleSS[,-ncol(sampleSS)]),
          				alpha= alpha, beta=J, 
          				seitr=seitr, convTime=convTime)
          	  
    return( data.frame(StableState.id=stablestate[,1], SSenergy=ssenergy,
                       energy.gap=energy.gap, SSentropy= ssent[,1]))
                       
    ## ================================ ##
    end <- proc.time()[3]
    cat(sprintf('Stability calculation done.\nElapsed time %.2f sec\n\n', end-start))                   
}


#' Energy landscape analysis with bootstrap resampled SA results
#' @description Running energy landscape analysis with bootstrap sampling of estimated saresults
#'
#' Performs \link{ELA} and \link{ELPruning} using SA fitting results based on 
#' the bootstrap resampled occurrence and environmental data using \link{bootstrap_SA}.
#' 
#' Basically, representative ELA results with original occurrence and environmental data should be given by \code{ela=ela},
#' alternatively the function can select a representative parameter where commonly appearing stable states are most frequent.
#'
#' The function computes bootstrap energy distributions for all stable states and tipping points
#' appearing in the specified pruned landscape. Energies that are not present in a given trial
#' are stored as \code{NaN} in the summary table.
#'
#' @param bs_params output of \link{bootstrap_SA}. list of estimated parameters using bootstrap resampled data set.
#' @param ocmat community composition matrix.
#' @param env  target environmental factor.
#' @param ela representative ELA result (output of \link{ELA}). If \code{NULL}, the function returns the most representative one.
#' @param enmat array of the vectors of environmental factors for each sample. Normalization required.
#' @param threads number of cores (threads) used for calculation.
#' @param reporting enable/disable the reporting function.
#'
#' @return A list with two elements:
#'   (1) representative pruned ELA result (stable states, energies, tipping points, tipping energies),
#'   (2) a data frame of bootstrap energies (rows = state IDs, columns = trials) with \code{NaN} for missing states.
#'
#' @useDynLib ELAplus, .registration=TRUE
#' @export
bootstrap_ELA <- function(bs_params,ocmat,ela=NULL,env=NULL,enmat=NULL,threads=1,reporting=FALSE){
  iterations <- length(bs_params)
  nspecies <- dim(ocmat)[[2]]
  if(reporting==TRUE){cat("Run bootstrap sampling of sa params.\n")}
  if(threads>1){cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
  parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
  parallel::clusterSetRNGStream(cluster)
  doParallel::registerDoParallel(cluster)
  on.exit(parallel::stopCluster(cluster), add = TRUE)}
  
  # summarize stable states after pruning
  if(is.null(ela)){
    if(threads>1){
    elaresults <- foreach(i = 1:iterations, 
                          .packages = c("ELAplus","doParallel","magrittr","gtools")) %dopar% {
                            rand_saparams <- list(bs_params[[i]],enmat)
                            # run ELA
                            elanp <- ELA(rand_saparams, env, threads=1, 
                                         FindingTip.itr = 5000, reporting=FALSE, fork=FALSE)
                            return(list(elanp,rand_saparams))
                          }
    }else{
      elaresults <- foreach(i = 1:iterations, 
                            .packages = c("ELAplus","doParallel","magrittr","gtools")) %do% {
                              rand_saparams <- list(bs_params[[i]],enmat)
                              # run ELA
                              elanp <- ELA(rand_saparams, env, threads=1, 
                                           FindingTip.itr = 5000, reporting=FALSE, fork=FALSE)
                              return(list(elanp,rand_saparams))
                            }
    }
    rand_ela_summary <- data.frame(matrix(NaN, nrow = 1, ncol = iterations))
    colnames(rand_ela_summary) <- paste0("trial", 1:iterations)
    rownames(rand_ela_summary) <- "0"
    default_row <- as.list(rep(NaN, ncol(rand_ela_summary)))
    for(ss in 1:length(elaresults)){
      ela_i <- elaresults[[ss]]
      sstates <- ela_i[[1]][[1]]
      ssenergy <- ela_i[[1]][[2]]
      
      for (s in 1:length(sstates)) {
        if (!(sstates[s] %in% rownames(rand_ela_summary))) {
          rand_ela_summary[sstates[s],] <- default_row
          rand_ela_summary[sstates[s],paste0("trial", ss)] <- ssenergy[s]}
        else{
          rand_ela_summary[sstates[s],paste0("trial", ss)] <- ssenergy[s]
        }
      }
    }
    rand_ela_summary <- rand_ela_summary[-1,]
    
    # select the most representative sample
    # (i.e., where the commonly appaeared stable states appeared most frequently)
    count_non_nan <- apply(rand_ela_summary,1,function(row) sum(!is.nan(row)))
    total_ss_score <- sapply(colnames(rand_ela_summary), function(x) 
    {si <- rand_ela_summary[,x] 
    si <- rownames(rand_ela_summary)[!is.nan(si)]
    return(sum(as.numeric(count_non_nan[si])))})
    rep_trial <- which.max(total_ss_score)
    rep_ela <- elaresults[[rep_trial]][[1]] # representative ELA results
    
  }else{
    rep_ela <- ela
  }
  # listing all tipping points and stable states in the representative resampling
  tipp_all <- unlist(rep_ela[[3]], use.names = FALSE)
  tipp_all <- unique(tipp_all[!(tipp_all=="Inf")])
  allpossible <- c(rep_ela[[1]],tipp_all)
  
  # Calculate qunatile distribution of energy
  if(reporting==TRUE){cat("calculate Energy.\n")}
  if(threads>1){
    energy_bootstrap <- foreach(i = 1:iterations,.packages = c("ELAplus","doParallel","magrittr","gtools")) %dopar% {
    rand_saparams <- sa2params(list(bs_params[[i]],enmat),env = env)
    hg_rand <- rand_saparams[[4]]
    J_rand <- rand_saparams[[2]]
    rand_cenergy <- sapply(allpossible, function(x){Energy(id2bin(x,nspecies),hg_rand,J_rand)})
    return(rand_cenergy)
    }
    }
  else{
      energy_bootstrap <- foreach(i = 1:iterations,.packages = c("ELAplus","doParallel","magrittr","gtools")) %do% {
        rand_saparams <- sa2params(list(bs_params[[i]],enmat),env = env)
        hg_rand <- rand_saparams[[4]]
        J_rand <- rand_saparams[[2]]
        rand_cenergy <- sapply(allpossible, function(x){Energy(id2bin(x,nspecies),hg_rand,J_rand)})
        return(rand_cenergy)}
  }
  rand_ene_summary <- data.frame(matrix(NaN, nrow = length(allpossible), ncol = iterations))
  colnames(rand_ene_summary) <- paste0("trial", 1:iterations)
  rownames(rand_ene_summary) <- allpossible
  for(r_ene in 1:length(energy_bootstrap)){
    energy_i <- energy_bootstrap[[r_ene]][allpossible]
    rand_ene_summary[,r_ene] <- energy_i
  }
  return(list(rep_ela,rand_ene_summary))
}

#' hamming_distance
#' @description Compute Hamming distance between two equal-length vectors.
#' @param x Vectors.
#' @param y Vectors.
#' @return Integer count of positions where x and y differ.
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
hamming_distance <- function(x, y) {
  sum(x != y)
}

#' gen_concompo
#' @description Merge overlapping index groups into connected components.
#' @param mm A list of vectors representing groups of indices.
#' @return A list of merged groups (connected components).
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
gen_concompo <- function(mm){
  concompo <- list()
  lapply(mm, function(mmm){
    matching_indices <- sapply(seq_along(concompo), function(i) {
      any(sapply(mmm, function(x) x %in% unlist(concompo[[i]])))
    })
    if (!any(matching_indices)) {
      concompo <- c(concompo, list(mmm))
    } else {
      matching_elements <- concompo[matching_indices]
      concompo <- c(concompo[!matching_indices], list(union(unlist(matching_elements), mmm)))
    }})
  return(concompo)
}
  
#' con_split
#' @description Split \code{data} into consecutive runs defined by \code{factor_variable}.
#' @param data Object to split
#' @param factor_variable Vector defining run-length groups.
#' @return A list of grouped subsets of \code{data} in run order.
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
con_split <- function(data, factor_variable){group_flag <- rle(factor_variable)$lengths
                        grouped_data <- split(data, rep(seq_along(group_flag), group_flag))
                        return(grouped_data)}

#' ssextend
#' @description subfunction for GELAObj
#' @param spli List of entries.
#' @param de Numeric/vector of values
#' @return A data.frame with columns (de, group_id, stb).
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
ssextend <- function(spli, de) {
  aaa <- sapply(spli, function(x) x[[1]])
  stb <- ifelse(de %in% aaa, 1, 0)
  data.frame(de, rep(spli[[1]][[3]], length(de)), stb)
}

#' replace_bin
#' @description Flip 0/1 values of \code{vec} at given indices.
#' @param vec Numeric/binary vector
#' @param index Integer indices to flip.
#' @return The modified vector with selected bits flipped.
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
replace_bin <- function(vec, index) {
  vec[index] <- abs(vec[index] - 1)
  return(vec)
}

#' foldList
#' @description Iteratively apply \code{func} over \code{vec}, returning all intermediate states.
#' @param func Binary function
#' @param init Initial value
#' @param vec Vector of inputs to fold over.
#' @return A list of accumulated results including the initial state.
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
foldList <- function(func, init, vec) {
  result <- list(init)
  for (i in seq_along(vec)) {
    result <- c(result, list(func(result[[length(result)]], vec[i])))
  }
  return(result)
}

#' MinTippingPath
#' @description Compute a minimum “tipping” energy path between two states.
#' @param s1 Another integer state ID
#' @param s2 Another integer state ID
#' @param hg explicit/implicit environmental factors
#' @param j species interaction matrix
#' @param tmax Iteration limit.
#' @return Numeric vector of energies along the selected path (or scalar if s1==s2).
#' @useDynLib ELAplus, .registration=TRUE
#' @export
MinTippingPath <- function(s1, s2, hg, j, tmax){
    n <- length(hg)
    ss1 <- id2bin(s1, n)
    ss2 <- id2bin(s2, n)
    dif <- ss2 - ss1
    pd <- which(abs(dif)==1)
    seq <- sample(pd)
    #############

    if(s1 == s2){return(Energy(ss1, hg, j))}
    if(length(seq) <= 7){
      seq_perms <- gtools::permutations(length(seq), length(seq), seq)

      pathes <- apply(seq_perms, 1, function(x) {
          samplepath <- foldList(replace_bin, ss1, x)
          sapply(samplepath, function(y) Energy(y, hg, j))
          })

      tote <- apply(pathes, 2, function(x) {
          dife <-  x[-1] - x[-length(x)]
          sum(dife[which(dife > 0)])})
      pe <- pathes[,which(tote==min(tote))]
      if(is.null(ncol(pe))){minpath <- pe}else{minpath <- rowMeans(pe)}
    }
    else{
    tt <- 0
    pree <- Inf
    mintote <- Inf

    while (tt < tmax) {
      tt <- tt + 1
      rpl <- sample(length(seq), 2) 
      seqnew <- seq 
      seqnew[rpl] <- seqnew[c(rpl[2], rpl[1])]
      samplepath <- foldList(replace_bin, ss1, seqnew)
      pathe <- sapply(samplepath, function(y) Energy(y, hg, j))
      dife <- diff(pathe)
      tote <- sum(dife[dife > 0])
      if (tote < mintote) {
        minpath <- pathe
        mintote <- tote
        pree <- tote
      }
      if (runif(1) < min(1, exp(pree) / exp(tote))) {
        seq <- seqnew
        pree <- tote
      }
    }}
    return(minpath)}


#' MarginePath
#' @description sub function of GELAObj
#' @param s1 Integer state ID
#' @param l Path length
#' @param hg explicit/implicit environmental factors
#' @param j species interaction matrix
#' @return Numeric vector giving the mean energy profile along the sampled paths.
#' @useDynLib ELAplus, .registration=TRUE
#' @noRd
MarginePath <- function(s1, l, hg, j) {
  n <- length(hg)
  ss1 <- id2bin(s1, n)
  
  pathes <- replicate(min(1024, 2^l), {
    samplepath <- replicate(l, {
        red <-  Reduce(function(vec, i) replace_bin(vec, sample(length(vec), 1)), seq_len(l), init = ss1, accumulate = TRUE)
        sapply(red, function(x) Energy(x, hg, j))
    })
    samplepath
  })
  
  mean_path <- apply(pathes, 1, mean)
  return(mean_path)}

#' GELAObj
#' @description Construct smoothed energy landscape surface objects from GELA output and parameters.
#' @param gela GELA object/list
#' @param sa Parameter list generated by \link{runSA}
#' @param threads Number of parallel workers.
#' @return A list containing surface grid data and state-specific surface points.
#' @useDynLib ELAplus, .registration=TRUE
#' @importFrom utils combn
#' @importFrom parallel makeCluster stopCluster clusterCall
#' @importFrom stats smooth.spline predict median embed
#' @importFrom dplyr bind_rows
#' @importFrom purrr map
#' @importFrom foreach foreach %do% %dopar%
#' @importFrom doParallel registerDoParallel
#' @importFrom magrittr %>%
#' @export
GELAObj <- function(gela, sa, threads = 1){
  de <- gela[[2]]
  s <- length(sa[1][[1]][,1])
  refenv <- gela[[3]]
  #derange <- range((sa[[2]])[,is.na(refenv)])
  #de0 <- seq(derange[[1]], derange[[2]], length.out = length(de))

  #elsurf <- foreach(x=gela[[1]], .packages=c("tidyverse","ELAplus","gtools","igraph")) %dopar% {GraphObj(x[[1]])}
  elsurf <- foreach(x=gela[[1]]) %do% {GraphObj(x[[1]])}
  ssp <- lapply(elsurf, function(x) {
    x[which(sapply(x$ccc_str, function(y) length(eval(parse(text = y))) == 1)), ]
  })
  sssp <- foreach(x=ssp, y=de) %do% {
      mat <- matrix(c(rep(y, nrow(x)), x[,3], x[,5]), nrow(x))
      split(mat, seq_len(nrow(mat)))}
  sssp <- unlist(sssp, recursive = FALSE)
  sslist <- unique(unlist(lapply(ssp, function(df) df[, 5])))
  ###
  if(length(sslist)>1){
    ss_combi <- combn(sslist, 2)
    #neibor states in assembly graph
    mm <- apply(ss_combi, 2, function(pair) {
      dist <- hamming_distance(id2bin(pair[1], s), id2bin(pair[2], s))
      if (dist == 1) return(pair)})
    mm <- Filter(function(x) !is.null(x), mm)
  }else{
    mm <- NULL
  }
  ###
  concompo <- gen_concompo(mm)
  grp <- c(concompo, sapply(setdiff(sslist, unlist(concompo)), function(x) list(x)))
  ssspa <- lapply(grp, function(gg) {
    lapply(sssp, function(ss) {
      ss[grepl(paste(gg, collapse = "|"), ss[3])]
    })})
  ssspa <- lapply(ssspa, function(x) x[sapply(x, function(y) length(y) > 0)]) #Remove empty elements
  if(length(grp)==1){
    ssspa <- lapply(ssspa, function(x) x[order(as.numeric(sapply(x, '[[', 1)))]) #Sorting within each element
  }else{
    ssspa <- sapply(ssspa, function(x) x[order(as.numeric(sapply(x, '[[', 1)))]) #Sorting within each element
  }
  ssspa <- ssspa[order(sapply(ssspa, length))] #Sort by length
  ssspa <- ssspa[order(sapply(ssspa, function(x){as.numeric(x[[1]][[1]])}))] #Sort by starting point
  ssspa_unlist <- unlist(ssspa, recursive = FALSE)
  spl <- con_split(ssspa_unlist, sapply(ssspa_unlist, '[[', 3))
  tssspastb <- lapply(spl, function(x) ssextend(x, de))
  ssid <- sapply(tssspastb, function(x){x[[2]][[1]]})
  if(length(ssid) > 1){
    hh <- apply(matrix(stats::embed(ssid, 2)[, 2:1],ncol=2),
                1, function(x) hamming_distance(id2bin(x[1], s), id2bin(x[2], s)))
    ml <- max(7, max(hh))
    
    posit <- Reduce(function(x, y) x + y, hh, accumulate = TRUE, init = ml + 1)
  }else{
    ml <- 7
    posit <- 8
  }
  
  cluster <- parallel::makeCluster(threads, type = if (.Platform$OS.type == "windows") "PSOCK" else "FORK")
  parallel::clusterCall(cluster, function(x) .libPaths(x), .libPaths())
  parallel::clusterSetRNGStream(cluster)
  doParallel::registerDoParallel(cluster)
  on.exit(parallel::stopCluster(cluster), add = TRUE)

  filledsurf <- foreach(i=seq(length(de)), .packages=c("ELAplus", "tidyverse", "purrr", "foreach", "gtools"),
                        .export = c("tssspastb","refenv","sa","posit","ml", "MinTippingPath", "MarginePath", "foldList", "replace_bin")) %dopar% {
      v <- lapply(tssspastb, function(x) x[i,])
      v <- do.call(rbind, lapply(v, function(x) as.matrix(x[,])))
      xde <- as.numeric(v[1,1])
      basenv <- refenv
      basenv[is.na(basenv)] <- xde
      basenv <- t(basenv)
      sa2p <- sa2params(sa, c(basenv))
      he <- sa2p[[1]]
      je <- sa2p[[2]]
      ge <- sa2p[[3]]
      hge <- sa2p[[4]]
      ###
      pstb <- as.numeric(v[,3])
      sid <- ssid[as.numeric(v[,3])==1]
      zx <- posit[as.numeric(v[,3])==1]
      ###
      scape <- unlist(map(seq_along(sid)[-length(sid)], function(j) {
          MinTippingPath(sid[j], sid[j + 1], hge, je, 10000)}))
      if(length(scape) == 0){
          scaledscp <- data.frame(z = zx,  energy = MinTippingPath(sid[[1]], sid[[1]], hge, je, 10000))}
      else{
          scaledscp <- data.frame(z = seq(zx[1], zx[length(zx)], length.out = length(scape) - 1), energy = scape[-length(scape)])}
      leftmargine <- MarginePath(ssid[1], zx[1] - 1, hge, je)
      leftmargine <- data.frame(
        z = seq_along(leftmargine[-1]),
        energy = rev(leftmargine[-1]))
      rightmargine <- MarginePath(ssid[length(ssid)], ml + (posit[length(posit)] - zx[length(zx)]), hge, je)
      rightmargine <- data.frame(
        z = seq(zx[length(zx)] + 1, posit[length(posit)] + 1 + ml),
        energy = rightmargine)
      rbind(leftmargine, scaledscp, rightmargine)
  }

  cc <- 0
  intpol <- lapply(filledsurf, function(df) {
    cc <<- cc + 1
    spline_fit <- smooth.spline(df$z, df$energy, spar = 0.2)  # Set adjustable smoothing parameter spar
    pol <- predict(spline_fit, seq(df$z[1], df$z[length(df$z)], length.out = 100))$y
    dx <- seq(1, df$z[length(df$z)], length.out = 100)
    dex <- rep(de[cc], length(pol))#dex <- rep(de0[cc], length(pol))
    data.frame(dx = dx, dex = dex, pol = pol)
  })
  surf3d <- lapply(intpol, function(df) {
    spline_fit <- smooth.spline(df$dx, df$pol, spar = 0.2)
    smoothed_pol <- predict(spline_fit, df$dx)$y
    filtered_pol <- stats::filter(smoothed_pol, filter = rep(1/5, 5), sides = 2, circular = TRUE)
    data.frame(x = df$dx, y = df$dex, z = filtered_pol)
  })
  surf3dx <- dplyr::bind_rows(surf3d)

  l3d <- lapply(tssspastb, function(y){
    si <- which(y$stb==1)
    xe <- posit[which(ssid==y[1,2])]
    xval <- surf3d[[1]][,1]
    nearest_x <- which.min(abs(xval - xe))
    data.frame(t(sapply(si, function(x) surf3d[[x]][nearest_x,])))})
  l3dx <- dplyr::bind_rows(lapply(seq_along(l3d), function(i) {
    df <- l3d[[i]]
    df$Index <- i
    df$ssid <- sapply(tssspastb,function(x){unique(x[,2])})[i]
    df <- df[,c(4,5,1,2,3)]
    return(df)}))
  return(list(surf3dx, l3dx))
}

---
title: "An introduction to ELAplus: Energy Landscape Analysis"
author: "Sotaro Takano, Kenta Suzuki, Hiroaki Fujita"
output: 
  rmarkdown::html_vignette:
    toc: true
    number_sections: true
vignette: >
  %\VignetteIndexEntry{An introduction to ELAplus: Energy Landscape Analysis(eval=FALSE)}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{=html}
<style>
body {
text-align: justify}
</style>
```

```{r setup, include=FALSE}
library(knitr)
opts_chunk$set(
  fig.align = "center",
  out.extra = 'style="max-width:100%; overflow-x:auto; white-space: nowrap;"',
  tidy = FALSE
)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  message  = FALSE,
  warning  = FALSE
)
```

# Overview

ELAplus provides tools for Energy Landscape Analysis (ELA), a framework for
analyzing dynamical systems represented as weighted networks. The
package is designed primarily for the analysis of ecological community
composition data, but the framework is applicable to other multistable
systems.

The typical workflow consists of:

**- Fitting community composition data to a (possibly extended) pairwise
maximum entropy model**

**- Constructing an energy landscape from the fitted model**

**- Identifying stable states and transition structures**

**- Visualizing the resulting landscape and interactions**

The theoretical background is described in [Suzuki et al.
(2021)](https://esajournals.onlinelibrary.wiley.com/doi/abs/10.1002/ecm.1469).

# Example: Basic Energy Landscape Analysis

This vignette demonstrates a standard ELA workflow using test dataset.

## Input dataset

-   `baseabtable`: **samples** - **species** abundance table (data.frame
    or matrix).
    -   Rows correspond to sampling units (sites, time points, subjects,
        etc.)
    -   Columns correspond to species (or OTUs/ASVs/features/taxa).
-   `basemetadata` (optional): a **samples** - **factors** metadata
    table (data.frame or matrix).
    -   Rows correspond to sampling units (same as those in
        `baseabtable`)
    -   Columns correspond to environmental parameters (e.g.,
        temperature, pH, etc.).

```{r}
# load package
library(ELAplus)
# load test data set
# species abundance table
data("baseabtable")
head(baseabtable)

# metadata (environmental factors)
data("basemetadata")
head(basemetadata)
```

## (Optional) Data generation using a Gibbs sampling (heat-bath algorithm)

The simulated community composition data are also available.

In this example, the data comprising `24` species and `512` samples were
generated through `100` iterations of Gibbs sampling (heat-bath
algorithm).

Both implicit/explicit environmental factors (`h.act` and `g.act`) are
considered. The number of explicit environmental factors is specified by
`ne` (`2` factors in this example).

`j.act` is the interaction matrix.

```{r}
# load package
library(ELAplus)
# Generate random parameters
hb_params <- hb.paramgen(24, ne = 2) # the number of environmental factors
h.act <- hb_params[[1]]
j.act <- hb_params[[2]]
g.act <- hb_params[[3]]

# Simulate community composition data
rand_data <- HeatBath(100, 512, h.act, j.act, g = g.act)
rand_mat   <- rand_data[[1]]
rand_enmat <- rand_data[[2]]

```

## Formatting the dataset

ELAplus expects the community data in a standardized representation.

`Formatting()` converts raw input tables into aligned occurrence
(binary) matrix and (optionally) environmental factor matrix.

### Inputs

-   `baseabtable`: **samples** - **species** abundance table (data.frame
    or matrix).

-   `basemetadata` (optional): a **samples** - **factors** metadata
    table (data.frame or matrix).

    -   This is used when you want to include environmental parameters
        (e.g., temperature, pH, treatment group).
    -   If you do not have environmental data, pass NULL (or an empty
        object).

```{r}
formatted <- Formatting(baseabtable = baseabtable, basemetadata = basemetadata)
```

### Key preprocessing options

#### Normalization (normalize)

-   If `normalize = 1`, the abundances within each sample are normalized
    so that each row sums to 1. This is useful when:
    -   total sequencing depth differs strongly between samples etc...
-   If `normalize = 0`, the original scale is preserved.

#### Occurrence binarization (the “0/1 threshold”)

-   ELA operates on binary presence/absence states (each species is
    present or absent).
-   `Formatting()` creates the occurrence matrix by applying the first
    element of `parameters`:
-   `parameters[1]` (“0/1 threshold”): a numeric cutoff used to define
    presence.
    -   Example: `parameters[1] = 0.05` means “abundance \> 0.05 is
        present; otherwise absent”.

#### Species filtering by prevalence (min/max occurrence thresholds)

-   Species that are too rare or almost always present typically add
    little information or inflate the results in construction of energy
    landscape.

-   `Formatting()` therefore filters species based on prevalence across
    samples using:

-   `parameters[2]` (“min occurrence threshold”)

-   `parameters[3]` (“max occurrence threshold”)

-   These are interpreted as proportions of samples in which the species
    is present **after binarization**.

-   Example:

    -   `parameters = c(0.01, 0.01, 0.99)` means:
    -   present(1) if abundance \> 0.01,
    -   keep species whose presence frequency is \> 0.01 and \< 0.99.

```{r, eval=FALSE}
formatted <- Formatting(baseabtable = baseabtable, basemetadata = basemetadata,
                        normalize = 1, parameters = c(0.01, 0.01, 0.99))
#> Processed 256 samples.
#> Relative abundance threshold = 0.01
#> Occurrence threshold (lower) = 0.01 
#> Occurrence threshold (upper) = 0.99
#> Selected  16  out of  16 species.
ocvecs <- formatted[[1]]  # community composition binary matrix
abvecs <- formatted[[2]]  # community abundance matrix (not binarized)
envecs <- formatted[[3]]  # environmental variables (optional)
```

```{r, echo=FALSE, out.width="80%"}
knitr::include_graphics("figures/ParameterMatrix.png")
```

## Parameter fitting

### Extended maximum pairwise entropy model

The energy $E(\sigma^{(k)},\varepsilon)$ of each community composition
$\sigma^{(k)}$ is assigned by imposing the potential structure to the
state space by introducing the extended pairwise maximum entropy model.
Briefly, the energy is determined by three parameters. For details,
please see [Appendix].

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/EPMEM_abstract2.png")
```

-   `h`: Implicit (unobserved) environmental parameter for each species

-   `J`: Species-species interaction matrix

-   `g`: Explicit environmental parameters for each species

### Parameter fitting algorithm

Fitting the parameters of the pairwise maximum entropy model is
performed using a stochastic approximation algorithm.

We approximate the intractable model expectation in the likelihood
gradient with a persistent (warm-started) Gibbs/heat-bath chain advanced
by one sweep per parameter iteration, corresponding to persistent
contrastive divergence with one sweep (PCD-1), also known as stochastic
maximum likelihood (SML).

It repeatedly simulates data under the current parameters and then
updates `h`, `J`, and `g` to reduce the discrepancies between observed
vs simulated metrics.

Two observed co-occurrence statistics are used for the fitting.

1.  Species–species co-occurrence: $y = ocvecs^{T}ocvecs$

2.  Environment–species co-occurrence: $y^{env} = envecs^{T}ocvecs$

Then, the algorithm repeatedly updates the parameters by comparing

-   observed vs simulated species–species co-occurrence
    $y_{diff} = y_{observed} - y_{simulated}$
-   observed vs simulated environment–species co-occurrence
    $y^{env}_{diff} = y^{env}_{observed} - y^{env}_{simulated}$

Each parameter independently updated by those metrics:

$$
\Delta{h} \propto diag(y_{diff})
$$ $$
\Delta{J} \propto y_{diff}
$$

$$
\Delta{g} \propto y^{env}_{diff}
$$

## Hyperparameter search for fitting

For simplicity, the usage of ELA without explicit environmental
parameters (`envecs`) is first explained.

Fitting the pairwise maximum entropy model with `runSA()` depends on
several optimization configurations, and the best settings can vary
across datasets (number of species, sample size, sparsity).

ELAplus provides a dedicated grid-search routine to choose hyperparameters
empirically.

`Findbp()` performs a grid search over three key hyperparameters:

\-`lmd` : **a sparsity penalty** to encourage sparse matrix. Larger
values typically lead to fewer nonzero parameters.

\-`we` : **AdamW weight decay**, a form of L2 regularization applied
during optimization.

\-`maxlrs` : **maximum learning rate** in fitting algorithm (SML).


```{r, eval=FALSE}
bpresult <- Findbp(ocvecs, rep = 2, threads = 2,
                   cvmode="10fold",nrepeat = 10, we=c(0.001),totalit=4000,
                   lmd=c(1e-05,1e-03,0.01),maxlrs = c(0.005),
                   tol=0.03, runadamW=TRUE,sparse=TRUE,fastfitting=TRUE)
#Start hyper-parameter search:
#There are 300 tasks
#Finished 300 tasks
#SA: elapsed time 15.08 sec

bp <- bpresult[[1]]  # typically best bpresult
bpm <- as.numeric(unlist(strsplit(names(bp)[1], split = "_")))
allresults <- bpresult[[2]]
```

### What Findbp does

**1. Splits samples into training and test subsets**

-   The data are split using cross-validation based on cvmode. 
-   For "10fold" or "5fold", samples are randomly assigned into k folds (repeated `nrepeat` times), and each fold is used as the test set while the remaining folds are used for training.
-   For "loo" (leave-one-out), each sample is used once as the test set while the rest form the training set.

**2. Fits the model on the training subset for each grid point**

-   For each combination of candidate hyperparameters, the function
    calls `runSA()`.

**3. Evaluates predictive performance on held-out samples**

-   Each fitted model is scored by its predictive error on the test
    subset (lower is better).
-   The best-performing settings are returned.
-   If `fastfitting=TRUE`, the function returns candidates with the
    lowest iterations within `tol` of the best observed error.

### Fitting results can be visualized by `plotSAtest`.

```{r plotSAtest, eval=FALSE}
plotSAtest(allresults)
# "id" corresponds to hyperparameter: lmd_we_maxlr
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/plotSAtest.png")
```

### Fitting the extended maximum pairwise entropy model

The best parameters found by `Findbp()` are used for real analyses by
`runSA()`.

```{r, eval=FALSE}
sa <- runSA(
  ocvecs,
  enmat   = NULL,
  rep     = 32,
  getall  = FALSE,
  lambda  = bpm[[1]],
  we      = bpm[[2]],
  maxlr = bpm[[3]],
  totalit = bpm[[4]]
)
#> Start parameter fitting:
#> .SA: elapsed time 0.13 sec
```

## Constructing the energy landscape

After fitting the (extended) pairwise maximum entropy model with
`runSA()`, the next step is to construct energy landscape over community
compositions using the fitted parameters.

In ELAplus, this is performed by `ELA()`, which identifies:

**1. Stable states** Here, the base of the basin (with minimal energy)
is regarded as a stable state, which is distinct from the equilibrium
point of the dynamical system or the attractor. For details, see [Suzuki
et al.
(2021)](https://esajournals.onlinelibrary.wiley.com/doi/abs/10.1002/ecm.1469),

**2. Tipping points (energy barriers connecting pairs of basins/stable
states).** A connecting pair of basins have basin boundary and the point
with the lowest energy barrier is regarded as tipping point here.

The fitted model defines an energy value for any community composition
(typically a binary presence/absence vector of community members).

```{r, eval=FALSE}
ela <- ELA(
  sa,
  env            = NULL,
  threads        = 1,
  SS.itr         = 20000,
  FindingTip.itr = 10000,
  reporting      = TRUE
)
#> Start ELA:
#> 7 stable states were found.
#> Checking 21 tipping points.
#> converting...
#> ELA: elapsed time 2.19 sec
```

## ELA output

After stable states and tipping points are computed, `ELA()` summarizes
and standardizes the outputs. This conversion step produces a compact
representation of the energy landscape structure:

-   [[1]] Stable state IDs (64-base index encoded from community
    composition array by `bin2id()`)
-   [[2]] Stable state energies
-   [[3]] Tipping point ID (for stable-state pairs)
-   [[4]] Tipping energies (corresponding barrier energies)

The output format is compatible to downstream steps such as pruning
(ELPruning) and visualization (e.g., disconnectivity graphs).

```{r, eval=FALSE}
ela[[1]] # stable state IDs
#> [1] "09x" "01t" "EWB" "EY8" "1uV" "1mL" "17s"
ela[[2]] # stable state energies
#> [1] -11.099418 -10.297931  -9.544050  -8.961634  -8.005040  -6.440385  -6.269140
```

### What ELA does

1.  Identifying stable states by repeated steepest descent
    (`SSestimate`)

-   The number of repetitions is controlled by `SS.itr` (default 20000).
-   More iterations increase the chance of discovering all major stable
    states, especially when the number of species is large, landscape
    has many stable states, etc...

2.  Finding tipping points between stable states

-   Once a set of stable states is obtained, `ELA()` evaluates
    transitions between them by searching for tipping points between all
    pairs of stable states.
-   The number of iterations is controlled by `FindingTip.itr`. As with
    `SS.itr`, increasing `FindingTip.itr` improves robustness at the
    cost of runtime. This step can be computationally heavy because it
    is performed across stable-state pairs.

## Pruning stable states

The energy landscape constructed by `ELA()` first identified stable
states as much as possible. Some of these stable states correspond to
very shallow basins.

Such stable states are often sensitive to noise and does not largely
explain dominant multistability of the system.

`ELPruning()` addresses this issue by iteratively pruning shallow basins
and merging them into deeper, neighboring basins.

```{r, eval=FALSE}
elap <- ELPruning(ela, th=0.1, type="relative")
#> Start pruning:
#> *...
#> ELPruning: elapsed time 0.05 sec
```

### Output and interpretation

`ELPruning()` returns a list with two elements:

[[1]] Pruned ELA object (same format as the output of `ELA()`):

-   [[1]] updated stable state IDs,
-   [[2]] updated stable state energies,
-   [[3]] updated tipping point IDs,
-   [[4]] updated tipping energies.

```{r, eval=FALSE}
elap[[1]][[1]] # updated stable state IDs
#> [1] "09x" "EWB" "EY8" "1uV"
elap[[1]][[2]] # updated stable state energies
#> [1] -11.099418  -9.544050  -8.961634  -8.005040
```

[[2]] Stable-state mapping table

-   ss.before.pruning: stable state IDs from the original (unpruned)
    landscape
-   ss.after.pruning: corresponding stable state IDs after pruning

```{r, eval=FALSE}
elap[[2]]
#>  ss.before.pruning ss.after.pruning
#> 1               09x              09x
#> 2               01t              09x
#> 3               EWB              EWB
#> 4               EY8              EY8
#> 5               1uV              1uV
#> 6               1mL              09x
#> 7               17s              09x
```

### Basin depth and pruning criterion

For each stable state, a basin depth is defined as the energy difference
between the energy of the stable state itself and the lowest tipping
point energy connecting that stable state to any other stable state
($\Delta E$).

The basin depth measures how “robust” a stable state is against
transitions: deeper basins correspond to more robust, while shallow
basins correspond to more unstable states.

In `ELPruning()`, there are two methods `relative` or `bootstrap`.

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/deltaE_scheme.png")
```

#### Relative depth pruning

If `type="relative"` is set, `th` (relative depth threshold) specifies
the minimum acceptable energy barrier relative to the deepest basin
barrier. $$
\Delta E_{\mathrm{min}} < th * \Delta E_{\mathrm{max}}
$$ For example, `th = 0.2` means that basins shallower than 20% of the
deepest basin ($\Delta E_{\mathrm{max}}$) are pruned. The results of
pruning is usually sensitive to this value.

```{r, eval=FALSE}
elap <- ELPruning(ela, th=0.2, type="relative")
#> Start pruning:
#> *...
#> ELPruning: elapsed time 0.05 sec
elap[[1]][[1]] # updated stable state IDs
#> [1] "09x" "EWB" "1uV"
```

#### bootstrap samples pruning

Alternatively, users can perform pruning of stable states by calculating
vulnerabilities in bootstrap samples in each stable state.

This method is explained in the next section:[Pruning with bootstrap
resampling].

## Visualization

### Disconnectivity graph

A disconnectivity graph is a compact visualization of the global
structure of an energy landscape.

-   It summarizes how stable states are separated by energy barriers
    (tipping points).
-   The vertical axis represents energy
-   Branches illustrate how stable states are connected and separated by
    tipping points.

```{r showDG, eval=FALSE}
# Before pruning
showDG(ela, ocvecs, "example")
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/DGexample.png")
```

```{r showDG_pruned, eval=FALSE}
# After pruning
showDG(elap[[1]], ocvecs, "example")
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/DGexample_pruned.png")
```

#### How to interpret the graph

-   Lower energies correspond to more stable configurations.
-   Branching structure indicates which basins are “close” in terms of
    energy barriers.
-   Barrier height as a transition difficulty proxy.

#### Optional checking bootstrap resampling

The reliability of stable states can be also examined by the bootstrap
resampling of training data sets (the occurrence matrix). The function,
`bootstrap_SA()`, performs parameter fitting with bootstrap resampling
of the data sets.

`bootstrap_ELA()` uses the results of `bootstrap_SA()` (i.e., estimated
parameters: h, J) for computing the variability of energy gaps between
tipping points and stable states arising from uncertainty due to finite
sample size.

If the estimated stable state is statistically reliable, the estimated
$\Delta E$ is expected to be consistently smaller than 0 across
bootstrap resamples of the training data.

Therefore, the probability of bootstrap samples with $\Delta E < 0$
reflects the statistical significance of the estimated stable state.
Here, this quantity is referred to as a "bootstrap-approximated p-value"
and is used to assess statistical reliability.

As an example, the distribution of $\Delta E$ between "01t" and "09/" is
shown as a histogram. More than 25% of bootstrap samples yield
$\Delta E > 0$ (i.e., p-val \> 0.25), and this case is regarded as
statistically not reliable.

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/deltaE_hist.png")
```

If `bootstrap_ene` is specified in `showDG()`, the function displays an
approximate p-value in each stable state. This option is helpful when
you want to see uncertainty in estimated energies in ELA.

```{r showDG_pvalue, eval=FALSE}
bs_params <- bootstrap_SA(ocvecs, rep = 16, threads=8,　bootstrap.it = 1000, 
                          lambda  = bpm[[1]],we = bpm[[2]],
                          maxlr = bpm[[3]], totalit = bpm[[4]])

bootstrap <- bootstrap_ELA(bs_params,ocvecs,ela=ela,threads = 1)
ela_rep <- bootstrap[[1]]
bootstrap_ene <- bootstrap[[2]]
showDG(ela_rep,ocvecs,bootstrap_ene = bootstrap_ene, min_pval = 0.01, 
       scale_labels = c(0.05,0.1,0.2,0.5))
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/DGexample_pval.png")
```

#### Pruning with bootstrap resampling

The pruning stable states is also possible using the bootstrap results.

The `ELPruning()` with `type="bootstrap"`used the distribution of
$\Delta E$ in bootstrap samples and evaluated whether the predicted
stable states are statistically reliable.

If the p-value is above threshold (e.g., 0.05), the function prunes such
unreliable and shallow basins (stable states) and merges them into
deeper, neighboring basins. The pruned graph is follows.

```{r showDG_IQR_pruned, eval=FALSE}
elap_rep <- ELPruning(ela, type = "bootstrap",bootstrap_ene=bootstrap_ene,lower_qr = 0.25)
showDG(ela_rep,ocvecs,bootstrap_ene = bootstrap_ene, min_pval = 0.01, 
       scale_labels = c(0.05,0.1,0.2,0.5))
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/DGexample_pval_pruned.png")
```

If `getGraohObj=TRUE`, `showDG()` function returns a data frame for
plotting a disconnectivity graph. In the following example,
disconnectivity graph is drawn using `ggplot()`. Tipping points and
stable states are shown as triangles and circles, respectively, and
colored by energies.

```{r showDG_manual, eval=FALSE}
grobj_to_plot <- showDG(ela_rep,ocvecs,bootstrap_ene = bootstrap_ene,getGraphObj = TRUE)
grobj_point <- grobj_to_plot[grobj_to_plot$point_str != "",]
g <- ggplot(
  grobj_to_plot,
  aes(x=.data$nodes2xposi, y=.data$energy, label=.data$point_str)) + 
  xlab("") + geom_text(hjust=0.75, vjust=2, aes(fontface=2)) + geom_path() +
  geom_point(
    data = grobj_point,
    mapping = aes(
      x = .data$nodes2xposi,
      y = .data$energy,
      color = .data$energy,
      shape = .data$type,
    ),size = 4
  ) +
  scale_color_viridis_c(option = "plasma",alpha=0.8,
                        direction=-1,na.value = "grey") +
  coord_cartesian(xlim=c(0.75,7.5), ylim=c(-5.8,-11.5))
plot(g)
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/DG_manual.png")
```

### Species interaction graph

`showIntrGraph()` visualizes the pairwise interaction structure inferred
by the fitted model in a 2D “interaction space”.

It extracts estimated parameters from the resulting interaction matrix
(jest).

-   Each species is plotted as a point at its PCoA coordinates.
-   Links with value \> `th` are classified as positive, while links
    with value \< `-th` are classified as negative.

The plot draws these two link types separately and colors them
differently (positive = red, negative = blue).

```{r ShowIntrGraph, eval=FALSE}
showIntrGraph(
  elap[[1]],
  sa,
  th = 0.01,
  annot_adj = c(0.75, 1.00)
)
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/IntrGraph.png")
```

# Environmental gradient analysis

The stability structure of a community is not fixed, but changes
continuously along an environmental gradient (e.g., temperature,
nutrient availability, disturbance intensity).

`GradELA()` is designed to capture this dependence by estimating energy
landscapes repeatedly across a controlled range of an environmental
factor and summarizing how stable states and transition barriers change.

When environmental variables (`envecs`) are included, ELAplus supports
**Gradient Energy Landscape Analysis** via `GradELA()`. The parameter
fitting process is same as that for normal ELA except for specifying
`enmat`.

```{r, eval=FALSE}
bpresult <- Findbp(ocvecs, enmat = envecs, rep = 2, threads = 2,
                   cvmode="10fold",nrepeat = 10, we=c(0.001),totalit=4000,
                   lmd=c(1e-05,1e-03,0.01),maxlrs = c(0.005),
                   tol=0.03, runadamW=TRUE,sparse=TRUE,fastfitting=TRUE)
bp <- bpresult[[1]]
bpm <- as.numeric(unlist(strsplit(names(bp)[1], split = "_")))

sa <- runSA(
  ocvecs,
  enmat   = envecs,
  rep     = 16,
  getall  = FALSE,
  lambda = bpm[[1]],
  we = bpm[[2]],
  maxlr = bpm[[3]],
  totalit = bpm[[4]]
)

```

## Concept of GradELA

`GradELA()` proceeds in three conceptual stages:

1.  Specification of an environmental gradient

-   One environmental factor `eid` is varied over a specified `range`
    and `steps`.
    -   `eid` is environmental factor for changing (one column in the
        environmental matrix `enmat`).
    -   `range` is the minimum and maximum values of the environmental
        factor.
    -   `steps` is how many discrete points are evaluated.

2.  Energy landscape construction at each step

-   For each environmental value along the gradient,`ELA()` and
    `ELPruning()` are performed.
-   When varying one environmental factor, all remaining factors must be
    assigned fixed values.
-   This is controlled by the `env` (it corresponds to a row in `enmat`)
    as follows:
    -   If `env` is provided, it is used as the reference vector for all
        environmental factors except `eid`.
    -   If `env` is not provided, the mean vector of `enmat` is used
        automatically.

```{r, eval=FALSE}
# This process takes few minutes with a small number of threads, and here pre-computed results are shown.
gela <- GradELA(
  sa      = sa,
  eid     = "factor.1",
  enmat   = envecs,
  env     = NULL,
  range   = c(-1, 1),
  steps   = 32,
  th = 0.05,
  threads = 8, 
  pruning_type = "relative", 
  bs_params = bs_params,
  ocmat = ocvecs)
```

Users can use bootstrap samples in ELPruning process by
`pruning_type="bootstrap"`. Other options are equivalent to
`ELPruning()`.

```{r, eval=FALSE}
# This process takes few minutes with a small number of threads, and here pre-computed results are shown.
bs_params <- bootstrap_SA(ocvecs, enmat = envecs,rep = 16, threads=8,bootstrap.it = 1000, 
                          lambda  = bpm[[1]],we = bpm[[2]],
                          maxlr = bpm[[3]], totalit = bpm[[4]])

gela <- GradELA(
  sa      = sa,
  eid     = "factor.1",
  enmat   = envecs,
  env     = NULL,
  range   = c(-1, 1),
  steps   = 32,
  lower_qr = 0.25,
  threads = 8, 
  pruning_type = "bootstrap", 
  bs_params = bs_params,
  ocmat = ocvecs)
```

## Output

`GradELA()` returns a list with three elements:

**1. Pruned ELA results for each step**

-   A list of energy landscape objects (same format as `ELA()` output,
    after pruning).

**2. Environmental value sequence**

-   The numeric vector of environmental values used for the target
    factor (`eid`).

**3. Reference environmental vector**

-   The full environmental vector used as a baseline for non-varied
    factors.

```{r, eval=FALSE}
gela[[1]][[1]][[1]][[1]] # stable state IDs at 1st environmental parameter
# [1] "0Ht" "EWB" "1uV"
gela[[1]][[20]][[1]][[1]] # stable state IDs at 20th environmental parameter
# [1] "09x" "EWB"
```

## Visualization of gradient ELA results

The resulting object can then be plotted using `showSSD()`, which
visualize the regime shift: how the number, identity, and energy of
stable states change along the environmental gradient.

-   The plot is drawn by overlaying multiple line traces, one per stable
    state.
    -   x-axis corresponds to the change of environmental parameters and
    -   y-axis corresponds to the energy of stable states
-   Changes in community state energy across the gradient reflect how
    resilient they are under changing conditions.
-   A plot discontinuity indicates a stable state that doesn't emerge
    under these conditions (or is pruned).

```{r showSSD, eval=FALSE, fig.width=8, fig.height=6, message=FALSE}
# GradELA result (ELPruning with type="relative").
showSSD(gela)
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/SSD.png")
```

```{r showSSD_bootstrap, eval=FALSE, fig.width=8, fig.height=6, message=FALSE}
# GradELA result (ELPruning with type="bootstrap").
showSSD(gela)
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/SSD_pruned.png")
```

If users want to modify the figure configuration of the diagram,
`showSSD_ggplot()` function returns a data frame for plotting. Then
users can the configuration manually using `ggplot()`.

```{r showSSD_ggplot, eval=FALSE, fig.width=8, fig.height=6, message=FALSE}
ssd_df <- showSSD_ggplot(gela,plot = FALSE,getSSDobj = TRUE)
g <- ggplot(ssd_df, aes(x = env, y = Energy, color = SS_ID)) +
            geom_point(shape = 19) +
            geom_line(aes(group = SS_ID))

```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/SSD_ggplot.png")
```

## 3D plot of gradient ELA results

`showGELA3D()` visualizes a smoothed energy landscape surface produced
by `GELAObj()` and overlays stable-state trajectories.

-   **Z axis (Energy):** model-predicted energy evaluated under the
    fitted parameters at each environmental parameter value.
-   **Y axis (y):** the environmental gradient value used in
    `GradELA()`.
-   **X axis (x):** stable states. Positions of stable states are
    determined by their nearest-neighbor relationships (in Hamming
    distance).

```{r GELA3D-demo, eval=FALSE, fig.width=8, fig.height=6, message=FALSE}
# This process takes few minutes, and here pre-computed result is drawn. 
gelaobj <- GELAObj(gela, sa=sa,threads=2)
showGELA3D(gelaobj,theta = 30,phi = 30)
```

```{r, echo=FALSE, out.width="95%"}
knitr::include_graphics("figures/GELA3D.png")
```

# Appendix

**Overview of extended maximum pairwise entropy model**

The model determines the probability of the occurrence of community
composition $\sigma^{(k)}$ in an environmental condition
$\varepsilon =  (\varepsilon_{1}, \varepsilon_{2}, ..., \varepsilon_{M})$,
environmental factors such as pH or temperature.

$$
E(\sigma^{(k)},\varepsilon)  = - \sum_{i=1}^{S} h_i \sigma_i^{(k)}
- \sum_{j=1}^{S} \sum_{i=1}^{M}
g_{ij} \varepsilon_i^{(k)} \sigma_j^{(k)}
- \sum_{i=1}^{S} \sum_{j=1,i \neq j}^{S} J_{ij} \sigma_i^{(k)} \sigma_j^{(k)}
$$ where:

-   $S$ is the number of species,\
-   $M$ is the number of environmental variables,\
-   $h_i$ is the implicit environmental variable for species $i$,\
-   $g_{ij}$ is the explicit effect of environmental variable $i$ on
    species $j$,\
-   $J_{ij}$ is the interaction strength between species $i$ and $j$,\
-   $\sigma_i^{(k)}$ is the state (occurrence) of species $i$ at sample
    $k$,\
-   $\varepsilon_i^{(k)}$ is the value of environmental variable $i$ at
    sample $k$.\
-   $J_{ij}$ is the interaction strength between species $i$ and $j$,

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{GradELA}
\alias{GradELA}
\title{GradELA}
\usage{
GradELA(
  sa = sa,
  pruning_type = "relative",
  eid = NULL,
  enmat = enmat,
  env = NULL,
  range = NULL,
  steps = 16,
  th = 0.05,
  threads = 1,
  SS.itr = 20000,
  FindingTip.itr = 10000,
  lower_qr = 0.05,
  bs_params = NULL,
  ocmat = NULL
)
}
\arguments{
\item{sa}{output of runSA function.}

\item{pruning_type}{the type of pruning \code{"relative"}(use with \code{th}) or \code{"bootstrap"} (use with \code{lower_qr},\code{ocmat} and \code{bs_params})}

\item{eid}{target environmental factor: the specified environmental factor to be changed.}

\item{enmat}{array/matrix of environmental factors for each sample (normalization required).}

\item{env}{array of environmental factor values. 
For the environmental parameters that are not specified for \code{eid}, these values are used as references.
If not specified, then the mean of values in \code{enmat} are used as references.}

\item{range}{range over which the environmental factor specified by eid is changed.}

\item{steps}{number of steps in the range.}

\item{th}{ratio of the depth of the basin to be pruned with respect to the deepest basin.}

\item{threads}{number of cores (threads) used for calculation.}

\item{SS.itr}{number of iterations to identify stable states.}

\item{FindingTip.itr}{number of iterations to identify tipping points.}

\item{lower_qr}{Numeric value. Quantile used for constructing confidence intervals from bootstrap samples to check 
whether the estimated energy gaps include zero. Default is \code{lower_qr = 0.05},necessary when \code{type="bootstrap"}.}

\item{bs_params}{output of \link{bootstrap_SA}. list of estimated parameters using bootstrap resampled data set.
necessary when \code{type="bootstrap"}.}

\item{ocmat}{community composition matrix.}
}
\value{
A list of three elements: (1) ELA results for each step (pruned),
  (2) the sequence of environmental values, and (3) the reference env vector.
}
\description{
Estimate energy landscape across environmental gradients.

Runs \link{ELA} and \link{ELPruning} comprehensively across specified environmental parameters.
One environmental parameter can be changed by specifying \code{eid} over \code{steps} within \code{range}. 
If \code{env} is not given, the mean vector of \code{enmat} is used as the reference environmental condition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{ELPruning}
\alias{ELPruning}
\title{ELPruning}
\usage{
ELPruning(
  ela,
  type = "relative",
  th = 0.05,
  lower_qr = 0.05,
  bootstrap_ene = NULL,
  threads = 1,
  reporting = TRUE,
  fork = TRUE
)
}
\arguments{
\item{ela}{output of ELA function.}

\item{type}{the type of pruning \code{"relative"}(use with \code{th}) or \code{"bootstrap"} (use with \code{lower_qr} and \code{bootstrap_ene})}

\item{th}{ratio of the depth of the basin to be pruned with respect to the deepest basin.}

\item{lower_qr}{Numeric value. Quantile used for constructing confidence intervals from bootstrap samples to check 
whether the estimated energy gaps include zero. Default is \code{lower_qr = 0.05}.}

\item{bootstrap_ene}{a data frame of bootstrap energies generated by \link{bootstrap_ELA}. Necessary when \code{type="quantile"}.}

\item{threads}{number of cores (threads) used for calculation.}

\item{reporting}{enable/disable the reporting function.}

\item{fork}{logical; if TRUE and \code{threads > 1}, a cluster is created internally for parallel execution.}
}
\value{
A list of two elements: (1) pruned ELA tuple (same format as \link{ELA} output),
  and (2) a data frame with columns \code{ss.before.pruning} and \code{ss.after.pruning}.
}
\description{
Pruning shallow stable states of energy landscape.

Iteratively prunes stable states (merges basins). 
If \code{type="relative"},  stable state whose energy depth is smaller than \code{th} times the deepest
basin depth are pruned and merged to the deeper basins. 
If \code{type="bootstrap"}, stable state whose bootstrap confidence interval (default: 95% quantile) of energy gap between tipping point
cross zero is pruned. 
The function updates stable states and corresponding tipping
structures, and returns both the pruned ELA object and a mapping table
from stable states before pruning to those after pruning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gels.R
\name{GELAObj}
\alias{GELAObj}
\title{GELAObj}
\usage{
GELAObj(gela, sa, threads = 1)
}
\arguments{
\item{gela}{GELA object/list}

\item{sa}{Parameter list generated by \link{runSA}}

\item{threads}{Number of parallel workers.}
}
\value{
A list containing surface grid data and state-specific surface points.
}
\description{
Construct smoothed energy landscape surface objects from GELA output and parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eval.R
\name{GradientDescent}
\alias{GradientDescent}
\title{Gradient descent estimation of model parameters}
\usage{
GradientDescent(data, maxit)
}
\arguments{
\item{data}{Numeric matrix (typically 0/1). Rows are samples and columns are species or taxa.}

\item{maxit}{Integer. Maximum number of gradient descent iterations.}
}
\value{
A list with elements:
\describe{
  \item{h.est}{Numeric vector of estimated implicit environmental parameters.}
  \item{j.est}{Numeric matrix of estimated interaction matrix.}
}
}
\description{
Estimates parameters \code{h} and \code{J} by iterative gradient updates to match
first and second moments of the observed binary data under an maximum
pairwise entropy model.
}
\details{
This function calls \code{Energy()} to compute energies for candidate states.
Ensure that \code{Energy} is available in the package namespace.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{showSSD}
\alias{showSSD}
\title{showSSD: draw stable state diagram}
\usage{
showSSD(gela)
}
\arguments{
\item{gela}{The output of \link{GradELA}. Object containing stable state lists, energies, the factor,
and environment labels as used in the code.}
}
\value{
NULL (produces a plot as side effect).
}
\description{
Plots energy values of stable states across a factor (e.g., environmental or
control parameter) using the output from \code{gela}. Each unique stable state is
drawn as a separate line with points, and a legend is added.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{bin2id}
\alias{bin2id}
\title{bin2id: map a binary vector to a 64base characters}
\usage{
bin2id(bin)
}
\arguments{
\item{bin}{A binary vector (0/1, community composition) to encode.}
}
\value{
A character scalar representing the encoded id.
}
\description{
Encode a binary vector into a base-64 id string index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{Stability}
\alias{Stability}
\title{stability}
\usage{
Stability(
  sa,
  ocmat,
  enmat = NULL,
  seitr = 1000,
  seconvTime = 10000,
  threads = 1,
  reporting = TRUE
)
}
\arguments{
\item{sa}{: output of runSA function.}

\item{ocmat}{: array of the vectors of presence/absence status for each sample.}

\item{enmat}{: array of the vectors of environmental factors for each sample. Normalization required.}

\item{seitr}{: Number of stochastic search iterations.}

\item{seconvTime}{: Maximum number of iterations for each search.}

\item{threads}{: number of cores (threads) used for calculation.}

\item{reporting}{: enable/disable the reporting function.}
}
\value{
A data frame with stability indices. Columns include at least
  \code{energy.gap}, \code{ss.entropy}, \code{e.realize}, \code{e.stable}, \code{state.id}, and \code{stable.state.id}
  (and may include \code{conv_time} and \code{convergence} depending on the branch).
}
\description{
Calculate energy gap and stable state entropy as stability indices and returns a dataframe.

For each sample state, this function identifies the basin (stable state) by \link{Bi},
then computes (i) the energy gap and (ii) stable state entropy
estimated by \link{SSentropy}. 
If \code{enmat} is provided and \code{sa} was estimated with environmental
factors, the basin assignment is computed for each sample-specific \code{env}. 

Note: the returned stable state id corresponds to the basin (stable state) the sample is assigned to.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eval.R
\name{hb.paramgen}
\alias{hb.paramgen}
\title{Randomly generate model parameters for heat-bath sampling}
\usage{
hb.paramgen(nn, ne = 0, j_connectivity = 1)
}
\arguments{
\item{nn}{Integer. Number of units (species, taxa, or members).}

\item{ne}{Integer. Number of environmental factors. If \code{ne > 0}, \code{g} is generated.}

\item{j_connectivity}{Numeric in [0, 1]. Fraction of possible pairwise
interactions set to non-zero.}
}
\value{
A list containing at least:
\describe{
  \item{h}{Named numeric vector of length \code{nn}. Implicit environmental parameters}
  \item{J}{Named numeric matrix (\code{nn x nn}), symmetric. Interaction matrix}
}
If \code{ne > 0}, also includes:
\describe{
  \item{g}{Named numeric matrix (\code{nn x ne}). Explicit environmental parameters}
}
}
\description{
Generates random implicit environmental parameters \code{h} and a interaction matrix \code{J}. 
Optionally generates explicit environmental coefficients \code{g}. 
The \code{j_connectivity} parameter controls the fraction of non-zero off-diagonal interactions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runSA.R
\name{saEndpoint}
\alias{saEndpoint}
\title{Extract endpoint parameters from \link{runSA} output}
\usage{
saEndpoint(fittingMat, ocmat, enmat = NULL)
}
\arguments{
\item{fittingMat}{A list of SA trajectories returned by \code{runSA(getall = TRUE)}.}

\item{ocmat}{Numeric matrix (0/1). Rows are samples and columns are species.}

\item{enmat}{Optional environmental parameter matrix used in fitting.}
}
\value{
A numeric matrix of median endpoint parameters across serials. 
Rownames correspond to samples and Colnames correspond to species list
}
\description{
Summarize the final parameter estimates across SA replicates by averaging
the last recorded block of parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{showIntrGraph}
\alias{showIntrGraph}
\title{showIntrGraph: draw interaction matrix on a PCoA plot}
\usage{
showIntrGraph(ela, sa, env = NULL, th = 0, annot_adj = c(0.75, 2))
}
\arguments{
\item{ela}{\link{ELA} output tuple/list (used to obtain stable states etc.; currently
mainly used for context).}

\item{sa}{Parameter set passed to \link{sa2params}.}

\item{env}{target environmental factor.}

\item{th}{Numeric threshold for drawing links (absolute value).}

\item{annot_adj}{Numeric length-2; horizontal/vertical adjustment for labels.}
}
\value{
NULL.
}
\description{
Computes a 2D PCoA embedding from the interaction matrix (\code{jest}) and overlays
interaction links above a threshold \code{th}. Positive and negative links are
drawn separately, and species names are annotated at their coordinates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_dataset.R
\docType{data}
\name{basemetadata}
\alias{basemetadata}
\title{Environmental metadata for 256 samples}
\format{
A data frame with 256 rows and 2 variables:
}
\source{
Generated for example purposes (replace with actual source if applicable)
}
\usage{
basemetadata
}
\description{
`basemetadata` is a data frame containing environmental parameters
associated with the same 256 samples as in \code{baseabtable}.
}
\details{
Each row corresponds to a sample, and each column represents
an environmental variable.


The rows in this data frame correspond to the same samples and
ordering as in \code{baseabtable}. These variables can be used
for ELA with environmental parameters such as \code{gradELA}
}
\examples{
data(baseabtable)
data(basemetadata)
head(basemetadata)
formatted <- Formatting(baseabtable = baseabtable, basemetadata = basemetadata)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gels.R
\name{MinTippingPath}
\alias{MinTippingPath}
\title{MinTippingPath}
\usage{
MinTippingPath(s1, s2, hg, j, tmax)
}
\arguments{
\item{s1}{Another integer state ID}

\item{s2}{Another integer state ID}

\item{hg}{explicit/implicit environmental factors}

\item{j}{species interaction matrix}

\item{tmax}{Iteration limit.}
}
\value{
Numeric vector of energies along the selected path (or scalar if s1==s2).
}
\description{
Compute a minimum “tipping” energy path between two states.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runSA.R
\name{Findbp}
\alias{Findbp}
\title{Search best hyperparameters for SA fitting (lambda and weight decay) by cross-validation}
\usage{
Findbp(
  ocmat,
  enmat = NULL,
  cvmode = c("10fold", "5fold", "loo"),
  nrepeat = 5,
  we = c(0.001),
  maxlrs = c(0.005),
  totalit = 4000,
  lmd = c(5e-04, 0.001, 0.0025, 0.005, 0.0075),
  rep = 8,
  intv = 100,
  threads = 1,
  tol = 0.03,
  runadamW = TRUE,
  sparse = TRUE,
  fastfitting = TRUE,
  reporting = TRUE
)
}
\arguments{
\item{ocmat}{Numeric matrix (0/1). occurrence matrix of the communities.}

\item{enmat}{Optional environmental parameter matrix.}

\item{cvmode}{Cross-validation mode. One of `"10fold"`, `"5fold"`, or `"loo"`.
`"loo"` performs leave-one-out cross-validation. Default is `"10fold"`.}

\item{nrepeat}{Number of repeated cross-validation runs for k-fold CV.
Ignored when `cvmode = "loo"`. Default is `10`.}

\item{we}{Numeric vector. Candidate weight decay values.}

\item{maxlrs}{Numeric vector. Maximum learning rate.}

\item{totalit}{Integer. Total iterations for SA.}

\item{lmd}{Numeric vector. Candidate \code{lambda} values.}

\item{rep}{Integer. Number of replicates for \link{runSA}.}

\item{intv}{Integer. Recording interval.}

\item{threads}{Integer. Number of cores.}

\item{tol}{Numeric. Error tolerance for finding the best hyperparameters (default: 0.03)}

\item{runadamW}{Logical. Use AdamW.}

\item{sparse}{Logical. Use sparsity penalty.}

\item{fastfitting}{Logical. If \code{TRUE}, returns early iterations within the range
specified by the parameter \code{tol} (default: 0.03) of best error.}

\item{reporting}{Logical. Print progress.}
}
\value{
A list with:
\describe{
  \item{best_candidates}{A named list of best hyperparameter settings and scores.}
  \item{allresults}{All validation results for each grid point.}
}
}
\description{
Repeating cross-validation over a grid of hyperparameters for
simulated annealing-based fitting, evaluates prediction/fitting errors on
held-out samples, and returns candidate best parameter/timepoint combinations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_dataset.R
\docType{data}
\name{baseabtable}
\alias{baseabtable}
\title{Abundance table for 16 taxa across 256 samples}
\format{
A data frame with 256 rows and 16 variables:
}
\source{
Generated for example purposes (replace with actual source if applicable)
}
\usage{
baseabtable
}
\description{
`baseabtable` is a data frame containing abundance for 16 taxa across 256 samples.
}
\details{
Each row corresponds to a sample, and each column represents
the abundance of a specific taxon.


The values represent abundance measurements for each taxon in each sample.
The exact definition of "abundance" (e.g., counts, relative abundance, or normalized values)
depends on the data preprocessing pipeline used.
}
\examples{
data(baseabtable)
head(baseabtable)
formatted <- Formatting(baseabtable = baseabtable)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{showSSD_ggplot}
\alias{showSSD_ggplot}
\title{showSSD_ggplot: draw stable state diagram using ggplot}
\usage{
showSSD_ggplot(gela, plot = TRUE, getSSDobj = TRUE)
}
\arguments{
\item{gela}{The output of \link{GradELA}. Object containing stable state lists, energies, the factor,
and environment labels as used in the code.}

\item{plot}{Logical; if TRUE, show SSD plot}

\item{getSSDobj}{Logical; if TRUE, returns data.frame object used for plotting SSD diagram
Users can manually change the graphical configurations.}
}
\value{
A data frame used for plotting SSD diagram.
}
\description{
Plots energy values of stable states across a factor (e.g., environmental or
control parameter) using the output from \code{gela}. Each unique stable state is
drawn as a separate line with points, and a legend is added.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{gStability}
\alias{gStability}
\title{gstability}
\usage{
gStability(
  sa,
  ocmat,
  enmat = NULL,
  seitr = 1000,
  seconvTime = 10000,
  pruning_type = "relative",
  th = 0.1,
  lower_qr = 0.05,
  threads = 1,
  reporting = TRUE,
  bs_params = NULL
)
}
\arguments{
\item{sa}{: output of runSA function.}

\item{ocmat}{: matrix of presence/absence status for each sample.}

\item{enmat}{: matrix of environmental factors for each sample. Normalization required.}

\item{seitr}{: Number of stochastic search iterations.}

\item{seconvTime}{: Maximum number of iterations for each search.}

\item{pruning_type}{the type of pruning \code{"relative"}(use with \code{th}) or \code{"bootstrap"} (use with \code{lower_qr} and \code{bootstrap_ene})}

\item{th}{: Numerical (0~1). Ratio of the depth of the basin to be pruned with respect to the deepest basin in each pair.}

\item{lower_qr}{Numeric value. Quantile used for constructing confidence intervals from bootstrap samples to check 
whether the estimated energy gaps include zero. Default is \code{lower_qr = 0.05},necessary when \code{type="bootstrap"}.}

\item{threads}{number of cores (threads) used for calculation.#' @param threads : number of cores (threads) used for calculation.}

\item{reporting}{: enable/disable the reporting function.}

\item{bs_params}{output of \link{bootstrap_SA}. list of estimated parameters using bootstrap resampled data set,
necessary when \code{type="bootstrap"}.}
}
\value{
A list of four elements: (1) pruned stability data frame, (2) non-pruned stability data frame,
  (3) auxiliary object for pruned landscape (parameters + stable states table), and
  (4) auxiliary object for sample-wise evaluation (inputs and ELA objects).
}
\description{
Calculate energy gap, energy barrier and stable state entropy.

This function evaluates stability indices for both (i) the pruned and (ii) the non-pruned
energy landscape. For each sample, it assigns the basin by \link{Bi} and computes:
 - energy gap (sample energy minus basin energy),
 - energy barrier (minimum tipping energy connected to the basin minus basin energy),
 - stable state entropy (\code{SSentropy}).

For basin connectivity: when only one stable state exists, tipping energy is treated as \code{Inf}
and therefore energy barrier is returned as \code{Inf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{ELA}
\alias{ELA}
\title{Energy landscape analysis in parallel mode}
\usage{
ELA(
  sa = sa,
  env = NULL,
  SS.itr = 20000,
  FindingTip.itr = 10000,
  threads = 1,
  reporting = TRUE,
  fork = TRUE
)
}
\arguments{
\item{sa}{output of runSA function.}

\item{env}{vector of environmental factors for specifying a single environmental condition (optional).}

\item{SS.itr}{number of iterations of the process to identify the stable states.}

\item{FindingTip.itr}{number of iterations of the process to identify the tipping points.}

\item{threads}{number of cores (threads) used for calculation.}

\item{reporting}{enable/disable the reporting function.}

\item{fork}{logical; if TRUE and \code{threads > 1}, a cluster is created internally for parallel execution.}
}
\value{
A list returned by \code{elaconvert}: (1) stable state IDs, (2) stable state energies,
  (3) tipping point ID matrix/list, and (4) tipping energies matrix/list.
}
\description{
Identifying stable states and tipping points.

This function estimates stable states by repeated steepest-descent searches
\code{SSestimate} and then checks tipping points between all pairs of stable
states using \code{FindingTippingpoint_cpp}. 
The result is summarized and converted by \code{elaconvert}
into a list of stable state IDs, their energies, tipping point IDs, and
tipping energies.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{SteepestDescent}
\alias{SteepestDescent}
\title{SteepestDescent}
\usage{
SteepestDescent(state, alpha, beta)
}
\arguments{
\item{state}{Binary vector representing the initial state.}

\item{alpha}{vector of implicit/explicit environmental effects.}

\item{beta}{Matrix of species interactions.}
}
\value{
A numeric vector of the reached binary stable state and its energy.
}
\description{
Perform steepest descent on the energy landscape.

Starting from an initial binary state, the steepest descent algorithm 
will find a local minimum of the energy landscape. 
The computation is delegated to the C++ implementation
\code{SteepestDescent_cpp}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runSA.R
\name{bootstrap_SA}
\alias{bootstrap_SA}
\title{Parameter fitting with bootstrap sampling of datasets}
\usage{
bootstrap_SA(
  ocmat,
  enmat = NULL,
  bootstrap.it = 2000,
  rep = 128,
  threads = 1,
  we = 0.001,
  totalit = 1000,
  lambda = 0.01,
  runadamW = TRUE,
  maxlr = 0.005,
  sparse = TRUE,
  reporting = TRUE
)
}
\arguments{
\item{ocmat}{Numeric matrix (0/1). Rows are samples and columns are species.}

\item{enmat}{Optional numeric matrix. Environmental variables for each sample.
Should be normalized/standardized beforehand.}

\item{bootstrap.it}{number of bootstrap trials (default: 1000).}

\item{rep}{Integer. Number of parallel runs (replicates) for optimization.}

\item{threads}{Integer. Number of cores used when running in parallel.}

\item{we}{Numeric. Weight decay parameter for AdamW.}

\item{totalit}{Integer. Total number of optimization iterations.}

\item{lambda}{Numeric. Sparsity penalty parameter (used when \code{sparse = TRUE}).}

\item{runadamW}{Logical. If \code{TRUE}, use AdamW-style optimization.}

\item{maxlr}{Numeric. Maximum learning rate.}

\item{sparse}{Logical. If \code{TRUE}, apply sparsity constraint.}

\item{reporting}{Logical. If \code{TRUE}, print progress messages.}
}
\value{
a list of SA parameters estimated with bootstrap resampling (length \code{bootstrap.it}).
}
\description{
Perform \link{runSA} algorithm and estimate parameters using bootstrap resampled
data set (\code{ocmat} and \code{enmat}). 
The basic configuration is same as \link{runSA}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{showGELA3D}
\alias{showGELA3D}
\title{GELA3D: draw 3D plot of energy landscape}
\usage{
showGELA3D(gelsobj, theta = 25, phi = 50)
}
\arguments{
\item{gelsobj}{List with two elements: surface data frame and line data frame.}

\item{theta}{numeric value. plot3D option: Azimuth angle (in degrees) controlling rotation around the z-axis.}

\item{phi}{numeric value. plot3D option: Elevation angle (in degrees) controlling the vertical viewing position.}
}
\value{
NULL (produces a 3D plot as side effect).
}
\description{
Draws a 3D surface of the energy landscape and overlays
line trajectories using \pkg{plot3D} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{PCplot}
\alias{PCplot}
\title{PCplot:Scatter plot of community compositions colored by basin label}
\usage{
PCplot(ocmat, sa = sa, env = NULL, ssrep = NULL, pruned = TRUE, export = FALSE)
}
\arguments{
\item{ocmat}{Data frame (or matrix) of community compositions (rows = samples,
columns = species/features).}

\item{sa}{Parameter set passed to \link{sa2params}.}

\item{env}{Optional environment object passed to \link{sa2params}.}

\item{ssrep}{Data frame-like object providing stable state label mapping with
two columns: \code{stablestates} and \code{stablestates2}.}

\item{pruned}{Logical; if TRUE, use \code{stablestates2} for coloring.}

\item{export}{Logical; if TRUE, return the data used for plotting.}
}
\value{
NULL invisibly (or a data frame when \code{export = TRUE}).
}
\description{
Performs PCA on the community composition matrix (\code{ocmat}) and plots the first
two PCs. Points are colored by the stable state (basin) label estimated
from \link{Bi}. If \code{pruned = TRUE}, colors correspond to the
pruned/merged basin labels in \code{ssrep} (\code{stablestates2}), otherwise to the
original labels (\code{stablestates}). When \code{export = TRUE}, the plotting data
frame (stable-state labels + PCs + merged labels) is returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{bootstrap_ELA}
\alias{bootstrap_ELA}
\title{Energy landscape analysis with bootstrap resampled SA results}
\usage{
bootstrap_ELA(
  bs_params,
  ocmat,
  ela = NULL,
  env = NULL,
  enmat = NULL,
  threads = 1,
  reporting = FALSE
)
}
\arguments{
\item{bs_params}{output of \link{bootstrap_SA}. list of estimated parameters using bootstrap resampled data set.}

\item{ocmat}{community composition matrix.}

\item{ela}{representative ELA result (output of \link{ELA}). If \code{NULL}, the function returns the most representative one.}

\item{env}{target environmental factor.}

\item{enmat}{array of the vectors of environmental factors for each sample. Normalization required.}

\item{threads}{number of cores (threads) used for calculation.}

\item{reporting}{enable/disable the reporting function.}
}
\value{
A list with two elements:
  (1) representative pruned ELA result (stable states, energies, tipping points, tipping energies),
  (2) a data frame of bootstrap energies (rows = state IDs, columns = trials) with \code{NaN} for missing states.
}
\description{
Running energy landscape analysis with bootstrap sampling of estimated saresults

Performs \link{ELA} and \link{ELPruning} using SA fitting results based on 
the bootstrap resampled occurrence and environmental data using \link{bootstrap_SA}.

Basically, representative ELA results with original occurrence and environmental data should be given by \code{ela=ela},
alternatively the function can select a representative parameter where commonly appearing stable states are most frequent.

The function computes bootstrap energy distributions for all stable states and tipping points
appearing in the specified pruned landscape. Energies that are not present in a given trial
are stored as \code{NaN} in the summary table.
}

\name{ELAplus-package}
\alias{ELAplus-package}
\alias{ELAplus}
\docType{package}
\title{
\packageTitle{ELAplus}
}
\description{
\packageDescription{ELAplus}
}
\details{
This package provides tools for energy landscape analysis (ELA),
a framework for analyzing dynamical systems represented as weighted
networks. The main functions support construction of energy landscapes,
by fitting the community composition data to a pairwise maximum entropy model
(or its extension) and subsequent identification of stable states, and
analysis of transitions between community compositions.
}
\author{
\packageAuthor{ELAplus}

Maintainer: \packageMaintainer{ELAplus}
}
\references{
Suzuki, K., Nakaoka, S., Fukuda, S., & Masuya, H. (2021). 
"Energy landscape analysis elucidates the multistability of ecological communities across environmental gradients." 
Ecological Monographs.
\doi{10.1002/ecm.1469}
\url{https://esajournals.onlinelibrary.wiley.com/doi/abs/10.1002/ecm.1469}
}
\keyword{models}
\keyword{multivariate}
\seealso{
GitHub: \url{https://github.com/kecosz/rELA},
Portal: \url{https://elaportal.org}
}
\examples{
# load sample dataset
data("baseabtable")
data("basemetadata")
# Formatting dataset
formatted <- Formatting(baseabtable, basemetadata, 0, c(0.05, 0.05, 0.95))
ocvecs <- formatted[[1]]
envecs <- formatted[[3]]
# Example (without environmental parameters)
# Fitting the data to the extended maximum pairwise entropy model
# without explicit environmental parameters
sa <- runSA(
  ocvecs,
  enmat   = NULL,
  rep     = 1,
  getall  = FALSE,
  lambda  = 0.01,
  we      = 0.001,
  maxlr = 0.005,
  totalit = 1000
)

# Using the fitted model, normal ELA is run.
ela <- ELA(sa, env=NULL,
           threads=1,SS.itr=2000,
           FindingTip.itr=1000,
           reporting=TRUE)
# Pruning stable states
elap <- ELPruning(ela, th=0.1)
# show disconnectivity graph
showDG(elap[[1]], ocvecs, "test")
# show species interaction graph
showIntrGraph(elap[[1]], sa, th=0.01,
              annot_adj=c(0.75, 2.00))
              
# Example (with environmental parameters)   
# Find best parameter sets for the machine learning (with envecs)
\donttest{
bpresult <- Findbp(ocvecs,enmat=envecs,rep=1,threads=1,nrepeat=1,cvmode="5fold")
bp <- bpresult[[1]]
bpm <- as.numeric(unlist(strsplit(names(bp)[1], split = "_")))
# Fitting the data to the extended maximum pairwise entropy model
# with explicit environmental parameters
sa <- runSA(
  ocvecs,
  enmat   = envecs,
  rep     = 16,
  getall  = FALSE,
  lambda  = bpm[[1]],
  we      = bpm[[2]],
  maxlr = bpm[[3]],
  totalit = bpm[[4]]
)
# Using the fitted model, gradient ELA is run.
gela <- GradELA(
  sa      = sa,
  eid     = "factor.1",
  enmat   = envecs,
  env     = NULL,
  range   = c(0, 1),
  steps   = 32,
  th      = 0.05,
  threads = 2
)
showSSD(gela)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{id2bin}
\alias{id2bin}
\title{id2bin: map a 64base characters to a binary vector}
\usage{
id2bin(id, s)
}
\arguments{
\item{id}{A character scalar encoded by \link{bin2id}.}

\item{s}{Integer number of bits to return.}
}
\value{
An integer vector of 0/1 bits of length s.
}
\description{
Decode a base-64 id string index to a binary vector of length s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runSA.R
\name{runSA}
\alias{runSA}
\title{Stochastic approximation for parameter fitting}
\usage{
runSA(
  ocmat,
  enmat = NULL,
  rep = 128,
  threads = 1,
  we = 0.001,
  totalit = 1000,
  lambda = 0.01,
  intv = 100,
  runadamW = TRUE,
  maxlr = 0.005,
  sparse = TRUE,
  getall = FALSE,
  reporting = TRUE
)
}
\arguments{
\item{ocmat}{Numeric matrix (0/1). Rows are samples and columns are species.}

\item{enmat}{Optional numeric matrix. Environmental variables for each sample.
Should be normalized/standardized beforehand.}

\item{rep}{Integer. Number of parallel runs (replicates) for optimization.}

\item{threads}{Integer. Number of cores used when running in parallel.}

\item{we}{Numeric. Weight decay parameter for AdamW.}

\item{totalit}{Integer. Total number of optimization iterations.}

\item{lambda}{Numeric. Sparsity penalty parameter (used when \code{sparse = TRUE}).}

\item{intv}{Integer. Interval of iterations at which intermediate results are recorded.}

\item{runadamW}{Logical. If \code{TRUE}, use AdamW-style optimization.}

\item{maxlr}{Numeric. Maximum learning rate.}

\item{sparse}{Logical. If \code{TRUE}, apply sparsity constraint.}

\item{getall}{Logical. If \code{TRUE}, return all replicate trajectories.}

\item{reporting}{Logical. If \code{TRUE}, print progress messages.}
}
\value{
If \code{getall = TRUE}, a list of SA trajectories (length \code{rep}).
Otherwise, a list with two elements, \code{c(params, enmat)} is returned.
\code{params} is the concatenated matrix of the estimated parameter set:
\describe{
  \item{h}{Implicit environmental effects on species occurrence}
  \item{J}{Species-species interaction matrix}
  \item{g}{(If \code{enmat} is not \code{NULL}) Explicit effects of the environmental factors.}
}
}
\description{
Estimate species interaction parameters and (optionally) environmental
preference parameters by stochastic approximation (SA). The function runs
multiple SA optimizations in parallel (when \code{threads > 1}) using
\pkg{foreach}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eval.R
\name{OnestepHBS}
\alias{OnestepHBS}
\title{Perform one heat-bath update step}
\usage{
OnestepHBS(y, logmo)
}
\arguments{
\item{y}{Numeric matrix of size \code{nproc x nn} (typically 0/1), current states.}

\item{logmo}{Numeric matrix of probabilities compatible with \code{y}. Each row
corresponds to a row in \code{y}; entries are used to sample 0/1 updates.}
}
\value{
A numeric matrix with the same dimensions as \code{y}, updated states.
}
\description{
Updates each row (sample) of a binary state matrix by choosing one random
position per row and resampling it according to precomputed heat-bath
probabilities.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{Energy}
\alias{Energy}
\title{Energy}
\usage{
Energy(state, alpha, beta)
}
\arguments{
\item{state}{Binary vector representing a state (community composition).}

\item{alpha}{vector of implicit/explicit environmental effects on species.}

\item{beta}{Matrix of species interactions.}
}
\value{
A numeric scalar giving the energy of the state.
}
\description{
Calculate energy of a given state.

Computing energy in \code{state} using given parameters,\code{alpha} and \code{beta}
based on the (extended) maximum pairwise entropy model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{SSentropy}
\alias{SSentropy}
\title{SSentropy}
\usage{
SSentropy(state, ss, alpha, beta, seitr = 1000, convTime = 10000)
}
\arguments{
\item{state}{Binary vector (or matrix with one row) representing the current state.}

\item{ss}{Matrix of stable states.}

\item{alpha}{Vector of implicit or explicit environmental effects.}

\item{beta}{Matrix of species interactions.}

\item{seitr}{Number of stochastic search iterations.}

\item{convTime}{Maximum number of iterations for each search.}
}
\value{
A numeric matrix with three columns:
  (1) stable state entropy,
  (2) convergence time,
  (3) number of converged searches.
}
\description{
Estimate stable state entropy for a given state.

Evaluating the entropy of transitions among stable states
starting from a given community state. 

The returned value includes the estimated entropy, the convergence time,
and the number of successful convergences within the given iteration limits.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eval.R
\name{HeatBath}
\alias{HeatBath}
\title{Heat-bath sampling of binary states}
\usage{
HeatBath(steps, nproc, h, J, g = NULL)
}
\arguments{
\item{steps}{Integer. Number of heat-bath iterations to run.}

\item{nproc}{Integer. Number of parallel samples (rows) to maintain.}

\item{h}{Numeric vector of length \code{nn}, implicit environmental parameters.}

\item{J}{Numeric matrix of size \code{nn x nn}, interaction matrix.}

\item{g}{Optional numeric matrix of size \code{nn x ne}, environmental coefficients.}
}
\value{
A list with components:
\describe{
  \item{y}{Numeric matrix (\code{nproc x nn}) of sampled states (0/1).}
  \item{enmat}{Numeric matrix (\code{nproc x ne}) of environmental values, or \code{NULL}
    if \code{g} is \code{NULL}.}
}
}
\description{
Generates samples of binary state vectors using heat-bath updates under
an extended maximum pairwise entropy model parameterized 
by implicit environmental parameters \code{h} and interaction matrix \code{J}.
Optionally includes environmental effects \code{g}, in which case an environmental
matrix is generated internally.
}
\examples{
set.seed(1)
nn <- 5
pars <- hb.paramgen(nn)
res <- HeatBath(steps = 10, nproc = 8, h = pars[[1]], J = pars[[2]])
str(res)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{Formatting}
\alias{Formatting}
\title{Formatting}
\usage{
Formatting(
  baseabtable,
  basemetadata = NULL,
  normalize = 1,
  parameters = c(0.05, 0.05, 0.95),
  grouping = 0,
  grouping_th = 0,
  reporting = TRUE
)
}
\arguments{
\item{baseabtable}{Abundance table (samples x species) as data.frame/matrix.}

\item{basemetadata}{Optional environmental metadata table (samples x factors); NULL/empty for none.}

\item{normalize}{Integer 0 or 1. If 1, normalize within each sample to sum to 1.}

\item{parameters}{Numeric vector/list of thresholds (following 3 parameters).
1. "0/1 threshold"
2. "min occurrence threshold"
3. "max occurrence threshold"
(e.g., c(0.05, 0.05, 0.95) : species abundance > 0.05 is regarded as present,
and species occurring in more than 0.05 and less than 0.95 in the all samples are kept.)}

\item{grouping}{Integer 0 or 1. If 1, group similar occurrence patterns.}

\item{grouping_th}{Numeric threshold for grouping (relative Hamming distance).}

\item{reporting}{Logical. If TRUE, print processing summary.}
}
\value{
A list: (ocmatrix, abmatrix, fctmatrix, samplelabel, specieslabel, factorlabel).
}
\description{
Generate occurrence(binary)/abundance/factor matrices and labels from input abundance/metadata tables.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runSA.R
\name{plotSAtest}
\alias{plotSAtest}
\title{plotSAtest: plot the results of SA fitting (Findbp output)}
\usage{
plotSAtest(sa_results, ylim = c("auto", "auto"))
}
\arguments{
\item{sa_results}{Output list from \link{Findbp}.}

\item{ylim}{Numeric length-2 vector (ymin, ymax). Default: \code{c("auto","auto")}.}
}
\value{
A \pkg{ggplot2} object.
}
\description{
Plot mean validation scores over time for the results returned by
\link{Findbp}, including \code{95\%} confidence intervals.
}
\details{
This function uses \pkg{ggplot2} and tidyverse-style data manipulation.
Ensure required packages are installed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runSA.R
\name{sa2params}
\alias{sa2params}
\title{Convert SA output to model parameter objects}
\usage{
sa2params(sa, env = NULL)
}
\arguments{
\item{sa}{A list in the form \code{list(param_matrix, enmat)} where
\code{param_matrix} contains estimated parameters.}

\item{env}{Optional numeric vector of environmental values (column names correspond
to environmental factors).Default is NULL.}
}
\value{
A list with elements \code{he} (h), \code{je} (J), \code{ge} (g or NULL),
  and \code{hge} (combined field).
}
\description{
Convert the output of \link{runSA} / \link{saEndpoint} into parameter
objects in (extended) maximum pairwise entropy model:
\code{h}, \code{J}, and optionally \code{g} when environmental parameter
matrix is specified.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{tptable}
\alias{tptable}
\title{tptable}
\usage{
tptable(ela, ocmat)
}
\arguments{
\item{ela}{output of \link{ELA} or the first return value of \link{ELPruning}.}

\item{ocmat}{community composition matrix.}
}
\value{
A data frame with columns \code{TP}, \code{SS1}, \code{SS2}, \code{Energy}, 
and one column per species (binary state vector).
If no tipping points exist, returns NULL (invisibly) after printing a message.
}
\description{
Summary of tipping points.

Builds a table of tipping points (transition states) and their energies from an ELA output.
This function lists only finite tipping points
(i.e., entries not equal to \code{Inf}), and returns the corresponding binary state vectors.

Two stable states are required at least.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{showDG}
\alias{showDG}
\title{showDG: draw DisconnectivityGraph}
\usage{
showDG(
  ela,
  oc,
  label = "",
  bootstrap_ene = NULL,
  min_pval = 0.05,
  plot = TRUE,
  scale_labels = c(0.05, 0.1, 0.2, 0.5),
  getGraphObj = FALSE
)
}
\arguments{
\item{ela}{ELA output tuple/list used by \code{GraphObj}.}

\item{oc}{Community composition matrix/data frame (used to infer \code{s = ncol(oc)}).}

\item{label}{Character; plot title.}

\item{bootstrap_ene}{Optional bootstrap energy matrix/data frame for computing statistical significance levels of energy gaps.}

\item{min_pval}{Numeric; minimum p-value shown in the visualization. This is
usually set to the minimum detection level in bootstrap sampling
(e.g., if the number of bootstrap iterations is 1000, the minimum level is
1/1000 = 0.001).}

\item{plot}{Logical; if TRUE, show disconnectivity graph}

\item{scale_labels}{Numeric vector of values between 0 and 1; labels for the
p-value scale. These values are converted to log-scaled values for plotting.}

\item{getGraphObj}{Logical; if TRUE, returns data.frame object used for plotting DisconnectivityGraph
Users can manually change the graphical configurations.}
}
\value{
A data frame containing energy values, xpositions, pvalues for energy gaps in identified stable states and tipping points.
}
\description{
Computing graph objects from \code{ela} (and optionally
bootstrap energies) and draw the disconnectivity graph for a given community
composition matrix \code{oc}. More than two stable state are required.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{Bi}
\alias{Bi}
\title{Bi}
\usage{
Bi(xi, h, j)
}
\arguments{
\item{xi}{Binary vector representing the initial state.}

\item{h}{vector of implicit/explicit environmental effects on species.}

\item{j}{Matrix of species interactions.}
}
\value{
A list with two elements:
  (1) integer stable state ID (basin assignment),
  (2) energy of the corresponding stable state.
}
\description{
Identify basin (stable state) and its energy for a given state.

Applies \link{SteepestDescent} to the input state and returns the basin (stable state) reached,
together with its energy. The stable state is encoded as an integer ID using \link{bin2id}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runELA.R
\name{sstable}
\alias{sstable}
\title{sstable}
\usage{
sstable(ela, ocmat)
}
\arguments{
\item{ela}{output of ELA or the first return value of ELPruning.}

\item{ocmat}{community composition matrix.}
}
\value{
A data frame with columns \code{ID}, \code{Energy}.
}
\description{
Summary of stable states.

Builds a table of stable states and their energies from an ELA output.
The returned data frame contains stable state IDs and the corresponding binary
presence/absence vector (per species). This function is typically used for
inspecting stable states after pruning (if \code{ela} is pruned) or before pruning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{DisconnectivityGraph}
\alias{DisconnectivityGraph}
\title{DisconnectivityGraph: Generate a disconnectivity graph}
\usage{
DisconnectivityGraph(
  grobj,
  s,
  DG_sample = "DG_sample",
  annot_adj = c(0.75, 2),
  min_pval = 0.05,
  scale_labels = c(0.05, 0.1, 0.2, 0.5),
  show_pvalue = FALSE,
  getGraphObj = FALSE,
  plot = TRUE
)
}
\arguments{
\item{grobj}{A data frame returned by \code{GraphObj}.}

\item{s}{Integer; number of species/features (used for labeling by \link{id2bin}).}

\item{DG_sample}{Character; title of the plot.}

\item{annot_adj}{Numeric vector of length 2; horizontal and vertical adjustment for labels.}

\item{min_pval}{Numeric (must be > 0); minimum p-value shown in the visualization. This is
usually set to the minimum detection level in bootstrap sampling
(e.g., if the number of bootstrap iterations is 1000, the minimum level is
1/1000 = 0.001).}

\item{scale_labels}{Numeric vector of values between 0 and 1; labels for the
p-value scale. These values are converted to log-scaled values for plotting.}

\item{show_pvalue}{Logical; if \code{TRUE}, plots approximate p-values based
on the energy gaps between each stable state and its corresponding tipping
point estimated by bootstrap resampling.}

\item{getGraphObj}{Logical; if \code{TRUE}, returns the data frame used for
plotting the disconnectivity graph, so that users can manually modify the
graphical settings.}

\item{plot}{Logical; if TRUE, show disconnectivity graph}
}
\value{
If \code{getGraphObj = TRUE}, returns a data frame containing energy values,
x-positions, and p-values for the energy gaps of the identified stable states
and tipping points. Otherwise, returns \code{NULL}.
}
\description{
Draws a disconnectivity graph from the data frame returned by \code{GraphObj}.
Nodes are plotted at their assigned x-positions and energy levels, and are
connected to their parent components by paths.

If \code{show_pvalue = TRUE}, approximate p-values for the energy gaps of
stable states are displayed. The p-values are shown on a log10 scale using
the values specified in \code{scale_labels}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{deltaE_histogram}
\alias{deltaE_histogram}
\title{Plot histograms of bootstrap deltaE values}
\usage{
deltaE_histogram(bootstrap_ene, grobj_to_plot, getData = TRUE)
}
\arguments{
\item{bootstrap_ene}{A data frame containing energies in stable states and tipping points,
output of \link{bootstrap_ELA}.}

\item{grobj_to_plot}{A data frame containing stable state - tipping point relationship,
output of \link{showDG}}

\item{getData}{Logical. If `TRUE`, return the computed deltaE values as a list.
If `FALSE`, return `NULL`.}
}
\value{
If `getData = TRUE`, a list of numeric vectors, one per deltaE for each stable state.
If `getData = FALSE`, `NULL`.
}
\description{
For each stable state, this function computes the difference between the energy values 
of the corresponding tipping point and the stable state across bootstrap samples,
then draws a histogram of the resulting deltaE values.
}
